/**
 * @file nblock_list.c
 *
 * Implementation of non-blocking linked list.
 *
 * This implementation supports concurrent linearizable insertion and deletion operation using low-level atomic 
 * primitive compare-and-swap (CAS). CAS isn't directly supported instruction on MIPS architecture, however it can 
 * be replaced using the instruction LL (load-linked) and SC (store-conditional) in the right way.
 * 
 * Algorithms was originally described by Timothy L. Harris:
 * Harris T. L.: A Pragmatic Implementation of Non-blocking Linked-Lists, Proceedings of the 15th International 
 * Conference on Distributed Computing, LNCS 2180, Springer-Verlag, 2001
 
 * @see http://research.microsoft.com/en-us/um/people/tharris/papers/2001-disc.pdf
 */
 
#include <include/c.h>
#include <mm/malloc.h>
#include <drivers/printer.h>
#include <adt/nblock_list.h>


/**
 * Implementation of atomic CAS operation on MIPS architecture.
 * The atomic instruction CAS is replaced using instructions LL and SC.
 *
 * @param accum The memory location.
 * @param dest The old value.
 * @param newval The new value.
 * @return True, if the operation was successful, false else.
 * @see http://en.wikipedia.org/wiki/Compare-and-swap
 */
bool CAS (void** accum, void* dest, void* newval);
bool CAS (void** accum, void* dest, void* newval) 
{
	/*
	 * bool CAS(void** accum, void* dest, void* newval)
	 * {
	 *	if(*accum == dest)
	 *	{
	 *		*accum = newval;
	 *		return true;
	 *	}
	 *	else
	 *	{
	 *		return false;
	 *	}
	 * }
	 */
	
	/* Just for associate with register. */
        native_t result = 1;
	native_t cas_tmp = 0;
        
        asm volatile
        (
                ".set push\n"
                ".set noat\n"
                ".set noreorder\n"

                "ll %[cas_tmp], %[accum]\n"
                "bne %[cas_tmp], %[dest], 1f\n"
                "nop\n"
                "move %[result], %[newval]\n"
                "sc %[result], %[accum]\n"
                "b 2f\n"
                "nop\n"
                
                "1:  li %[result], 0\n"
                
                "2:  sync\n"
                
                ".set pop\n"
                :
                [result] "=r" (result),
                [cas_tmp] "+r" (cas_tmp),
                [dest] "+r" (dest),
                [newval] "+r" (newval),
                [accum] "+m" (*accum)
                :
                /* No input. */
                :
                "memory"
        );
        
        return result;         
}


/**
 * Checks, if the reference is marked.
 *
 * @param ref The reference.
 * @return True if the given reference is marked, else false.
 * @see NBLOCK_MARKED_REF
 */
bool nblock_is_marked_ref (void** ref);
bool nblock_is_marked_ref (void** ref)
{
	return (bool) ((uint32_t)ref & NBLOCK_MARKED_REF);
}


/**
 * Converts the given reference to marked reference.
 *
 * @param ref The reference.
 * @return The marked reference.
 * @see NBLOCK_MARKED_REF
 */
void** nblock_get_marked_ref (void** ref);
void** nblock_get_marked_ref (void** ref)
{        
        return (void**)((uint32_t) ref | NBLOCK_MARKED_REF);
}


/**
 * Converts the given reference to unmarked reference.
 *
 * @param ref The reference.
 * @return The unmarked reference.
 * @see NBLOCK_MARKED_REF
 */
void** nblock_get_unmarked_ref (void** ref);
void** nblock_get_unmarked_ref (void** ref)
{
        return (void**)((uint32_t) ref & (~NBLOCK_MARKED_REF));
}


/**
 * Simple debug tool for dumping the non-blocking linked list.
 * Prints the contents (pointer, value, marked) of all items in the list to the console.
 *
 * @param list The non-blocking linked list.
 */
void nblock_list_dump (nblock_list_t* list);
void nblock_list_dump (nblock_list_t* list)
{
	nblock_item_t* it = list->head->next;
	while(it != list->tail)
	{
		printk_not_sync ("Item ptr %p, value %x, marked %u.\n", it, it->key, nblock_is_marked_ref ((void**)it));
		it = it->next;
	}
}


/** 
 * Initialize an item of the list.
 *
 * @param item The item to initialize.
 * @param key The key.
 * @param data The data.
 */
void nblock_item_init (nblock_item_t* item, unsigned long key, void* data);
void nblock_item_init (nblock_item_t* item, unsigned long key, void* data)
{
	item->key = key;
	item->data = data;
	item->next = NULL;
}


/** 
 * Initialize a list.
 *
 * @param list The list to initialize.
 */
void nblock_list_init (nblock_list_t* list)
{
	/* Alloc the head of list with magic control key and data. */
	list->head = malloc (sizeof(nblock_item_t));
	nblock_item_init (list->head, 0x00000000, NULL);
	
	/* Alloc the tail of list with magic control key and data. */
	list->tail = malloc (sizeof(nblock_item_t));
	nblock_item_init (list->tail, 0xffffffff, NULL);
	
	list->head->next = list->tail;
}


/**
 * Internal search function for the non-blocking linked list.
 * The list is sorted by the key value and the search operation finds the left and right items for a given search 
 * key.
 *
 * @param list The list.
 * @param search_key The key for search.
 * @param left_item The searched left item.
 * @return The searched right item.
 * @see nblock_is_marked_ref
 * @see nblock_get_marked_ref
 * @see nblock_get_unmarked_ref
 * @see CAS
 */
nblock_item_t* nblock_list_search (nblock_list_t* list, unsigned long search_key, nblock_item_t** left_item);
nblock_item_t* nblock_list_search (nblock_list_t* list, unsigned long search_key, nblock_item_t** left_item)
{
	nblock_item_t* left_item_next = NULL;
	nblock_item_t* right_item = NULL;
	
	search_again:
		do
		{
			nblock_item_t* t = list->head;
			nblock_item_t* t_next = list->head->next;
			
			/* Find left_item and right_item. Marked items are ignored. */
			do
			{
				if (!(nblock_is_marked_ref ((void**)t_next)))
				{
					*left_item = t;
					left_item_next = t_next;
				}
				
				/* We can't use the reference until is unmarked. */
				t = (nblock_item_t*)(nblock_get_unmarked_ref ((void**)t_next));
				if (t == list->tail)
				{
					break;
				}
				
				t_next = t->next;
			}
			while ((nblock_is_marked_ref ((void**)t_next)) || (t->key < search_key));
			right_item = t;
			
			/* Check items are adjacent. */
			if (left_item_next == right_item)
			{
				if ((right_item != list->tail) && 
					(nblock_is_marked_ref ((void**)right_item->next)))
				{
					goto search_again;	
				}
				else
				{
					return right_item;
				}
			}
			
			/* Remove one or more marked items. */
			if (CAS ((void**)&((*left_item)->next), (void*)left_item_next, (void*)right_item))
			{
			
				if ((right_item != list->tail) && nblock_is_marked_ref ((void**)right_item->next))
				{
					free (right_item);
					goto search_again;
				}
				else
				{
					return right_item;
				}
			}
		}
		while (true);
}


/** 
 * Insert a new item to the list.
 *
 * @param list The list.
 * @param key The new key.
 * @param data The data.
 * @return True, if the new key was successfully added, false else.
 * @see CAS
 * @see nblock_list_search
 */
bool nblock_list_insert (nblock_list_t* list, unsigned long key, void* data)
{
	nblock_item_t* right_item = NULL;
	nblock_item_t* left_item = NULL;
	
	do
	{
		right_item = nblock_list_search (list, key, &left_item);
		if ((right_item != list->tail) && (right_item->key == key))
		{
			/* The key already exists. */
			return false;
		}
		
		/* Create new structure.*/
		nblock_item_t* new_item = malloc (sizeof (nblock_item_t));
		nblock_item_init (new_item, key, data);
		
		new_item->next = right_item;
		if (CAS ((void**)&(left_item->next), (void*)right_item, (void*)new_item))
		{
			return true;
		}
		
	}
	while (true);
}


/** 
 * Delete the item with given key.
 * Return data in the given @data_ptr - must be valid. 
 *
 * @param list The list.
 * @param search_key The key.
 * @param data_ptr Pointer to data.
 * @return True, if the item with given key was successfully removed, false else.
 * @see nblock_list_search
 * @see nblock_is_marked_ref
 * @see CAS
 */
bool nblock_list_remove (nblock_list_t* list, unsigned long search_key, void** data_ptr)
{
	nblock_item_t* left_item = NULL;
	nblock_item_t* right_item = NULL;
	nblock_item_t* right_item_next = NULL;

	do
	{
		right_item = nblock_list_search (list, search_key, &left_item);
		if ((right_item == list->tail) || (right_item->key != search_key))
		{
			/* There isn't any item with given key. */
			return false;
		}
		right_item_next = right_item->next;
		
		if (!(nblock_is_marked_ref ((void**)right_item_next)))
		{
			if (CAS ((void**)&(right_item->next), (void*)right_item_next, 
				(nblock_get_marked_ref ((void**)right_item_next))))
			{
				break;
			}
		}
	}
	while (true);
	
	/* Return the data. */
	if (data_ptr)
	{
		(*data_ptr) = right_item->data;	
	}
	
	if (!(CAS ((void**)&(left_item->next), (void*)right_item, (void*)right_item_next)))
	{
		/* Call only for removing. */
		right_item = nblock_list_search (list, right_item->key, &left_item);
	}
	else
	{
		free (right_item);
	}
		
	return true;
}


/** 
 * Find the item with given key.
 * Return the data in @data_ptr - must be valid.
 *
 * @param list The list.
 * @param search_key The key.
 * @param data_ptr The pointer to data.
 * @return True, if the item with given key was successfully removed, false else.
 * @see nblock_list_search
 */
bool nblock_list_find (nblock_list_t* list, unsigned long search_key, void** data_ptr)
{
	nblock_item_t* left_item = NULL;
	nblock_item_t* right_item = NULL;
	
	right_item = nblock_list_search (list, search_key, &left_item);
	if ((right_item == list->tail) || (right_item->key != search_key))
	{
		return false;
	}
	else
	{
		if (data_ptr)
		{
			(*data_ptr) = right_item->data;	
		}
		return true;
	}
}
