/**
 * @file nblock_list.h
 *
 * Implementation of non-blocking linked list.
 *
 * This implementation supports concurrent linearizable insertion and deletion operation using low-level atomic 
 * primitive compare-and-swap (CAS). CAS isn't directly supported instruction on MIPS architecture, however it can 
 * be replaced using the instruction LL (load-linked) and SC (store-conditional) in the right way.
 * 
 * Algorithms was originally described by Timothy L. Harris:
 * Harris T. L.: A Pragmatic Implementation of Non-blocking Linked-Lists, Proceedings of the 15th International 
 * Conference on Distributed Computing, LNCS 2180, Springer-Verlag, 2001
 
 * @see http://research.microsoft.com/en-us/um/people/tharris/papers/2001-disc.pdf
 */

#ifndef NBLOCK_LIST_H_
#define NBLOCK_LIST_H_

/**
 * The logical delete notation.
 * The reference contained in the next field of an item may be in one of two states: Marked or unmarked. An item 
 * is marked if and only if its next field is marked. Marked references are indicated by an otherwise-unused 
 * low-order bit in each reference, in this case the first low-order bit. Intuitively a marked item is one which 
 * should be ignored because some process is deleting it.
 */
#define NBLOCK_MARKED_REF	(1 << 0)


/**
 * Structure of an item of non-blocking linked list.
 */
typedef struct nblock_item 
{
	/** The key. */
	unsigned long key; 
	
	/** The data. */
	void* data;
	
	/** The next item in the list. */
	struct nblock_item * next;
	
} nblock_item_t;


/**
 * Structure of non-blocking linked list.
 */
typedef struct nblock_list 
{
	/** The first item in the list. */
	nblock_item_t* head;
	
	/** The last item in the list. */
	nblock_item_t* tail;
	
} nblock_list_t;


/* Commented with implementation. */

void 		nblock_list_init 	(nblock_list_t* list);
bool 		nblock_list_insert 	(nblock_list_t* list, unsigned long key, void* data);
bool 		nblock_list_remove 	(nblock_list_t* list, unsigned long search_key, void** data_ptr);
bool 		nblock_list_find 	(nblock_list_t* list, unsigned long search_key, void** data_ptr);
void 		nblock_list_dump 	(nblock_list_t* list);

#endif
