/**
 * @file sync_list.c
 *
 * Implementation of global synchronized single linked list.
 * Synchronization is provided using the mutex synchronization primitive. There is a global mutex, which was locked,
 * when someone operates with the list.
 * @link kernel/sync/mutex.h Kernel mutex implementation @endlink
 */
 

#include <include/c.h>
#include <mm/malloc.h>
#include <drivers/printer.h>
#include <sync/mutex.h>
#include <adt/sync_list.h>


/** A global mutex for locking the list. */
struct mutex sync_list_lock;

/** 
 * Initialize an item of the list.
 *
 * @param item The item to initialize.
 * @param key The key.
 * @param data The data.
 */
void sync_item_init (sync_item_t* item, unsigned long key, void* data);
void sync_item_init (sync_item_t* item, unsigned long key, void* data)
{
	item->key = key;
	item->data = data;
	item->next = NULL;
}


/** 
 * Initialize a synchronized linked list.
 *
 * @param list The list to initialize.
 */
void sync_list_init (sync_list_t* list)
{
	mutex_init (&sync_list_lock);
	
	mutex_lock (&sync_list_lock);
	
	/* Alloc the head of list with magic control key and data. */
	list->head = malloc (sizeof (sync_item_t));
	sync_item_init (list->head, 0x00000000, NULL);
	
	/* Alloc the tail of list with magic control key and data. */
	list->tail = malloc (sizeof (sync_item_t));
	sync_item_init (list->tail, 0xffffffff, NULL);
	
	list->head->next = list->tail;
	
	mutex_unlock (&sync_list_lock);
}


/** 
 * Internal search function for the global synchronized linked list.
 * The list is sorted by the key value and the search operation finds the left and right items for a given search key.
 *
 * @param list The list.
 * @param search_key The key for search.
 * @param left_item The searched left item.
 * @return The searched right item.
 */
sync_item_t* sync_list_search (sync_list_t* list, unsigned long search_key, sync_item_t** left_item);
sync_item_t* sync_list_search (sync_list_t* list, unsigned long search_key, sync_item_t** left_item)
{
	sync_item_t* t = list->head;
	sync_item_t* t_next = list->head->next;

	/* Find left_item and right_item. */
	do
	{
		*left_item = t;
		t = t_next;
		t_next = t->next;
	}
	while ((t != list->tail) && (t->key < search_key));
	
	return t;
}


/** 
 * Insert a new item to the synchronized linked list.
 *
 * @param list The list.
 * @param key The new key.
 * @param data The data.
 * @return True, if the new key was successfully added, false else.
 * @see sync_list_search
 */
bool sync_list_insert (sync_list_t* list, unsigned long key, void* data)
{
	mutex_lock (&sync_list_lock);
	
	sync_item_t* right_item = NULL;
	sync_item_t* left_item = NULL;
	
	right_item = sync_list_search (list, key, &left_item);
	if ((right_item != list->tail) && (right_item->key == key))
	{
		/* The key already exists. */
		mutex_unlock (&sync_list_lock);
		return false;
	}

	/* Create new structure.*/
	sync_item_t* new_item = malloc (sizeof (sync_item_t));
	sync_item_init (new_item, key, data);

	new_item->next = right_item;
	left_item->next = new_item;
	
	mutex_unlock (&sync_list_lock);
	return true;
}


/** 
 * Delete the item with the given key. 
 * Return data in the given @data_ptr - must be valid. 
 *
 * @param list The list.
 * @param search_key The key.
 * @param data_ptr Pointer to data.
 * @return True, if the item with given key was successfully removed, false else.
 * @see sync_list_search
 */
bool sync_list_remove (sync_list_t* list, unsigned long search_key, void** data_ptr)
{
	mutex_lock (&sync_list_lock);
	
	sync_item_t* left_item = NULL;
	sync_item_t* right_item = NULL;

	right_item = sync_list_search (list, search_key, &left_item);
	if ((right_item == list->tail) || (right_item->key != search_key))
	{
		/* There isn't any item with given key. */
		mutex_unlock (&sync_list_lock);
		return false;
	}

	left_item->next = right_item->next;	
	if (data_ptr)
	{
		(*data_ptr) = right_item->data;	
	}
	
	free (right_item);
	
	mutex_unlock (&sync_list_lock);
	return true;
}


/** 
 * Find the item with given key in the synchronized linked list.
 * Return in the @data_ptr - must be valid.
 *
 * @param list The list.
 * @param search_key The key.
 * @param data_ptr The pointer to data.
 *
 * @return True, if the item with given key was successfully found, false else.
 * @see sync_list_search
 */
bool sync_list_find (sync_list_t* list, unsigned long search_key, void** data_ptr)
{
	mutex_lock (&sync_list_lock);
	
	sync_item_t* left_item = NULL;
	sync_item_t* right_item = NULL;
	
	right_item = sync_list_search (list, search_key, &left_item);
	if ((right_item == list->tail) || (right_item->key != search_key))
	{
		/* There isn't any item with given key. */
		mutex_unlock (&sync_list_lock);
		return false;
	}
	else
	{
		if (data_ptr)
		{
			(*data_ptr) = right_item->data;	
		}
		mutex_unlock (&sync_list_lock);
		return true;
	}
}


/**
 * Simple debug tool for dumping the global synchronized linked list.
 * Prints the contents (pointer, key, data) of all items in the list to the console.
 *
 * @param list The global synchronized linked list.
 */
void sync_list_dump (sync_list_t* list)
{
	sync_item_t* it = list->head->next;
	while(it != list->tail)
	{
		printk_not_sync ("Item ptr %p, key %x, data %x.\n", it, it->key, it->data);
		it = it->next;
	}
}
