/**
 * @file keyboard.c
 *
 * keyboard driver and handling
 *
 */


#include <include/shared.h>
#include <drivers/keyboard.h>
#include <include/error_codes.h>
#include <adt/list.h>
#include <proc/thread.h>
#include <sync/mutex.h>
#include <sched/sched.h>
#include <drivers/dorder.h>
#include <drivers/printer.h>



/** keyboard mutex */
static struct mutex keyboard_mutex;
/** readers blocked by keyboard */
list_t waiting_readers;
/** keyboard buffer */
char keyboard_buffer [KEYBOARD_BUFFER_SIZE];
/** top buffer index */
uint32_t top_buffer_index;
/** bottom buffer index */
uint32_t bot_buffer_index;


/** blocked readers */
struct reader
{
	/* readers are in list */
	item_t     item;
	/* blocked thread */
	thread_t   reader_thread;
	/* how many chars it wants to read */
	uint32_t   requested_count;
	/* how many chars it already read */
	uint32_t   readed_count;
	/* buffer for that characters */
	char     * readed_chars;
};


/** initializes keyboard */
void keyboard_init (void)
{
	/* lock for ensuring atomicity */
	mutex_init (&keyboard_mutex);
	/* intializion of readers list */
	list_init (&waiting_readers);
	/* pointer (index) to the top of the buffer */
	top_buffer_index = 0;
	/* pointer (index) to the bottom of the buffer */
	bot_buffer_index = 0;
};


/** read char; if there is no char, block thread
 *
 * @return returns readed char
 *
 */
char getc (void)
{
	char c;
	gets (&c, 1);
	return c;
};


/** gets char from keyboard buffer and removes it
 *
 * @param out_char char from buffer
 *
 * @return returns EOK (if there is some char in buffer) or EWOULDBLOCK (if there is nothing to return)
 *
 */
static inline int read_buffer (char * out_char);
static inline int read_buffer (char * out_char)
{
	/* there is nothing in the buffer if top and bottom index are equal */
	if (top_buffer_index == bot_buffer_index)
	{
		return EWOULDBLOCK;
	}
	else
	{
		/* add input into the buffer */
		if (NULL != out_char)
		{
			*out_char = keyboard_buffer[bot_buffer_index];
		}
		bot_buffer_index++;
		bot_buffer_index %= KEYBOARD_BUFFER_SIZE;
		return EOK;
	}
};


/** try to read char; if there is no char, carry on
 *
 * @return returns readed char or EWOULDBLOCK
 *
 */
char getc_try (void)
{
	char c;
	int result;
	ipl_t state = query_and_disable_interrupts ();
	result = read_buffer (&c);
	conditionally_enable_interrupts (state);
	if (EWOULDBLOCK != result)
	{
		return c;
	}
	else
	{
		return result;
	}
};


/** transfers chars from buffer into waiting threads if there are threads and chars */
void process_chars (void);
void process_chars (void)
{
	struct reader * rdr = (struct reader *)(waiting_readers.head);
	while ((NULL        != rdr                                                 ) &&
	       (EWOULDBLOCK != read_buffer(&(rdr->readed_chars[rdr->readed_count]))))
	{
		if (('\n' == ((rdr->readed_chars)[rdr->readed_count])) ||
		    (rdr->readed_count + 1 == rdr->requested_count   ))
		{
			mutex_lock (&keyboard_mutex);
			list_remove (&waiting_readers, &(rdr->item));
			mutex_unlock (&keyboard_mutex);
			if (NULL == rdr->reader_thread->item.list)
			{
				runnable (rdr->reader_thread);
			}
		}
		rdr->readed_count++;
		rdr = (struct reader *)(waiting_readers.head);
	}
	
};


/** reads string terminated by \n or count
 *
 * @param str pointer to string destination
 * @param len maximum length of string if not terminated by \n earlier
 *
 * @return returns real string length
 *
 */
ssize_t gets (char * str, size_t len)
{
	if (0 == len)
	{
		return 0;
	}

	ipl_t state = query_and_disable_interrupts ();

	struct reader rdr;
	item_init (&(rdr.item));
	rdr.reader_thread   = thread_get_current ();
	rdr.requested_count = len;
	rdr.readed_count    = 0;
	rdr.readed_chars    = str;
	mutex_lock (&keyboard_mutex);
	list_append (&(waiting_readers), &(rdr.item));
	mutex_unlock (&keyboard_mutex);

	process_chars ();
	
	if ((!(rdr.readed_count == rdr.requested_count           )) &&
	    (!('\n' == ((rdr.readed_chars)[rdr.readed_count - 1]))))
	{
		disabled (rdr.reader_thread);
		scheduled ();
	}

	conditionally_enable_interrupts (state);

	return rdr.readed_count;
};


/** keyboard interrupt handling */
void keyboard_interrupt (void)
{
	keyboard_buffer[top_buffer_index] = (*((volatile uint8_t *)(KEYBOARD_ADDRESS)));
	top_buffer_index++;
	top_buffer_index %= KEYBOARD_BUFFER_SIZE;
	if (top_buffer_index == bot_buffer_index)
	{
		bot_buffer_index++;
		bot_buffer_index %= KEYBOARD_BUFFER_SIZE;
	}
	process_chars ();
};
