/**
 * @file printer.c
 *
 * Simple console support.
 *
 */


#include <drivers/printer.h>

#include <include/shared.h>
#include <include/c.h>
#include <sync/mutex.h>


/** Printer mutex */
static struct mutex printer_mutex;


/** init printers */
void printer_init (void)
{
	mutex_init (&printer_mutex);
}


/** Print one character, NOT SYNCHRONIZED (call in debug mode etc.).
 *
 * @param chr The character to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t putc_not_sync (const char chr)
{
	/* writting a char to printer address prints it */
	(*((volatile uint8_t *)(PRINTER_ADDRESS))) = chr;

	/* this function never fails, it always writes one char */
	return 1;
}


/** Print string, NOT SYNCHRONIZED (call in debug mode etc.).
 *
 * @param str The string to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t puts_not_sync (const char * str)
{
	/* printed chars */
	uint32_t result = 0;

	/* '\0' is not always 0 (depends on platform) */
	while ('\0' != (*str))
	{
		/* print char, add it to result and proceed */
		result += putc_not_sync (*(str++));
	}

	/* returns count of printed chars */
	return result;
}


/** Print unsigned integer, NOT SYNCHRONIZED (call in debug mode etc.).
 *
 * @param uint The unsigned integer to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t putu_not_sync (uint32_t uint);
size_t putu_not_sync (uint32_t uint)
{
	/* printed chars */
	size_t result = 0;

	/* 32bit unsigned integer has max 10 chars as string */
	char output[10];

	/* index of array */
	uint32_t loop1 = 0;

	/* if zero, number seems to be consumed */
	if (0 == uint)
	{
		/* print zero */
		putc_not_sync ('0');
		/* zero is one char width */
		return 1;
	}

	/* go until number is consumed */
	while (uint)
	{
		/* move last cipher to array */
		output[loop1++] = uint % 10;
		/* cut last cipher away */
		uint /= 10;
	}

	/* index means count of printed chars */
	result = loop1--;

	/* the number is reverted in array */
	do
	{
		/* print number as char */
		putc_not_sync ('0' + output[loop1]);
	}
	/* repeat until array printed */
	while (0 != loop1--);

	/* returns count of printed chars */
	return result;
}


/** Print signed integer, NOT SYNCHRONIZED (call in debug mode etc.).
 *
 * @param sint The signed integer to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t putd_not_sync (int32_t sint);
size_t putd_not_sync (int32_t sint)
{
	/* printed chars count */
	uint32_t result = 0;

	/* lt zero, print '-' and signed integer */
	if (0 > sint)
	{
		/* print '-' and count it */
		result += putc_not_sync ('-');
		/* revert negative integer to positive */
		sint = -1 * sint;
		/* print number and count it */
		result += putu_not_sync (sint);
	}
	else
	{
		/* gt zero, print it as unsigned */
		result += putu_not_sync (sint);
	}

	/* return count of printed chars */
	return result;
}


/** Print hexadecimal unsigned integer, NOT SYNCHRONIZED (call in debug mode etc.).
 *
 * @param hint The hexadecimal unsigned integer to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t puth_not_sync (uint32_t hint);
size_t puth_not_sync (uint32_t hint)
{
	/* printed chars */
	size_t result = 0;

	/* 32bit unsigned integer has max 8 chars as hexa string */
	char output[8];

	/* index of array */
	uint32_t loop1 = 0;

	/* if zero, number seems to be consumed */
	if (0 == hint)
	{
		/* print zero */
		putc_not_sync ('0');
		/* zero is one char width */
		return 1;
	}

	/* go until number is consumed */
	while (hint)
	{
		/* move last cipher to array */
		output[loop1++] = hint % 16;
		/* cut last cipher away */
		hint /= 16;
	}

	/* index means count of printed chars */
	result = loop1--;

	/* the number is reverted in array */
	do
	{
		/* if lt 0xa, print as digit */
		if (10 > output[loop1])
		{
			/* print number as char */
			putc_not_sync ('0' + output[loop1]);
		}
		else
		{
			/* print hexachar */
			putc_not_sync ('a' + output[loop1] - 10);
		}
	}
	/* repeat until array printed */
	while (0 != loop1--);

	/* returns count of printed chars */
	return result;
}


/** Function for printing formated string; it can be used by synchronized or unsynchronized version of printk.
 *
 * @param fst The formated string to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t printk_variadic_list (const char * fst, __builtin_va_list variadic_args)
{
	/* printed chars */
	uint32_t result = 0;

        /* '\0' is not always 0 (depends on platform) */
	while ('\0' != (*fst))
	{
		/* if there is %, special char will follow */
		if ('%' == *fst)
		{
			/* skip that % */
			fst++;
			/* look at special char type */
			switch (*fst)
			{
				/* print char */
				case 'c':
					/* print char from arg and count it */
					result += putc_not_sync (__builtin_va_arg(variadic_args, int));
					/* of case */
					break;
				/* print string */
				case 's':
					/* print string from arg and count it */
					result += puts_not_sync (__builtin_va_arg(variadic_args, char *));
					/* of case */
					break;
				/* print signed integer */
				case 'd':
					/* print signed integer from arg and count it */
					result += putd_not_sync (__builtin_va_arg(variadic_args, int32_t));
					/* of case */
					break;
				/* print unsigned integer */
				case 'u':
					/* print unsigned integer from arg and count it */
					result += putu_not_sync (__builtin_va_arg(variadic_args, uint32_t));
					/* of case */
					break;
				/* print signed integer */
				case 'i':
					/* print signed integer from arg and count it */
					result += putd_not_sync (__builtin_va_arg(variadic_args, int32_t));
					/* of case */
					break;
				/* print unsigned integer as hexa */
				case 'x':
					/* print unsigned integer as hexa from arg and count it */
					result += puth_not_sync (__builtin_va_arg(variadic_args, uint32_t));
					/* of case */
					break;
				/* print pointer */
				case 'p':
					/* print pointer from arg and count it */
					result += puts_not_sync ("0x") + puth_not_sync (__builtin_va_arg(variadic_args, uint32_t));
					/* of case */
					break;
				/* print real % */
				case '%':
					/* print char and count it */
					result += putc_not_sync ('%');
					/* of case */
					break;
				/* ignore other cases */
				default:
					/* of case */
					break;
			}
			/* skip special char */
			fst++;
		}
		/* standard char */
		else
		{
			/* print char and count it */
			result += putc_not_sync (*(fst++));
		}
	}

	/* returns count of printed chars */
	return result;
}


/** Print formated string, NOT SYNCHRONIZED (call in debug mode etc.).
 *
 * @param fst The formated string to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t printk_not_sync (const char * fst, ...)
{
	/* count of printed chars */
	size_t result;

	/* take args to pass them into internal function */
	__builtin_va_list args;
	/* initialize args list */
	__builtin_va_start (args, fst);

	/* call internal function with args for this function */
	result = printk_variadic_list (fst, args);

	/* clean up variadic args */
	__builtin_va_end (args);

	/* return count of printed chars */
	return result;
}


/** Print one character.
 *
 * @param chr The character to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t putc (const char chr)
{
	size_t result;
	mutex_lock (&printer_mutex);
	result = putc_not_sync (chr);
	mutex_unlock (&printer_mutex);
	return result;
}


/** Print string.
 *
 * @param str The string to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t puts (const char * str)
{
	size_t result;
	mutex_lock (&printer_mutex);
	result = puts_not_sync (str);
	mutex_unlock (&printer_mutex);
	return result;
}


/** Print formated string.
 *
 * @param fst The formated string to print.
 *
 * @return Returns count of printed characters.
 *
 */
size_t printk (const char * fst, ...)
{
	/* count of printed chars */
	size_t result;

	/* take args to pass them into internal function */
	__builtin_va_list args;
	/* initialize args list */
	__builtin_va_start (args, fst);

	/* synchronize with other printing */
	mutex_lock (&printer_mutex);
	/* call internal function with args for this function */
	result = printk_variadic_list (fst, args);
	/* unsynchronize with other printing */
	mutex_unlock (&printer_mutex);

	/* clean up variadic args */
	__builtin_va_end (args);

	/* returns count of printed chars */
	return result;
}


/** Print a zero terminated string
 *
 * @param text The string to print.
 *
 */
void print (const char *text)
{
	puts (text);
} 