/**
 * @file timer.c
 *
 * Timer queue and structures implementation.
 *
 */

#include <drivers/timer.h>

#include <drivers/dorder.h>
#include <drivers/printer.h>
#include <exc/assert.h>
#include <include/c.h>
#include <sched/sched.h>


/**
 * Statically set ratio to have deterministic system and easier debugging.
 * To be a more realistic system this value can be set by ticks_per_second function.
 */
uint32_t ticks_per_s;

/**
 * This value determines how many seconds can be converted at once to ticks
 * by @link ticks_per_s @endlink while not overflowing native unsigned integer. 
 */
uint32_t s_per_fullticks;

/** 
 * Head of the timer sorted queue used to pick the timer
 * with the smallest remaining quantum to run its handler.
 */
timer* timer_normal [MAX_CPU];


/**
 * Initialize timer constants @link ticks_per_s @endlink and 
 * @link s_per_full_ticks @endlink.
 */
void init_timer (void)
{
	ticks_per_s = 1;
	s_per_fullticks = 0xffffffff / ticks_per_s;

	printk_not_sync ("Ticks per second %u, ticks for full cycle %u.\n", ticks_per_s, s_per_fullticks);
}


/**
 * Initialize timer queue for current CPU, timers can be started.
 */
void init_timer_cpu (void)
{
	timer_normal [cpuid ()] = NULL;
}


/**
 * Initialize passed timer structure with supplied values.
 * Prepares timer to be started.
 * 
 * @param tmr
 *	Timer to initialize.
 *
 * @param usec 
 *	Seconds to wait before firing handler function.
 *
 * @param handler
 *	Function to run when time elapse.
 * 
 * @param data
 *	Data to pass into handler function.
 *
 * @return
 * 	EOK - initialization finished successfully.
 *	EINVAL - initialization failed.
 *
 */
int32_t timer_init (
	struct timer* tmr,
	const uint32_t usec,
	void (*handler) (struct timer*, void*), 
	void* data)
{
	if (tmr->handler == handler && tmr->data == data)
	{
		return EINVAL;
	}

	tmr->handler = handler;	
	tmr->data = data;
	tmr->usec = usec;
	tmr->remain_usec = 0;
	tmr->remain_ticks = 0;
	tmr->enabled = false;
	tmr->free_me = false;
	tmr->prev = NULL;
	tmr->next = NULL;
	
	return EOK;
}


/**
 * Function starts and enqueues timer.
 *
 * @param tmr
 * 	Timer to start.
 *
 * @param queue
 *	Queue to add timer to.
 */
void timer_start_queue (struct timer* tmr, struct timer** queue);
void timer_start_queue (struct timer* tmr, struct timer** queue)
{
	/* Prepares structure for time countdown. */
	tmr->remain_usec = tmr->usec;
	if (tmr->remain_usec >= s_per_fullticks)
	{
		tmr->remain_ticks = s_per_fullticks * ticks_per_s;
		tmr->remain_usec -= s_per_fullticks;
	}
	else
	{
		tmr->remain_ticks = tmr->remain_usec * ticks_per_s;
		tmr->remain_usec = 0;
	}
	tmr->enabled = true;

	/* Add timer in the front if queue is empty. */
	if (*queue == NULL)
	{
		*queue = tmr;
	} 
	/* Add timer in the front if fires faster than current front. */
	else if ( 	(*queue)->remain_usec > tmr->remain_usec || 
			((*queue)->remain_usec == tmr->remain_usec && (*queue)->remain_ticks > tmr->remain_ticks))
	{
		timer* tmp = *queue;
		*queue = tmr;
		tmr->next = tmp;		
	}
	/* Lets find the correct spot by the remaining time. */
	else
	{
		timer* tmp = *queue;

		while (true)
		{
			/* If timer fits at the end of the queue. */
			if (tmp->next == NULL)
			{
				tmp->next = tmr;
				tmr->prev = tmp;
				
				break;
			}

			/* If we have just found the correct spot. */
			if ( 	tmp->next->remain_usec > tmr->remain_usec || 
				(tmp->next->remain_usec == tmr->remain_usec && tmp->next->remain_ticks > tmr->remain_ticks))
			{
				timer* next = tmp->next;
				
				next->prev = tmr;
				tmr->next = next;
				tmr->prev = tmp;
				tmp->next = tmr;
				
				break;
			}

			tmp = tmp->next;
		}
	}
}


/**
 * Function starts timer on current CPU. 
 *
 * @param tmr
 * 	Timer to start.
 */
void timer_start (struct timer* tmr)
{
	ipl_t state = query_and_disable_interrupts();
	
	timer_start_queue (tmr, &timer_normal [cpuid ()]);
	
	conditionally_enable_interrupts (state);
}


/**
 * Function clears timer structure and removes it from queue
 * if timer is still enabled.
 *
 * @param tmr
 * 	Timer to destroy.
 *
 * @param queue
 *	Queue to remove from if necessary.
 */
void timer_destroy_queue (struct timer* tmr, struct timer** queue);
void timer_destroy_queue (struct timer* tmr, struct timer** queue)
{	
	timer* prev = tmr->prev;
	timer* next = tmr->next;
	
	if (tmr->enabled == true)
	{
		if (prev == NULL)
		{
			*queue = tmr->next;
			
		}
		else if (next == NULL)
		{
			tmr->prev->next = NULL;	
		}
		else
		{
			prev->next = next;
			next->prev = prev;
		}
	}

	tmr->handler = NULL;
	tmr->data = NULL;	
	tmr->remain_usec = 0;
	tmr->remain_ticks = 0;
	tmr->enabled = false;
	tmr->prev = NULL;
	tmr->next = NULL;
}


/**
 * Function clears timer structure and stops timer if necessary.
 *
 * @param tmr
 * 	Timer to destroy.
 *
 * @param queue
 *	Queue to remove from if necessary.
 */
void timer_destroy (struct timer* tmr)
{
	ipl_t state = query_and_disable_interrupts ();

	timer_destroy_queue (tmr, &timer_normal [cpuid ()]);
	
	conditionally_enable_interrupts (state);
}


/**
 * Determines if timer is ticking or not.  
 *
 * @param tmr
 * 	Timer to query.
 *
 * @return
 *	True - if timer is running.
 * 	False - if is not.
 */
int32_t timer_pending (struct timer *tmr)
{
	return tmr->enabled;
}


/**
 * Function lowers remaining time of all timers in the queue.
 *
 * @param ticks
 *	Ticks passed.
 * 
 * @param
 *	Queue to lower timers form. 
 */
void timer_ticks_queue (const uint32_t ticks, struct timer** queue);
void timer_ticks_queue (const uint32_t ticks, struct timer** queue)
{
	/* TODO rewrite - split into functions */ 

	timer* tmr = *queue;
	while (tmr != NULL)
	{
		uint32_t tmpticks = ticks;	
	
		/* We need to try decreasing remamining time
		 * in cycles, for we may need to convert remaming seconds
		 * to remaining ticks few times. */
		while (tmpticks > 0)
		{	
			/* Enough remaining ticks, no need to convert. */
			if (tmr->remain_ticks > tmpticks)
			{
				tmr->remain_ticks -= tmpticks;
				tmpticks = 0;
			}
			else if (tmr->remain_ticks <= tmpticks)
			{
				/* Timer is ready to run handler. */ 
				if (tmr->remain_usec == 0)
				{
					tmr->remain_ticks = 0;
					tmpticks = 0;
				}
				/* We need to convert remamining seconds to ticks. */
				else
				{
					/* Maximum we can decrease by. */
					tmpticks -= tmr->remain_ticks;
				
					/* If remains more seconds than we can convert at once. */	
					if (tmr->remain_usec >= s_per_fullticks)
					{
						tmr->remain_ticks = s_per_fullticks * ticks_per_s;
						tmr->remain_usec -= s_per_fullticks;
					}
					else
					{
						tmr->remain_ticks = tmr->remain_usec * ticks_per_s;
						tmr->remain_usec = 0;
					}
				}
			}
		}
		tmr = tmr->next;		
	}
}


/**
 * Adjust timers on current CPU by time elapsed.
 * 
 * @param ticks
 *	Ticks to remove from all timers on current CPU.
 */
void timer_ticks (uint32_t ticks)
{
	ipl_t state = query_and_disable_interrupts ();
	
	timer_ticks_queue (ticks, &timer_normal [cpuid ()]);
	
	conditionally_enable_interrupts (state);
}


/**
 * Determines if there's any timer ready in the queue. 
 *
 * @param queue
 *	Queue to look for some ready timer.
 *
 * @return
 *	True - if there's ready timer.
 *	False - if there's not.
 */
bool timer_is_ready_queue (struct timer** queue);
bool timer_is_ready_queue (struct timer** queue)
{
	if ((*queue) != NULL && (*queue)->remain_usec == 0 && (*queue)->remain_ticks == 0)
		return true;
	else
		return false;
}


/**
 * Determines if there's any timer ready on current CPU. 
 *
 * @return
 *	True - if there's ready timer.
 *	False - if there's not.
 */
bool timer_is_ready (void)
{
	ipl_t state = query_and_disable_interrupts ();
	
	bool ret = timer_is_ready_queue (&timer_normal [cpuid ()]);
	
	conditionally_enable_interrupts (state);

	return ret;
}


/**
 * Returns ready timer from the queue. 
 *
 * @param queue
 *	Queue to look for ready timer.
 *
 * @return
 *	Pointer to timer - if there's any timer ready.
 *	NULL - if there's none.
 */
struct timer* timer_ready_queue (struct timer** queue);
struct timer* timer_ready_queue (struct timer** queue)
{
	if ((*queue) == NULL || (*queue)->remain_usec != 0 || (*queue)->remain_ticks != 0)
	{
		return NULL;
	}
	else
	{
		if ( 	(*queue)->remain_usec == 0 &&
			(*queue)->remain_ticks == 0)
		{
			timer* tmr = (*queue);
			(*queue) = tmr->next;
		
			if ((*queue) != NULL)
			{
				(*queue)->prev = NULL;	
				if ((*queue)->next != NULL)
					(*queue)->next->prev = NULL;
			}			

			tmr->remain_usec = 0;
			tmr->remain_ticks = 0;
			tmr->enabled = false;
			tmr->prev = NULL;
			tmr->next = NULL;

			return tmr;
		}
		else
		{
			return NULL;
		}
	}
}


/**
 * Returns ready timer on the current CPU. 
 *
 * @param queue
 *	Queue to look for ready timer.
 *
 * @return
 *	Pointer to timer - if there's any timer ready.
 *	NULL - if there's none.
 */
struct timer* timer_ready(void)
{
	ipl_t state = query_and_disable_interrupts ();
	
	struct timer* tmr = timer_ready_queue (&timer_normal [cpuid ()]);

	conditionally_enable_interrupts (state);

	return tmr;
}


/**
 * Function for debug purpose.
 * Prints queue of timers on the current CPU.
 */
void timer_dump (void)
{
	ipl_t state = query_and_disable_interrupts ();
	
	timer* tmr = timer_normal [cpuid ()];
	while (tmr != NULL)
	{
		printk ("timer %x usec %d rusec %d rticks %d\n", tmr, tmr->usec, tmr->remain_usec, tmr->remain_ticks);
		tmr = tmr->next; 
	}
	
	conditionally_enable_interrupts (state);
}

