/**
 * @file example.c
 *
 * Simple example with multiple threads.
 *
 * Kalisto
 *
 * Copyright (c) 2001-2010
 *   Department of Distributed and Dependable Systems
 *   Faculty of Mathematics and Physics
 *   Charles University, Czech Republic
 *
 */

 
#include <include/shared.h>
#include <include/c.h>

#include <adt/nblock_list.h>
#include <mm/vmm.h>
#include <mm/malloc.h>
#include <mm/vmalloc.h>
#include <adt/list.h>
#include <proc/thread.h>
#include <sched/sched.h>
#include <drivers/printer.h>
#include <exc/assert.h>
#include <sync/mutex.h>
#include <drivers/dorder.h>
#include <drivers/dtime.h>

#ifdef KERNEL_TEST
	#include <api.h>
#endif

#include <example.h>


#ifndef KERNEL_TEST
static thread_t thread_user;
#endif

#if !defined(KERNEL_TEST) && !defined(USER_TEST)

/** Random seed for the example */
static unsigned int seed = 0;

/** Activity threads */
static thread_t thread_one;
static thread_t thread_two;

static struct mutex synchronizer;


/** Example thread activity
 *
 * The function simply prints a thread signature
 * and a few additional characters to show
 * thread activity.
 *
 * @param signature The thread signature string.
 * @param activity  The thread activity character.
 *
 */
static void activity (const char *signature, const char activity)
{
	for (unsigned int i = 0 ; i < 100 ; i++) {

		mutex_lock(&synchronizer);	

		/* Print the thread signature with interrupts disabled
		   to make sure it is not mixed with output of other
		   threads. */
		
		ipl_t state = query_and_disable_interrupts ();
		puts_not_sync (signature);
		conditionally_enable_interrupts (state);
		
		/* Print a random number of activity characters. */
		for (unsigned int j = 0 ; j < 10 + random (&seed) % 10; j++)
			putc_not_sync (activity);

		mutex_unlock(&synchronizer);

	}
}


/* Since we have no way of passing arguments to
   threads, we implement the individual example
   threads as separate procedures. */

/** Example thread one
 *
 */
static void * thread_one_function (void * data)
{
//	while( 1);
	activity (data, '+');
	return 0;
}

/** Example thread two
 *
 */
static void * thread_two_function (void * data)
{
//	while( 1);
	activity (data, '*');
	return 0;
}

#endif /* !KERNEL_TEST */


/** Main example function
 *
 * Use three threads to print gibberish on console
 * or run the kernel test.
 *
 */
void * example (void * data)
{
  	thread_sleep(1);

#ifdef KERNEL_TEST
	
	/* If kernel test is compiled in, run it. */
	run_test ();
	
#else /* KERNEL_TEST */

  	thread_create(&thread_user, NULL, NULL, TF_USERLAND | TF_USERINIT | TF_NEW_VMM);

#ifndef USER_TEST

	mutex_init(&synchronizer);

	thread_create(&thread_one, thread_one_function, " [U] ", 0);
	thread_create(&thread_two, thread_two_function, " [K] ", 0);

	/* Exhibit some activity of our own. */
	activity (" [M] ", '-');

#endif

#endif /* KERNEL_TEST */

#ifndef KERNEL_TEST
	thread_join(thread_user);
#endif

	thread_sleep(8);

	/* The machine is halted as soon as any thread finishes. */
	msim_halt ();
}
