/**
 * @file assert.h
 *
 * Simple assert and debug tools support.
 *
 * There are implemented some macros and debug tools, mainly the panic function, which is called when the system has run 
 * aground of a fatal error. The other tools are macro ASSERT and DPRINTK. The ASSERT macro is useful for checking some 
 * assumption (if the evaluation fails, the panic function is called). DPRINTK macro prints the message given as the 
 * argument to the console. It can be used as simple log trace debug tool. The implemented macros are used only if the 
 * standard C-assertions is disabled (!(NDEBUG)).
 */
 
#ifndef ASSERT_H_
#define ASSERT_H_

#include <include/c.h>
#include <drivers/printer.h>
#include <drivers/dorder.h>
#include <proc/thread.h>


/** 
 * The number of a wait cycles.
 * When the panic function is called, we need some time to stop the other CPUs. There is a block of code with zero 
 * effect, only to provide some extra time. This block of code is executed repeatedly (in cycles).
 */
#define WAIT_CYCLES 100


/**
 * The panic function - the system has run aground of a fatal error.
 *
 * If the panic originates in the kernel, we are kind of stuck. At first all the CPU are blocked and then the contents 
 * of the registers the current CPU are printed to the console. The current thread ID and CPU ID is printed too.
 *
 * We need to send the interprocessor message, that the panic function was called. For that reason we use the 
 * interprocessor interrupt (IPI) - every CPU that receives the IPI, stops the execution. Unfortunately, there can be a 
 * problem, if some of the other CPUs are in this time in the critical section and has disabled the interruption. For 
 * that reason we use the waiting cycles to get the extra time to leave all the critical sections. Its only assumption, 
 * that every CPU managed to leave the critical section, but in fact we cannot use some reliable confirmation in panic 
 * function.
 * @link kernel/drivers/dorder.h Interprocessor device implementation @endlink
 * 
 * In the end the simulator is stopped.
 *
 * @param msg Panic message.
 * @see cpu_send_ipi_all
 */
static inline __attribute__((noreturn)) void panic (const char *msg, ...)
{
	
        /* We need to block all the other CPUs - sending interprocessor interrupt to all the other CPUs. There is disabled 
         * interrupts in the current CPU, so this one isn't blocked. */
         
        query_and_disable_interrupts ();
        cpu_send_ipi_panic_all ();
        
        /* We need to wait some time until all the CPUs receive the interrupt. */
        for(int i=0; i<WAIT_CYCLES; ++i)
        {
        	printk_not_sync ("");
        }
        
        printk_not_sync ("\n[ [ [ KERNEL PANIC ] ] ]\n");

	/* Print the message @msg with variadic args to the console. */
        __builtin_va_list args;
	__builtin_va_start (args, msg);
	printk_variadic_list (msg, args);
	__builtin_va_end (args);

        /* Prints the identification of current thread and CPU. */
        printk_not_sync ("\n\nThread %p, CPU %u.\n", thread_get_current (), cpuid ());
                
        /* Prints contents of registers the current CPU to the console. */
        msim_reg_dump ();

        /* Stop simulation. */    
        msim_halt ();        
}

/* NDEBUG disables standard C-assertions */
#if !(NDEBUG)

/* GCC Compiler provides variable __FUNCTION__, which contains the name of current function. Unfortunatelly it is only GCC 
 * Compiler support, so we have to check if we work with GCC Compiler. There was a variable __func__ in older versions of GCC. */
#if __STDC_VERSION__ < 199901L
#if __GNUC__ >= 2
#define __func__ __FUNCTION__
#else /*__GNUC__*/
#define __func__ "<unknown>"
#endif /*__GNUC__*/
#endif /*__STDC_VERSION__*/


/**
 * Standard macro for assert function.
 * If evaluation fails, the panic function is called. The panic function prints the current file, fuction and line where the 
 * evaluation fails.
 *
 * @param expr The expression for evaluation. 
 */
#define ASSERT(expr) \
	(!(expr) ? \
        panic ("Assert error in file %s, function %s, line %u.\n", __FILE__, __func__, __LINE__) : \
	(void) 0); \


/**
 * Simple debug tool for log trace debugging.
 * DPRINTK macro prints name of the current file, the name of function, the line number and the formatted text entered as the 
 * argument of DPRINTK macro to the console. It's only a debug tool, we don't use synchronized print functions.
 *
 * @param msg The formatted text to print.
 */
#define DPRINTK(msg, ...) \
	printk_not_sync ("Dprintk message in file %s, function %s, line %i:\n", __FILE__, __func__, __LINE__); \
 	__builtin_va_list args; \
	__builtin_va_start (args, msg); \
	printk_variadic_list (msg, args); \
	__builtin_va_end (args); \
 
#else /* NDEBUG */

#define ASSERT(expr)            (void) 0
#define DPRINTK(args, ...)      (void) 0

#endif /* NDEBUG */

#endif /* ASSERT_H */
