/**
 * @file exc.c
 *
 * Exception handling.
 *
 * Kalisto
 *
 * Copyright (c) 2001-2010
 *   Department of Distributed and Dependable Systems
 *   Faculty of Mathematics and Physics
 *   Charles University, Czech Republic
 *
 */

#include <include/shared.h>
#include <include/c.h>
#include <mm/tlb.h>
#include <exc/int.h>
#include <exc/syscall.h>
#include <drivers/printer.h>
#include <exc/assert.h>

#include <exc/exc.h>


/** Handle an exception
 *
 * The function is called from the assembler exception handler.
 * Interrupts are disabled and the various System Control
 * Coprocessor registers identify what exception is
 * being handled.
 *
 * @param registers The registers at the time of the exception.
 *
*/
void wrapped_general (context_t *registers)
{
	thread_t curr_thr = thread_get_current();
		
	/* The handling of the exception depends on its cause */
	switch (CP0_CAUSE_EXCCODE (registers->cause)) {
	case CP0_CAUSE_EXCCODE_INT:
		/* Interrupt Exception:
		   Interrupts are handled by a separate procedure. */
		interrupt (registers);
		break;
	case CP0_CAUSE_EXCCODE_SYS:
		/* System Call Exception:
		   System calls are handled by a separate procedure. */
		syscall (registers);
		break;
	case CP0_CAUSE_EXCCODE_TLBL:
		wrapped_tlbrefill( registers);
		break;
	case CP0_CAUSE_EXCCODE_TLBS:			
		wrapped_tlbrefill( registers);
		break;
	case CP0_CAUSE_EXCCODE_ADEL:
	case CP0_CAUSE_EXCCODE_ADES:
		/* Address Error Exceptions:
		   Code accessed an invalid address.
		   
		   Something is very wrong, we basically just halt. */
		if (curr_thr->flags & TF_USERLAND)
		{
			printk_not_sync("Killing process - Address error exception.\n");
			registers->a0 = SYSCALL_EXIT;
			syscall(registers);
		}
		else
		{
			panic ("Unhandled Address Error exception.");
		}
		break;
	case CP0_CAUSE_EXCCODE_BP:
	case CP0_CAUSE_EXCCODE_TR:
	case CP0_CAUSE_EXCCODE_OV:
		/* Ignore debug exception if we are not in bd slot. */
		if ((read_cp0_cause () & CP0_CAUSE_BD_MASK) == 0)
		{
			registers->epc += 4;
		}
		/* We are in branch delay slot. */
		else
		{
			panic ("Debug exception in branch delay slot.");
		}
		break;
	
	/* Some other exceptions can occur but mostly they just
	   mean that something is terribly wrong and the kernel
	   halts. */
	
	case CP0_CAUSE_EXCCODE_CPU:
		if (curr_thr->flags & TF_USERLAND)
		{
			printk_not_sync("Killing process - Unhandled Coprocessor Unusable exception.\n");
			registers->a0 = SYSCALL_EXIT;
			syscall(registers);
		}
		else
		{
			panic ("Unhandled Coprocessor Unusable exception.");
		}
		break;
	case CP0_CAUSE_EXCCODE_RI:
		panic ("Unhandled Reserved Instruction exception.");
		break;
	case CP0_CAUSE_EXCCODE_IBE:
	case CP0_CAUSE_EXCCODE_DBE:
		panic ("Unhandled Physical Address Access exception.");
		break;
	default:
		printk_not_sync( "Unhandled exception cause %d", CP0_CAUSE_EXCCODE (registers->cause));
		panic ("Unhandled exception.");
	}
}
