/**
 * @file int.c
 *
 * Interrupt handling.
 *
 * Kalisto
 *
 * Copyright (c) 2001-2010
 *   Department of Distributed and Dependable Systems
 *   Faculty of Mathematics and Physics
 *   Charles University, Czech Republic
 *
 */


#include <exc/int.h>


#include <adt/list.h>
#include <drivers/dorder.h>
#include <drivers/keyboard.h>
#include <drivers/timer.h>
#include <exc/assert.h>
#include <include/shared.h>
#include <include/c.h>
#include <mm/tlb.h>
#include <proc/thread.h>
#include <sched/sched.h>


/** 
 * Cause of inter process interrupt.
 */
ATOMIC_DECLARE( ipi_panic, 0);

/** 
 * Cause of inter process interrupt.
 */
ATOMIC_DECLARE( ipi_cause, 0);

/** Handle an interrupt
 *
 * The function is called from the exception handler when the exception
 * is identified as being caused by an interrupt request. Interrupts
 * are disabled and the IP bits in the Cause register identify which
 * interrupt requests are active. More than one bit can be set.
 *
 * The handler has to do whatever is necessary to satisfy the interrupt
 * request, otherwise the exception will be generated immediately after
 * interrupts are enabled.
 *
 * @param registers The registers at the time of the interrupt.
 */
void interrupt (context_t *registers)
{
	/* Get the IP bits from the Cause register. */
	int cause = registers->cause;
	current_registers = registers;	
	
	if (cause & CP0_CAUSE_IP7_MASK) 
	{
		/* IP7 is a timer interrupt. */		
		schedule( cpuid());
	} 
	else if (cause & CP0_CAUSE_IP5_MASK) 
	{
		/* IP5 is a keyboard interrupt.
		
		   To disable iterrupt infinite loop just read from the keyboard address. 
		*/
		keyboard_interrupt();
		
	}
	else if (cause & CP0_CAUSE_IP6_MASK) 
	{
		/* IP6 is an interprocessor interrupt. */
		
		/* Interprocessor panic interrupt, just run forever. */
		if (atomic_get (&ipi_panic))
		{
			while (true);
		}
	
		switch (atomic_get (&ipi_cause))
		{

		case CP0_CAUSE_IPI_TLB:
			tlb_clear_local ();
			cpu_ack_ipi (cpuid ());	
			break;

		default: 
			panic ("Unknown interprocessor interrupt exception.");
			break;
		}
	}
	else 
	{
		panic ("Unhandled interrupt exception.");
	}

	/* All interrupt requests should be handled by now. */
}
