/**
 * @file syscall.c
 *
 * System call interface.
 *
 */

#include <drivers/keyboard.h>
#include <drivers/printer.h>
#include <exc/assert.h>
#include <include/c.h>
#include <include/error_codes.h>
#include <include/shared.h>
#include <mm/vmm.h>
#include <mm/vmalloc.h>
#include <proc/thread.h>
#include <sync/mutex.h>


#include <exc/syscall.h>


/** syscall endless loop for kernel stack of thread */
void syscall_function (void)
{
	uint32_t syscall_ID;
	uint32_t syscall_arg0;
	uint32_t syscall_arg1;
	uint32_t syscall_arg2;
	uint32_t syscall_result;

	thread_t this_thread = thread_get_current();

	while (1)
	{
		syscall_ID     = this_thread->registers->a0;
		syscall_arg0   = this_thread->registers->a1;
		syscall_arg1   = this_thread->registers->a2;
		syscall_arg2   = this_thread->registers->a3;
		syscall_result = 0;

		switch (syscall_ID)
		{
			case SYSCALL_THREAD_CREATE:
			{
				syscall_result = u_thread_create(
					(void**) syscall_arg0, 
					(void*) syscall_arg1, 
					(void*) syscall_arg2,
					TF_USERLAND );
				break;
			}
			case SYSCALL_THREAD_SELF:
			{
				thread_t curr_thread = thread_get_current(); 
				
				*((void**)syscall_arg0) = curr_thread;
				*((void**)syscall_arg1) = curr_thread->thread_function;
				*((void**)syscall_arg2) = curr_thread->data;
				
				break;
			}
			case SYSCALL_THREAD_JOIN:
			{
				int result = u_thread_valid((void*) syscall_arg0);
				if (result == EOK)
				{
					result = thread_join((thread_t) syscall_arg0);
					*((int*)syscall_arg1) = ((thread_t) syscall_arg0)->return_value;
				}
				
				/* User thread_join_timeout can't return EKILLED. */
				if(result == EKILLED)
				{
					result = EINVAL;
				}
				
				syscall_result = result;
				break;
			}
			case SYSCALL_THREAD_JOIN_TIMEOUT:
			{
				int result = u_thread_valid((void*) syscall_arg0);
				if (result == EOK)
				{
					result = thread_join_timeout((thread_t) syscall_arg0, (unsigned int) syscall_arg1);
					*((int*)syscall_arg2) = ((thread_t) syscall_arg0)->return_value;
				}
				
				/* User thread_join_timeout can't return EKILLED. */
				if(result == EKILLED)
				{
					result = EINVAL;
				}
				
				syscall_result = result;
				break;
			}
			case SYSCALL_THREAD_DETACH:
			{
				int result = u_thread_valid((void*) syscall_arg0);
				if (result == EOK)
				{
					result = thread_detach((thread_t) syscall_arg0);
				}
				syscall_result = result;
				break;
			}
			case SYSCALL_THREAD_CANCEL:
			{
				int result = u_thread_valid((void*) syscall_arg0);
				if (result == EOK)
				{
					result = thread_kill((thread_t) syscall_arg0);
				}
				syscall_result = result;
				break;
			}
			case SYSCALL_THREAD_SLEEP:
			{
				thread_sleep((unsigned int)syscall_arg0);  
				break;
			}
			case SYSCALL_THREAD_USLEEP:
			{
				thread_usleep((unsigned int)syscall_arg0);  
				break;
			}
			case SYSCALL_THREAD_YIELD:
			{
				thread_yield();
				break;
			}
			case SYSCALL_THREAD_SUSPEND:
			{
				thread_suspend();
				break;
			}
			case SYSCALL_THREAD_FINISH_OVER:
			{
				thread_t curr_thr = thread_get_current();
				curr_thr->return_value = (int)(syscall_arg0);
				thread_finish_over(curr_thr);
				break;
			}
			case SYSCALL_THREAD_EXIT:
			{
				thread_t curr_thr = thread_get_current();
				curr_thr->return_value = (int)(syscall_arg0);
				thread_finish_over(curr_thr);
				break;
			}
			case SYSCALL_EXIT:
			{
				u_process_kill(thread_get_current());
				break;
			}
			case SYSCALL_THREAD_WAKEUP:
			{
				int result = u_thread_valid((void*) syscall_arg0);
				if (result == EOK)
				{
					thread_wakeup((thread_t) syscall_arg0);
				}
				syscall_result = result;
				break;
			}
			case SYSCALL_MUTEX_INIT:
			{
				void** kernel_mutex = (void **)(syscall_arg0);
				syscall_result = u_mutex_init(kernel_mutex);
				break;
			}
			case SYSCALL_MUTEX_DESTROY:
			{
				void* kernel_mutex = (void*)(syscall_arg0);
				syscall_result = u_mutex_destroy(kernel_mutex);
				break;
			}
			case SYSCALL_MUTEX_LOCK:
			{
				void* kernel_mutex = (void*)(syscall_arg0);
				syscall_result = u_mutex_lock(kernel_mutex);
				break;
			}
			case SYSCALL_MUTEX_LOCK_TIMEOUT:
			{
				syscall_result = u_mutex_lock_timeout(
					(void*)syscall_arg0,
					(unsigned long)syscall_arg1);
				break;
			}
			case SYSCALL_MUTEX_UNLOCK:
			{
				void* kernel_mutex = (void*)(syscall_arg0);
				syscall_result = u_mutex_unlock(kernel_mutex);
				break;     
			}
			case SYSCALL_MALLOC:
			{
				printk_not_sync( "");
				uint32_t malloc_size = *(uint32_t*)(syscall_arg0);
				void* malloc_ptr = virtual_malloc( malloc_size);
				*((uint32_t *)(syscall_arg1)) = (uint32_t)malloc_ptr;

				break;
			}
			case SYSCALL_FREE:
			{
				printk_not_sync( "");
				void* free_ptr = (void*)(syscall_arg0);
				virtual_free( free_ptr);
				
				break;
			}
			case SYSCALL_PUTC:
			{
				printk_not_sync( "");
				char putc_chr = *(char*)(syscall_arg0);
				syscall_result = putc_not_sync( putc_chr);
				
				break;
			}
			case SYSCALL_PUTS:
			{
				printk_not_sync( "");
				char* puts_str = (char*)(syscall_arg0);
				syscall_result = puts_not_sync( puts_str);

				break;
			}
			case SYSCALL_GETC:
			{
				printk_not_sync( "");
				syscall_result = (uint32_t)getc();

				break;
			}
			default:
			{
				panic("\nUNKNOWN SYSCALL (ID %u, arg0 %u, arg1 %u, arg2 %u)\n", syscall_ID, syscall_arg0, syscall_arg1, syscall_arg2);
				break;
			}
		}


		/* interrupts must be disabled to avoid scheduler to plan out this thread
		 * before it switch contexts from kernel stack to userland */
		ipl_t status = query_and_disable_interrupts();
		this_thread->registers->a0 = syscall_result;
		this_thread->is_in_kernel_mode = 0;
		switch_cpu_context(&(this_thread->stack_top     ),
		                   &(this_thread->user_stack_top),
				   read_cp0_entryhi());
		conditionally_enable_interrupts(status);
	}
}


/** Handle a system call.
 *
 * The function is called from the exception handler when the exception
 * is identified as being caused by a system call. Interrupts are
 * disabled and the various registers identify what system
 * call is requested.
 *
 * Before returning, the routine adjusts the program counter to
 * point to the instruction following the system call.
 *
 * @param registers The registers at the time of the interrupt.
 *
 */
void syscall (context_t *registers)
{
	/* arguments are passed in registers stored in the userland context */
	
	thread_t this_thread = thread_get_current();

	this_thread->registers    = registers;

	/* interrupts must be disabled to avoid scheduler to plan out this thread
	 * before it switch contexts from kernel stack to userland */
	ipl_t status = query_and_disable_interrupts(); 
	this_thread->is_in_kernel_mode = 1;
	switch_cpu_context(&(this_thread->user_stack_top),
	                   &(this_thread->stack_top     ),
			   read_cp0_entryhi());
	conditionally_enable_interrupts(status);

	registers->epc += 4;
}
