/**
 * @file main.c
 *
 * Kernel mainline.
 *
 * Kalisto
 *
 * Copyright (c) 2001-2010
 *   Department of Distributed and Dependable Systems
 *   Faculty of Mathematics and Physics
 *   Charles University, Czech Republic
 *
 */

#include <include/shared.h>
#include <include/c.h>
#include <adt/list.h>
#include <mm/malloc.h>
#include <proc/thread.h>
#include <sched/sched.h>
#include <sync/mutex.h>
#include <mm/tlb.h>
#include <drivers/printer.h>
#include <drivers/keyboard.h>
#include <exc/assert.h>
#include <mm/memtest.h>
#include <drivers/dorder.h>
#include <drivers/dtime.h>
#include <example.h>
#include <mm/physmem.h>
#include <mm/vmm.h>
#include <mm/tlb.h>
#include <mm/vmalloc.h>


#include <main.h>


/** Idle thread
 *
 * This thread is being scheduled by the application
 * processors and does nothing (it only spins in
 * an infinite loop).
 *
 * Since Kalisto lacks the implementation of any
 * synchronization primitives, it is hard to do
 * any real job on the application processors.
 * Even the output to the dprinter device would
 * just create a mess.
 *
 */


/** The kernel entry point in C code for bootstrap processor (BSP)
 *
 * This function is called by the assembler code shortly after bootstrap,
 * with disabled interrupts and temporary stack. The function initializes
 * the kernel data structures and only then switches to a standard thread.
 *
 * After the initialization is complete, application processors entry
 * points are allowed to proceed.
 *
 */
void bsp_start (void)
{
	/* Say hello :-) We write a small message after each
	   initialization stage to make it easier to see
	   where things go wrong. Should that happen,
	   of course. */
	printk_not_sync("..:: INITIALIZING KALISTO ::..\n\n");

	printk_not_sync("DTime....\n");
	//ticks_per_second();
	printk_not_sync("DONE\n\n");

	printk_not_sync("Memory....\n");
	memory_test();
	printk_not_sync("DONE\n\n");

	printk_not_sync("Physical memory mapping....\n");
	physical_memory_mapping_init();
	printk_not_sync("DONE\n\n");
	
	/* Initialize TLB. We are running in an unmapped
	   segment, initializing TLB is therefore not
	   strictly necessary, but it might help
	   as an example for later assignments. */
	printk_not_sync ("Address translation....\n");
	printk_not_sync("DONE\n\n");
	
	/* Initialize heap. */
	printk_not_sync("Heap structure....\n");
	init_memory();
	init_virtual_memory();
	printk_not_sync("DONE\n\n");
	
	/* printer */
	printk_not_sync("Printer....\n");
	printer_init();
	printk_not_sync("DONE\n\n");

	/* keyboard */
	printk_not_sync("Keyboard....\n");
	keyboard_init();
	printk_not_sync("DONE\n\n");

	/* Threading. */
	printk_not_sync("Threading....\n");
	init_threads ();
	init_u_threads ();
	init_u_sync ();
	printk_not_sync("DONE\n\n");
	
	printk_not_sync("TLB....\n");
	init_tlb_cpu();
	printk_not_sync("DONE\n\n");

	printk_not_sync("Timer....\n");
	init_timer();
	init_timer_cpu();
	printk_not_sync("DONE\n\n");

	printk_not_sync("Scheduler....\n");
	init_scheduler();
	init_scheduler_cpu();
	printk_not_sync("DONE\n\n");

	printk_not_sync("..:: KERNEL IS WORKING ::..\n\n");
	
	/* Everything is ready for moving
	   to a standard thread. */
	thread_t thread;
	thread_create_on_cpu(&thread, example, NULL, TF_NEW_VMM, 0);

	/* Allow the first AP to run. It is critical to
	   do this after finishing all memory
	   allocations since there is no synchronization
	   (locking) of the memory management. */
	atomic_add(&cpu_ready, 1);

	start_scheduler_cpu();
	switch_to_thread (thread);
	
	/* The execution should never
	   reach this point. */
	panic ("Execution returned to bsp_start().");
}


/** The kernel entry point in C code for application processor (AP)
 *
 * This function is called by the assembler code shortly after bootstrap,
 * with disabled interrupts and temporary stack. The function waits until
 * the global kernel structures are initialized by bsp_start () and only
 * then switches to a standard thread.
 *
 */
void ap_start (void)
{
	/* Wait until we are ready to run. */
	while (atomic_get(&cpu_ready) < (native_t) cpuid());
	
	/* Initialize local CPU resources. */
	printk_not_sync("TLB on CPU %d ....\n", cpuid());
	init_tlb_cpu();
	printk_not_sync("DONE\n\n");

	printk_not_sync("Timer on CPU %d ....\n", cpuid());
	init_timer_cpu();
	printk_not_sync("DONE\n\n");

	printk_not_sync("Scheduler on CPU %d ....\n", cpuid());
	thread_t thread = init_scheduler_cpu();
	printk_not_sync("DONE\n\n");
	
	printk_not_sync("..:: CPU %d IS READY ::..\n\n", cpuid());
	
	/* Everything is ready for moving
	   to a standard idle thread scheduling. */
	
	/* Allow the next AP to run. */
	atomic_add(&cpu_ready, 1);
	
	start_scheduler_cpu();
	switch_to_thread (thread);
	
	/* The execution should never
	   reach this point. */
	panic ("Execution returned to ap_start().");
}
