/**
 * @file malloc.c
 *
 * Kernel memory allocator.
 *
 * The allocation policy is first fit, blocks are merged on free.
 *
 */

#include <include/shared.h>
#include <include/c.h>

#include <drivers/printer.h>
#include <exc/assert.h>
#include <mm/memtest.h>
#include <sync/mutex.h>
#include <mm/physmem.h>
#include <mm/vmm.h>
#include <include/error_codes.h>

#include <mm/malloc.h>

/** Some maximum heap block size to avoid overflow. */
#define HEAP_BLOCK_SIZE_MAX    0x10000000

/** Generic memory alignment */
#define ALIGNMENT  4


/** heap block structure
 *
 */
typedef struct
{
	/* true for free block, false for used block */
	bool   is_free;
	/* size of whole block including header structure */
	size_t block_size;
	/* pointer to previous block */
	void * previous_block;
	/* pointer to next block */
	void * next_block;
} heap_block_t;


struct mutex heap_mutex;


/** Address of first block */
static void * heap_start;


/** Initialize the heap allocator
 *
 * Finds how much physical memory we have and creates
 * the heap management structures that mark the whole
 * physical memory as a single free block.
 *
 */
void init_memory (void)
{
	heap_start = NULL;
	mutex_init(&heap_mutex);
}


/** Initialize a heap block
 *
 * Fills in the structures related to a heap block.
 *
 * @param addr Address of the block.
 * @param size Size of the block including the header and the footer.
 * @param free Indication of a free block.
 * @param prev pointer to previous block
 * @param next pointer to next block
 *
 */
static void block_init (void * addr, size_t size, bool free, void * prev, void * next);
static void block_init (void * addr, size_t size, bool free, void * prev, void * next)
{
	heap_block_t * heap_block = (heap_block_t *)(addr);

	heap_block->block_size     = size;
	heap_block->is_free        = free;
	heap_block->previous_block = prev;
	heap_block->next_block     = next;
}


/** finds the next block
 *
 * @param this_block actual block for finding the next
 *
 * @return returns the next block
 *
 */
heap_block_t * get_next_block (void * this_block);
heap_block_t * get_next_block (void * this_block)
{
	return (heap_block_t *)(((heap_block_t *)(this_block))->next_block);
}


/** finds previous block
 *
 * @param this_block actual block for finding the next
 *
 * @return returns previous block
 *
 */
heap_block_t * get_previous_block (void * this_block);
heap_block_t * get_previous_block (void * this_block)
{
	return (heap_block_t *)(((heap_block_t *)(this_block))->previous_block);
}


/** splits a free block to two blocks according to size of the first one which is set NOT free
 *
 * @param address address of the block to divide
 * @param real_size size of the first block (including header structure)
 *
 */
void block_split (void * address, size_t real_size);
void block_split (void * address, size_t real_size)
{
	heap_block_t * old_block = (heap_block_t *)(address);
	size_t         old_size  = old_block->block_size;

	assert(old_block->is_free);

	if (old_block->block_size < real_size + sizeof(heap_block_t))
	{
		old_block->is_free = 0;
	}
	else
	{
		void * prev = old_block->previous_block;
		void * next = old_block->next_block;
		void * scnd = address + real_size;
		block_init(address, real_size, 0, prev, scnd);
		block_init(scnd, old_size - real_size, 1, address, next);
		if (((heap_block_t *)(scnd))->next_block)
		{
			((heap_block_t *)(((heap_block_t *)(scnd))->next_block))->previous_block = scnd;
		}
	}
}


/** frees physical frame from some big free block to reduce kernel heap size; split that big block
 *
 * @param free_block free block to release something inside
 *
 */
void try_to_physical_free (heap_block_t * free_block);
void try_to_physical_free (heap_block_t * free_block)
{
	assert(free_block);
	assert(free_block->is_free);

	uint32_t block_start = (uint32_t)free_block;
	uint32_t frame_start = ALIGN_UP(block_start, PHYSMEM_PAGE_SIZE);
	if ((block_start != frame_start) && (sizeof(heap_block_t) < frame_start - block_start))
	{
		frame_start += PHYSMEM_PAGE_SIZE;
	}

	uint32_t block_end   = block_start + free_block->block_size;
	if (block_end < frame_start)
	{
		return;
	}
	uint32_t frame_count = (block_end - frame_start) / PHYSMEM_PAGE_SIZE;
	if (!(0 < frame_count))
	{
		return;
	}
	uint32_t frame_end   = ALIGN_DOWN(block_end, PHYSMEM_PAGE_SIZE);
	if ((block_end != frame_end) && (sizeof(heap_block_t) < block_end - frame_end))
	{
		frame_end -= PHYSMEM_PAGE_SIZE;
		frame_count--;
		if (!(0 < frame_count))
		{
			return;
		}
	}

	heap_block_t * prev = free_block->previous_block;
	heap_block_t * next = free_block->next_block;

	if ((block_start == frame_start) && (block_end == frame_end))
	{
		if (prev)
		{
			prev->next_block     = (void *)(next);
		}
		if (next)
		{
			next->previous_block = (void *)(prev);
		}
	}
	else
	{
		if ((block_start != frame_start) && (block_end != frame_end))
		{
			block_init((void *)block_start, frame_start - block_start, 1, (void *)prev, (void *)frame_end);
			block_init((void *)frame_end, block_end - frame_end, 1, (void *)block_start, (void *)next);
			if (prev)
			{
				prev->next_block     = (void *)block_start;
			}
			if (next)
			{
				next->previous_block = (void *)frame_end;
			}
		}
		else
		{
			if (block_start != frame_start)
			{
				block_init((void *)block_start, frame_start - block_start, 1, (void *)prev, (void *)next);
				if (prev)
				{
					prev->next_block     = (void *)block_start;
				}
				if (next)
				{
					next->previous_block = (void *)block_start;
				}
			}
			else
			{
				block_init((void *)frame_end, block_end - frame_end, 1, (void *)prev, (void *)next);
				if (prev)
				{
					prev->next_block     = (void *)frame_end;
				}
				if (next)
				{
					next->previous_block = (void *)frame_end;
				}
			}
		}
	}

	frame_free((void *)frame_start - 0x80000000, frame_count);
}


/** joins block with its free neighbours and returns that huge block
 *
 * @param this_block address of block for joining neighbours blocks
 *
 * @return returns a new huge free block
 *
 */
heap_block_t * join_blocks (heap_block_t * this_block);
heap_block_t * join_blocks (heap_block_t * this_block)
{
	assert (this_block);
	assert (this_block->is_free);

	heap_block_t * next_block = get_next_block((void *)this_block);
	if ((next_block) && (next_block->is_free) && ((void *)this_block + this_block->block_size == (void *)next_block))
	{
		block_init((void *)this_block, this_block->block_size + next_block->block_size, 1, this_block->previous_block, next_block->next_block);
		if (this_block->next_block)
		{
			((heap_block_t *)(this_block->next_block))->previous_block = (void *)this_block;
		}
	}

	heap_block_t * prev_block = get_previous_block((void *)this_block);
	if ((prev_block) && (prev_block->is_free) && ((void *)prev_block + prev_block->block_size == (void *)this_block))
	{
		block_init((void *)prev_block, prev_block->block_size + this_block->block_size, 1, prev_block->previous_block, this_block->next_block);
		if (prev_block->next_block)
		{
			((heap_block_t *)(prev_block->next_block))->previous_block = (void *)prev_block;
		}
		this_block = prev_block;
	}

	return this_block;
}


/** prints a kernel heap for debugging purposes
 *
 */
void kernel_heap_print (void);
void kernel_heap_print (void)
{
	printk_not_sync("KERNEL HEAP (starts at %p)\n", heap_start);
	heap_block_t * block = (heap_block_t *)(heap_start);
	while (block)
	{
		printk_not_sync("[address %p; size %u; free %u; prev %p; next %p]\n",
		                 block,
		                 block->block_size,
		                 block->is_free,
		                 block->previous_block,
		                 block->next_block);
		block = get_next_block((void *)block);
	}
}


/** Allocate a memory block
 *
 * @param size The size of the block to allocate.
 * @return The address of the block or NULL when not enough memory.
 *
 */
void * malloc (const size_t size)
{
	/* Checking for maximum size avoids errors due to
	   overflow, which would be hard to debug. */
	assert (size <= HEAP_BLOCK_SIZE_MAX);
	
	mutex_lock(&heap_mutex);
	
	/* We have to allocate a bit more to have room for header. */
	size_t real_size = ALIGN_UP (size, ALIGNMENT) + sizeof (heap_block_t);
	
	void * result = NULL;
	
	heap_block_t * pos        = (heap_block_t *) heap_start;
	heap_block_t * last_block = NULL;
	
	while ((!result) && (pos)) {
		/* Try to find a block that is free and large enough. */
		if ((pos->is_free) && (pos->block_size >= real_size)) {
			/* We have found a suitable block.
			   See if we should split it. */
			block_split((void *)pos, real_size);
			
			/* Either way we have our result. */
			result = ((void *) pos) + sizeof (heap_block_t);
		}
		/* Advance to the next block. */
		last_block = pos;
		pos        = get_next_block((void *)pos);
	}
	if (!result)
	{
		void * new_frame;
		size_t frames_count = (real_size - 1) / PHYSMEM_PAGE_SIZE + 1;
		if (EOK != frame_alloc(&new_frame, frames_count, VF_AT_KSEG0))
		{
			result = NULL;
		}
		else
		{
			new_frame += 0x80000000;
			if (last_block)
			{
				block_init(new_frame, frames_count * PHYSMEM_PAGE_SIZE, 1, (void *)last_block, last_block->next_block);
				last_block->next_block = new_frame;
			}
			else
			{
				heap_start = new_frame;
				block_init(new_frame, frames_count * PHYSMEM_PAGE_SIZE, 1, NULL, NULL);			
			}
			block_split(new_frame, real_size);
			result = new_frame + sizeof(heap_block_t);
		}
	}

	mutex_unlock(&heap_mutex);	

	return result;
}


/** Allocate a memory block
 *
 * Unlike standard malloc, this routine never returns NULL to
 * indicate an out of memory condition. Recovering from an out
 * of memory condition in the kernel is difficult and we never
 * really need it in the example kernel.
 *
 * @param size The size of the block to allocate.
 * @return The address of the block.
 *
 */
void *safe_malloc (const size_t size)
{
	void *result = malloc (size);
	assert (result != NULL);
	return (result);
}


/** Free a memory block
 *
 * @param addr The address of the block.
 */
void free (const void * addr)
{
	/* Calculate the position of the header. */
	heap_block_t * heap_block = (heap_block_t *) (addr - sizeof (heap_block_t));

	assert (!(heap_block->is_free));

	mutex_lock(&heap_mutex);

	/* Mark the block itself as free. */
	heap_block->is_free = true;

	heap_block_t * free_block = join_blocks(heap_block);
	try_to_physical_free(free_block);

	mutex_unlock(&heap_mutex);
}


