/**
 * @file memtest.c
 *
 * Detecting memory on system start by writting patterns to it and reading them.
 * Because there can be more memory modules in system, we do not assume one solid
 * memory block, there can be holes in the address space. Each solid block is 
 * count as one hardware memory module. We assume that there is no more than 
 * PHYSICAL_MEMORY_MODULES_LIMIT in the system. We left a free space in memory for
 * kernel binary and userland thread binary. This memory is treat as not available,
 * so it can not be allocated or freed (as like as address ranges where are no 
 * memory modules).
 *
 */


#include <mm/memtest.h>

#include <include/c.h>
#include <include/shared.h>
#include <drivers/printer.h>
#include <mm/physmem.h>


#define MEMORY_TESTING_BLOCK 		PHYSMEM_PAGE_SIZE
#define MEMORY_ATOMIC_BLOCK  		4
#define MEMORY_TESTING_BLOCKS_ALIGNMENT (1 << 20)

#define MEMTEST_VALUES(index) \
( \
	(0 == index ? 0x00000000 : \
	(1 == index ? 0xffffffff : \
	(2 == index ? 0xdeadbabe : \
	(3 == index ? 0xace5feed : \
	0x2decade5)))) \
)
#define MEMTEST_VALUES_COUNT 4


uint8_t  available_memory_map          [PHYSICAL_MEMORY_LIMIT_FOR_4K_FRAMES / 8] = {0};
uint32_t first_frames_of_memory_moduls [PHYSICAL_MEMORY_MODULES_LIMIT          ] = {0};
uint32_t frames_count_in_memory_moduls [PHYSICAL_MEMORY_MODULES_LIMIT          ] = {0};


/** set frame as valid
 *
 * @param frame_number a number of valid frame
 *
 */
void set_frame_valid (uint32_t frame_number);
void set_frame_valid (uint32_t frame_number)
{
	available_memory_map[frame_number / 8] |= 1 << (7 - frame_number % 8);
}


/** set frame as invalid
 *
 * @param frame_number a number of invalid frame
 *
 */
void set_frame_invalid (uint32_t frame_number);
void set_frame_invalid (uint32_t frame_number)
{
	available_memory_map[frame_number / 8] &= ~(1 << (7 - frame_number % 8));
}


/** tlb mapping virtual address to physical 1:1 
 *
 * @param address address to map
 *
 */
void map_address_to_physical (uintptr_t address);
void map_address_to_physical (uintptr_t address)
{
	/* The size of the page to map. */
	write_cp0_pagemask (CP0_PAGEMASK_4K);
	
	/* The physical address is simply the same as the virtual address. */
	uintptr_t frame_addr = address >> 12;
	uintptr_t frame_pair_addr = frame_addr & ~1;
	uintptr_t frame_pair_addr_lo =
	    frame_pair_addr & ((1 << CP0_ENTRYLO_PFN_BITS) - 1);
	uintptr_t frame_pair_addr_hi =
	    frame_pair_addr >> CP0_ENTRYLO_PFN_BITS;
	
	write_cp0_entrylo0 (((frame_pair_addr_lo + 0) << CP0_ENTRYLO_PFN_SHIFT)
	    | CP0_ENTRYLO_D_MASK | CP0_ENTRYLO_V_MASK);
	write_cp0_entrylo1 (((frame_pair_addr_lo + 1) << CP0_ENTRYLO_PFN_SHIFT)
	    | CP0_ENTRYLO_D_MASK | CP0_ENTRYLO_V_MASK);
	write_cp0_entryhi (frame_pair_addr_hi << CP0_ENTRYHI_VPN2_SHIFT);
	
	/* Fill a random TLB entry. */
	TLBWR ();
}


/** tlb mapping virtual null to physical frame
 *
 * @param frame_number the number of frame to map
 *
 */
void map_virtual_null_to_physical_frame_start (uintptr_t frame_number);
void map_virtual_null_to_physical_frame_start (uintptr_t frame_number)
{
	/* The size of the page to map. */
	write_cp0_pagemask (CP0_PAGEMASK_4K);
	
	/* The physical address is simply the same as the virtual address. */
	uintptr_t frame_addr = frame_number;
	uintptr_t frame_pair_addr = frame_addr & ~1;
	uintptr_t frame_pair_addr_lo =
	    frame_pair_addr & ((1 << CP0_ENTRYLO_PFN_BITS) - 1);
	uintptr_t frame_pair_addr_hi =
	    0;
	
	write_cp0_entrylo0 (((frame_pair_addr_lo + 0) << CP0_ENTRYLO_PFN_SHIFT)
	    | CP0_ENTRYLO_D_MASK | CP0_ENTRYLO_V_MASK);
	write_cp0_entrylo1 (((frame_pair_addr_lo + 1) << CP0_ENTRYLO_PFN_SHIFT)
	    | CP0_ENTRYLO_D_MASK | CP0_ENTRYLO_V_MASK);
	write_cp0_entryhi (frame_pair_addr_hi << CP0_ENTRYHI_VPN2_SHIFT);
	
	/* Fill a random TLB entry. */
	write_cp0_index(0);
	TLBWI ();
}


/** test four bytes of physical memory (aligned, ofc)
 *
 * @param ptr physical address to test
 *
 * @return returns true if memory byte is OK or false if it is not available or broken
 *
 */
bool test_four_bytes (uint32_t ptr);
bool test_four_bytes (uint32_t ptr)
{
	for (uint32_t loop1 = 0; loop1 < MEMTEST_VALUES_COUNT; loop1++)
	{
		(*(volatile uint32_t *)(ptr)) = MEMTEST_VALUES(loop1);
		if (MEMTEST_VALUES(loop1) != (*(volatile uint32_t *)(ptr)))
		{
			return 0;
		}
	}

	return 1;
}


/** tests a physical frame if it is valid (testing by first and last four bytes)
 *
 * @param frame_number number of frame to test
 *
 * @return returns true for valid frame or false for invalid frame
 *
 */
bool test_frame (uint32_t frame_number);
bool test_frame (uint32_t frame_number)
{
	map_virtual_null_to_physical_frame_start (frame_number);

	uintptr_t frame_start;
	uintptr_t frame_end;

	if (frame_number % 2)
	{
		frame_start = MEMORY_TESTING_BLOCK;
		frame_end   = 2 * MEMORY_TESTING_BLOCK - MEMORY_ATOMIC_BLOCK;
	}
	else
	{
		frame_start = 0;
		frame_end   = MEMORY_TESTING_BLOCK - MEMORY_ATOMIC_BLOCK;
	}

	return (test_four_bytes(frame_start)) && (test_four_bytes(frame_end));
}


/** gets a pointer and determine its memory block start and end
 *
 * @param ptr pointer to physical memory
 * @param start if not NULL, function fills it by block-start address
 * @param end if not NULL, function fills it by block-end address
 *
 */
void block_limits (uint32_t ptr, uint32_t * start, uint32_t * end);
void block_limits (uint32_t ptr, uint32_t * start, uint32_t * end)
{
	if (start)
	{
		(*start) = MEMORY_TESTING_BLOCK * (ptr / MEMORY_TESTING_BLOCK);
	}
	if (end)
	{
		(*end)   = MEMORY_TESTING_BLOCK * (ptr / MEMORY_TESTING_BLOCK) + MEMORY_TESTING_BLOCK - MEMORY_ATOMIC_BLOCK;
	}
}


/** test block of physical memory (aligns it itself)
 *
 * @param ptr some address of that block
 *
 * @return returns true for valid block or false for currupted or invalid block
 *
 */
bool test_block (uint32_t ptr);
bool test_block (uint32_t ptr)
{
	uint32_t block_start;
	uint32_t block_end;

	block_limits(ptr, &block_start, &block_end);

	return (test_four_bytes(block_start)) && (test_four_bytes(block_end));
}


/** get first address of nth block
 *
 * @param block_number counting number of block
 *
 * @return returns first address of nth block
 *
 */
uint32_t get_address_by_block_number (uint32_t block_number);
uint32_t get_address_by_block_number (uint32_t block_number)
{
	return MEMORY_TESTING_BLOCK * block_number;
}


/** get counting number of block
 *
 * @param address address somewhere in block
 *
 * @return returns block counting number
 *
 */
uint32_t get_block_number_by_address (uint32_t address);
uint32_t get_block_number_by_address (uint32_t address)
{
	return address / MEMORY_TESTING_BLOCK;
}


/** testing memory and detecting its size */
void memory_test (void)
{
	ipl_t status = read_cp0_status();
	write_cp0_status(status & ~CP0_STATUS_ERL_MASK);

	uint32_t bot_frame        = ((uint32_t)(&_kernel_end) - 0x80000000) / MEMORY_TESTING_BLOCK + 1;
	uint32_t user_bot_frame   = ((uint32_t)(USER_BIN_START)) / MEMORY_TESTING_BLOCK;
	uint32_t user_top_frame   = ((uint32_t)(USER_BIN_START + USER_BIN_LEN)) / MEMORY_TESTING_BLOCK + 1;
	uint32_t available_memory = (bot_frame + user_top_frame - user_bot_frame) * MEMORY_TESTING_BLOCK;
	uint32_t memory_modul     = 0;
	uint8_t  perc             = 0;
	bool     last_frame_OK    = 0;

	uint32_t loop1            = bot_frame;
	while (loop1 < PHYSICAL_MEMORY_LIMIT_FOR_4K_FRAMES)
	{
		if (test_frame(loop1))
		{
			set_frame_valid(loop1);
			available_memory += MEMORY_TESTING_BLOCK;
			if ((!last_frame_OK) && (memory_modul < PHYSICAL_MEMORY_MODULES_LIMIT))
			{
				first_frames_of_memory_moduls[memory_modul] = loop1;
			}
			frames_count_in_memory_moduls[memory_modul]++;
			last_frame_OK = 1;
		}
		else
		{
			if (last_frame_OK)
			{
				memory_modul++;
			}
			last_frame_OK = 0;
		}
		loop1++;
		if (user_bot_frame == loop1)
		{
			loop1 = user_top_frame;
			if (last_frame_OK)
			{
				memory_modul++;
			}
			last_frame_OK = 0;
		}
		if (100 * loop1 / PHYSICAL_MEMORY_LIMIT_FOR_4K_FRAMES != perc)
		{
			perc = 100 * loop1 / PHYSICAL_MEMORY_LIMIT_FOR_4K_FRAMES;
			printk_not_sync("\rMemory test (%u%% done)....", perc);			
		}
	}
	printk_not_sync("\rDetected %u KB; system will use %u frames (4KB size).\n", available_memory / 1024, available_memory / MEMORY_TESTING_BLOCK);

	available_memory = 0;
	for (loop1 = 0; loop1 < PHYSICAL_MEMORY_MODULES_LIMIT; loop1++)
	{
		available_memory += MEMORY_TESTING_BLOCK * frames_count_in_memory_moduls[loop1]; 
		if (0 < frames_count_in_memory_moduls[loop1])
		{
			printk_not_sync("Block of %u free frames found at address %p.\n", frames_count_in_memory_moduls[loop1], get_address_by_block_number(first_frames_of_memory_moduls[loop1]));
		}
	}
	printk_not_sync("\rSummary %u KB of free space.\n", available_memory / 1024);

	write_cp0_entrylo0(0);
	write_cp0_entrylo1(0);
	write_cp0_index(0);
	TLBWI();

	write_cp0_status(status);
}