/**
 * @file malloc.c
 *
 * Kernel memory allocator.
 *
 * The allocator expects the heap to reside in a continuous physical memory range.
 * Heap blocks have headers and footers and are kept next to each other so that
 * traversing can be done by adding or subtracting the appropriate size.
 *
 * The allocation policy is first fit, blocks are merged on free.
 *
 * Kalisto
 *
 * Copyright (c) 2001-2010
 *   Department of Distributed and Dependable Systems
 *   Faculty of Mathematics and Physics
 *   Charles University, Czech Republic
 *
 */

#include <include/shared.h>
#include <include/c.h>

#include <drivers/printer.h>
#include <exc/assert.h>
#include <mm/memtest.h>
#include <sync/mutex.h>

#include <mm/malloc.h>


/** Magic used in heap headers. */
#define HEAP_BLOCK_HEAD_MAGIC  0xBEEF0001

/** Magic used in heap footers. */
#define HEAP_BLOCK_FOOT_MAGIC  0xBEEF0002

/** Some maximum heap block size to avoid overflow. */
#define HEAP_BLOCK_SIZE_MAX    0x10000000


/** Generic memory alignment */
#define ALIGNMENT  4


/** Header of a heap block
 *
 */
typedef struct {
	/** Size of the block including header and footer */
	size_t size;
	
	/** Indication of a free block */
	bool free;
	
	/**
	 * A magic value to detect overwrite of heap header.
	 * The value is at the end of the header because
	 * that is where writing past block start will
	 * do damage.
	 */
	uint32_t magic;
} heap_block_head_t;

/** Footer of a heap block
 *
 */
typedef struct {
	/**
	 * A magic value to detect overwrite of heap footer.
	 * The value is at the beginning of the footer
	 * because that is where writing past block
	 * end will do damage.
	 */
	uint32_t magic;
	
	/** Size of the block */
	size_t size;
} heap_block_foot_t;


struct mutex heap_mutex;


/** Address of heap start */
static void *heap_start;

/** Address of heap end */
static void *heap_end;


#ifdef NDEBUG

/** Dummy heap block check
 *
 * Do not actually do any checking if we are building
 * a non-debug kernel.
 *
 */
#define block_check(addr)

#else /* NDEBUG */

/** Check a heap block
 *
 * Verifies that the structures related to a heap block still contain
 * the magic constants. This helps detect heap corruption early on.
 *
 * @param addr Address of the block.
 *
 */
static void block_check (void *addr)
{
	/* Calculate the position of the header. */
	heap_block_head_t *head = (heap_block_head_t *) addr;
	
	/* Make sure the header is still intact. */
	assert (head->magic == HEAP_BLOCK_HEAD_MAGIC);
	
	/* Calculate the position of the footer. */
	heap_block_foot_t *foot =
	    (heap_block_foot_t *) (addr + head->size - sizeof (heap_block_foot_t));
	
	/* Make sure the footer is still intact. */
	assert (foot->magic == HEAP_BLOCK_FOOT_MAGIC);
	
	/* And one extra check for the fun of it. */
	assert (head->size == foot->size);
}

#endif /* NDEBUG */


/** Initialize a heap block
 *
 * Fills in the structures related to a heap block.
 *
 * @param addr Address of the block.
 * @param size Size of the block including the header and the footer.
 * @param free Indication of a free block.
 *
 */
static void block_init (void *addr, size_t size, bool free)
{
	/* Calculate the position of the header and the footer. */
	heap_block_head_t *head = (heap_block_head_t *) addr;
	heap_block_foot_t *foot =
	    (heap_block_foot_t *) (addr + size - sizeof (heap_block_foot_t));
	
	/* Fill the header. */
	head->size = size;
	head->free = free;
	head->magic = HEAP_BLOCK_HEAD_MAGIC;
	
	/* Fill the footer. */
	foot->magic = HEAP_BLOCK_FOOT_MAGIC;
	foot->size = size;
}


/** Find an end of a physical memory block
 *
 * To find the end of a physical memory block, we simply try
 * modifying memory at addresses beyond the end of the kernel
 * image. When our modifications stop being accepted, we
 * consider that as indication of reaching the end of
 * the physical memory block.
 *
 * @param start The starting address of the memory block.
 *
 */
static void *find_block_end (void *start)
{
	volatile uint8_t *pos = (uint8_t *) start;
	const unsigned int step = 1024;
	
	while (true) {
		/* Try write two values, if either is
		   not read back, we have reached the end. */
		
		(*pos) = 0x55;
		if ((*pos) != 0x55)
			break;
		
		(*pos) = 0xAA;
		if ((*pos) != 0xAA)
			break;
		
		/* The increment is chosen for
		   this to be reasonably fast. */
		pos += step;
	}
	
	return ((void *) pos - step);
}


/** Initialize the heap allocator
 *
 * Finds how much physical memory we have and creates
 * the heap management structures that mark the whole
 * physical memory as a single free block.
 *
 */
void init_memory (void)
{
	/* We assume that our physical memory comes in a single
	   block that starts with the kernel image. We find
	   the end of the block by trial and error.
	   
	   Memory regions should be 4 bytes aligned to avoid
	   unaligned memory access exceptions while accessing
	   the header and footer structures. */
	heap_start = (void *) ALIGN_UP ((uintptr_t) &_kernel_end, ALIGNMENT);
//	heap_end = (void *) ALIGN_DOWN ((uintptr_t) find_block_end (heap_start), ALIGNMENT);
heap_end = find_block_end(heap_start);
// tohle je tu provizorne proto, aby nebyl heap pres celou pamet
// NEZ TOHLE NEKDO PREPISE, PROSIM, KONZULTUJTE SE MNOU (URZOU)
heap_end = (void *) ALIGN_DOWN ((uint32_t)heap_start + KERNEL_HEAP_SIZE, ALIGNMENT);
	
	/* Make the entire area one large block. */
	block_init (heap_start, heap_end - heap_start, true);
}


/** Allocate a memory block
 *
 * @param size The size of the block to allocate.
 * @return The address of the block or NULL when not enough memory.
 *
 */
void *malloc (const size_t size)
{
	/* Checking for maximum size avoids errors due to
	   overflow, which would be hard to debug. */
	assert (size <= HEAP_BLOCK_SIZE_MAX);
	
	/* Disable interrupts while accessing shared structures. */
	ipl_t state = query_and_disable_interrupts ();
	
	/* We have to allocate a bit more to have room for
	   header and footer. The size of the memory block
	   also has to be 4 bytes aligned to avoid unaligned
	   memory access exception while accessing the
	   footer structure. */
	size_t real_size = ALIGN_UP (size, ALIGNMENT) +
	    sizeof (heap_block_head_t) + sizeof (heap_block_foot_t);
	
	void *result = NULL;
	
	heap_block_head_t *pos = (heap_block_head_t *) heap_start;
	
	while ((result == NULL) && ((void *) pos < heap_end)) {
		/* Make sure the heap is not corrupted. */
		block_check (pos);
		
		/* Try to find a block that is free and large enough. */
		if ((pos->free) && (pos->size >= real_size)) {
			/* We have found a suitable block.
			   See if we should split it. */
			size_t split_limit = real_size +
			    sizeof (heap_block_head_t) + sizeof (heap_block_foot_t);
			
			if (pos->size > split_limit) {
				/* Block big enough -> split. */
				void *next = ((void *) pos) + real_size;
				block_init (next, pos->size - real_size, true);
				block_init (pos, real_size, false);
			} else {
				/* Block too small -> use as is. */
				pos->free = false;
			}
			
			/* Either way we have our result. */
			result = ((void *) pos) + sizeof (heap_block_head_t);
		}
		/* Advance to the next block. */
		pos = (heap_block_head_t *) (((void *) pos) + pos->size);
	}
	
	conditionally_enable_interrupts (state);
	
	return result;
}


/** Allocate a memory block
 *
 * Unlike standard malloc, this routine never returns NULL to
 * indicate an out of memory condition. Recovering from an out
 * of memory condition in the kernel is difficult and we never
 * really need it in the example kernel.
 *
 * Warning: Memory management implementation is not SMP safe.
 *
 * @param size The size of the block to allocate.
 * @return The address of the block.
 *
 */
void *safe_malloc (const size_t size)
{
	void *result = malloc (size);
	assert (result != NULL);
	return (result);
}


/** Free a memory block
 *
 * @param addr The address of the block.
 */
void free (const void *addr)
{
	/* Calculate the position of the header. */
	heap_block_head_t *head =
	    (heap_block_head_t *) (addr - sizeof (heap_block_head_t));
	
	/* Make sure the block is not corrupted and not free. */
	block_check (head);
	assert (!head->free);
	
	/* Disable interrupts while accessing shared structures. */
	ipl_t state = query_and_disable_interrupts ();
	
	/* Mark the block itself as free. */
	head->free = true;
	
	/* Look at the next block. If it is free, merge the two. */
	heap_block_head_t *next_head =
	    (heap_block_head_t *) (((void *) head) + head->size);
	
	if ((void *) next_head < heap_end) {
		block_check (next_head);
		if (next_head->free)
			block_init (head, head->size + next_head->size, true);
	}
	
	/* Look at the previous block. If it is free, merge the two. */
	if ((void *) head > heap_start) {
		heap_block_foot_t *prev_foot =
		    (heap_block_foot_t *) (((void *) head) - sizeof (heap_block_foot_t));
		
		heap_block_head_t *prev_head =
		    (heap_block_head_t *) (((void *) head) - prev_foot->size);
		
		block_check (prev_head);
		
		if (prev_head->free)
			block_init (prev_head, prev_head->size + head->size, true);
	}
	
	conditionally_enable_interrupts (state);
}
