/**
 * @file physmem.h
 *
 * Physical memory management.
 *
 * Pyhsical memory is managed by a bitmap. There is allowed to have more than one
 * physical memory module (more physical devices). Memory map as well as modules
 * starts and ends addresses are initialized in memtest.h. Here is just allocating
 * and freeing frames, all of it is very easy to understand. Free blocks are found
 * by first-fit strategy.
 *
 */


#include <mm/physmem.h>

#include <sync/mutex.h>
#include <include/c.h>
#include <include/shared.h>
#include <include/error_codes.h>
#include <drivers/printer.h>


struct mutex physmem_lock;


/** initializes physical memory mapping */
void physical_memory_mapping_init (void)
{
	mutex_init (&physmem_lock);
}


/** set physical frame unconditionally free
 *
 * @param frame_number number of frame to set free
 *
 */
void set_physical_frame_free (uint32_t frame_number);
void set_physical_frame_free (uint32_t frame_number)
{
	available_memory_map[frame_number / 8] |= 1 << (7 - frame_number % 8);
}


/** set physical frame unconditioanlly NOT free
 *
 * @param frame_number number of frame to set NOT free
 *
 */
void set_physical_frame_not_free (uint32_t frame_number);
void set_physical_frame_not_free (uint32_t frame_number)
{
	available_memory_map[frame_number / 8] &= ~(1 << (7 - frame_number % 8));
}


/** find out if given frame is free 
 *
 * @param frame_number number of frame, NOT ADDRESS, just number
 *
 * @return returns true if the frame is free or false otherwise
 *
 */
bool is_physical_frame_free (uint32_t frame_number);
bool is_physical_frame_free (uint32_t frame_number)
{
	return available_memory_map[frame_number / 8] & (1 << (7 - frame_number % 8));
}


/** find out if given frame is in some memory module
 *
 * @param frame_number number of frame, NOT ADDRESS, just number
 *
 * @return returns true if the frame is available or false otherwise
 *
 */
bool is_physical_frame_in_memory_module (uint32_t frame_number);
bool is_physical_frame_in_memory_module (uint32_t frame_number)
{
	for (uint32_t loop1 = 0; loop1 < PHYSICAL_MEMORY_MODULES_LIMIT; loop1++)
	{
		if ((first_frames_of_memory_moduls[loop1] <= frame_number                                      ) &&
		    (frame_number < first_frames_of_memory_moduls[loop1] + frames_count_in_memory_moduls[loop1]))
		{
			return 1;
		}
	}
	return 0;
}


/** find out if given frame is allocated
 *
 * @param frame_number number of frame, NOT ADDRESS, just number
 *
 * @return returns true if the frame is allocated or false otherwise
 *
 */
bool is_physical_frame_allocated (uint32_t frame_number);
bool is_physical_frame_allocated (uint32_t frame_number)
{
	return ((is_physical_frame_in_memory_module(frame_number)) && (!is_physical_frame_free(frame_number)));
}


/** find out if given frame could be allocated
 *
 * @param frame_number number of frame, NOT ADDRESS, just number
 *
 * @return returns true if the frame could be allocated or false otherwise
 *
 */
bool can_be_physical_frame_allocated (uint32_t frame_number);
bool can_be_physical_frame_allocated (uint32_t frame_number)
{
	return ((is_physical_frame_in_memory_module(frame_number)) && (is_physical_frame_free(frame_number)));
}


/** find out if given frame could be deallocated
 *
 * @param frame_number number of frame, NOT ADDRESS, just number
 *
 * @return returns true if the frame could be deallocated or false otherwise
 *
 */
bool can_be_physical_frame_deallocated (uint32_t frame_number);
bool can_be_physical_frame_deallocated (uint32_t frame_number)
{
	return is_physical_frame_allocated (frame_number);
}


/** allocate free frames between given addresses
 *
 * @param lower_address address for allocation start (or at the nearest higher frame after)
 * @param higher_address address higher limit for the allocation (or the nearest lower frame before)
 * @param count count of frames in a row
 * @param address starting address of allocation
 *
 * @return returns EOK if allocated or ENOMEM otherwise
 *
 */
int allocate_physical_frames_block (uint32_t lower_address, uint32_t higher_address, uint32_t count, void ** address);
int allocate_physical_frames_block (uint32_t lower_address, uint32_t higher_address, uint32_t count, void ** address)
{
	/* Here is just first fit finding free blocks.... nothing complicated here, there
	 * are just limits by physical memory modules, so we do not seek in spaces where
         * is no memory device. */
	
	uint32_t aligned_lower_address  = ALIGN_UP (lower_address, PHYSMEM_PAGE_SIZE);
	uint32_t aligned_higher_address = ALIGN_DOWN (higher_address, PHYSMEM_PAGE_SIZE);
	uint32_t lower_frame            = aligned_lower_address / PHYSMEM_PAGE_SIZE;
	uint32_t higher_frame           = aligned_higher_address / PHYSMEM_PAGE_SIZE;
	uint32_t lower_module_frame     = 0;
	uint32_t higher_module_frame    = 0;
	uint32_t count_temp             = 0;

	if ((!(0 < count)) || (higher_address < lower_address))
	{
		return ENOMEM;
	}

	mutex_lock (&physmem_lock);
	for (uint32_t loop1 = 0; loop1 < PHYSICAL_MEMORY_MODULES_LIMIT; loop1++)
	{
		if (lower_frame < first_frames_of_memory_moduls[loop1])
		{
			lower_module_frame = first_frames_of_memory_moduls[loop1];
		}
		else
		{
			lower_module_frame = lower_frame;
		}
		if (higher_frame < first_frames_of_memory_moduls[loop1] + frames_count_in_memory_moduls[loop1] - 1)
		{
			higher_module_frame = first_frames_of_memory_moduls[loop1] + frames_count_in_memory_moduls[loop1] - 1;
		}
		else
		{
			higher_module_frame = higher_frame;
		}
		if (lower_module_frame + count <= higher_module_frame + 1)
		{
			count_temp = 0;
			for (uint32_t loop2 = lower_module_frame; loop2 <= higher_module_frame; loop2++)
			{
				if (is_physical_frame_free(loop2))
				{
					count_temp++;
				}
				else
				{
					count_temp = 0;
				}
				if (count_temp == count)
				{
					for (uint32_t loop3 = count_temp; 0 < loop3; loop3--)
					{
						set_physical_frame_not_free (loop2 - loop3 + 1);
					}
					if (address)
					{
						*address = (void *)((loop2 - count_temp + 1) * (PHYSMEM_PAGE_SIZE));
					}
					mutex_unlock (&physmem_lock);
					return EOK;
				}
			}
		}
	}

	mutex_unlock (&physmem_lock);
	return ENOMEM;
}


/** allocates physical memory in one block
 *
 * @param paddr pointer to allocated memory
 * @param cnt count of allocated frames
 * @param flags specification for allocation details
 *
 * @return returns ENOMEM when fails or EOK otherwise
 *
 */
int frame_alloc (void ** paddr, const size_t cnt, const vm_flags_t flags);
int frame_alloc (void ** paddr, const size_t cnt, const vm_flags_t flags)
{	
	if (!(0 < cnt))
	{
		return ENOMEM;
	}

	
	if (VF_VA_AUTO & flags)
	{
		int res = allocate_physical_frames_block (0x00000000, 0xffffffff, cnt, paddr); 
		return res;
	}

	if ((VF_VA_USER & flags) && (paddr))
	{
		if (ENOMEM == allocate_physical_frames_block((uint32_t)(*paddr), 0xffffffff, cnt, paddr))
		{
			int res = allocate_physical_frames_block (0x00000000, (uint32_t)(*paddr), cnt, paddr); 
			return res;
		}
		else
		{
			return EOK;
		}
	}

	if ((VF_AT_KSEG0 | VF_AT_KSEG1) & flags)
	{
		int res = allocate_physical_frames_block(0x00000000, 0x1fffffff, cnt, paddr);
		return res;
	}

	if ((VF_AT_KUSEG | VF_AT_KSSEG | VF_AT_KSEG3) & flags)
	{
		if (ENOMEM == allocate_physical_frames_block(0x20000000, 0xffffffff, cnt, paddr))
		{
			int res = allocate_physical_frames_block (0x00000000, 0x20000000, cnt, paddr);
			return res;
		}
		else
		{
			return EOK;
		}
	}

	return ENOMEM;
}


/** frees physical memory allocated by frame_alloc
 *
 * @param paddr starting address for free
 * @param cnt count of frames for free
 *
 */
int frame_free (const void * paddr, const size_t cnt)
{
	uint32_t address      = (uint32_t)(paddr);
	uint32_t frame_number = address / PHYSMEM_PAGE_SIZE;
	uint32_t end_frame    = frame_number + cnt;
	if (PHYSICAL_MEMORY_LIMIT_FOR_4K_FRAMES < end_frame)
	{
		return EINVAL;
	}

	if (!(0 < cnt))
	{
		return EINVAL;
	}

	if (frame_number * PHYSMEM_PAGE_SIZE != address)
	{
		return EINVAL;
	}

	mutex_lock (&physmem_lock);
	for (uint32_t loop1 = frame_number; loop1 < end_frame; loop1++)
	{
		if (!is_physical_frame_allocated(loop1))
		{
			return EINVAL;
		}
	}

	for (uint32_t loop1 = frame_number; loop1 < end_frame; loop1++)
	{
		set_physical_frame_free (loop1);
	}
	mutex_unlock (&physmem_lock);

	return EOK;
}