/**
 * @file tlb.c
 *
 * Address translation handling.
 *
 */

#include <mm/tlb.h>

#include <drivers/dorder.h>
#include <drivers/printer.h>
#include <exc/syscall.h>
#include <include/shared.h>
#include <include/c.h>
#include <mm/utils.h>
#include <mm/vmm.h>


/** 
 * Counter determing which asid to pick if none is free.
 * If all ASIDs are used then pick this ASID and increment the counter. 
 */
uint8_t last_picked_asid [MAX_CPU];

/**
 * ASID #255 is considered unused by thread and is used
 * to wipe out TLB.
 */
#define UNUSED_ASID 255

/** 
 * @see tlb
 */
struct tlb_entry;
typedef struct tlb_entry {
	uint8_t asid;
	uint32_t virtual_page;
} tlb_entry;

/** 
 * Structure used to keep track processor TLB so we
 * don't have to query registers when clearing page entries
 * for passed ASID or ASID and range of pages.
 */
tlb_entry tlb [MAX_CPU] [CP0_INDEX_INDEX_COUNT];


/** 
 * Initialize address translation structures on current CPU.
 * Clears TLB, structures used to keep track of used ASIDs
 * and structures used to clear TLB.
 */
void init_tlb_cpu (void)
{
	last_picked_asid [cpuid ()] = 0;
	
	/* Init asid mapping array. */
	for (uint32_t i = 0; i < ASID_LENGTH; ++i) 
	{
		asid_used [cpuid ()] [i] = 0;
		asid_thread [cpuid ()] [i] = NULL;
	}

	/* The Wired register contains the number of entries that
	 * are never selected by the random TLB replacement
	 * algorithm. We need none of those now. */
	write_cp0_wired (0);
	
	/* The size of the page to map. */
	write_cp0_pagemask (CP0_PAGEMASK_4K);
	
	tlb_clear_local ();
}


/**
 * Clears all entries in TLB on all CPUs.
 *
 * @see cpu_send_ipi_cause_all
 * @see tlb_clear_local
 */
void tlb_clear( void)
{
	cpu_send_ipi_cause_all (CP0_CAUSE_IPI_TLB);		

	tlb_clear_local ();
}


/** 
 * Clears all entries in TLB on current CPU.
 */
void tlb_clear_local (void)
{
	/* It would be mostly unwise to interrupt this method
	 * due to ASID swaping. */
	ipl_t state = query_and_disable_interrupts ();
	
	/* Current address space identifier. */
	uint32_t current_asid = read_cp0_entryhi () & CP0_ENTRYHI_ASID_MASK;

	/* Shifted already. */
	uint32_t invalid_page = 0;

	/* Unused address space identifier is the important thing.
	 * Used virtual adress is not as long as flags are set invalid. */
	write_cp0_entryhi (invalid_page | UNUSED_ASID);
	
	/* Physical address of the first of the pair of mapped pages. 
	 * With valid bit and dirty bit disabled. */
	write_cp0_entrylo0 (0);
	
	/* Physical address of the second of the pair of mapped pages. 
	 *  With valid bit and dirty bit disabled. */
	write_cp0_entrylo1( 0);
	
	/* Copy this to all TLB entries. */
	for (uint32_t i = 0; i < CP0_INDEX_INDEX_COUNT; i++) 
	{
		tlb [cpuid ()] [i].asid = UNUSED_ASID;
		tlb [cpuid ()] [i].virtual_page = invalid_page;	
		write_cp0_index (i);
		TLBWI ();
	}

	/* Restore current address sapce identifier. */
	write_cp0_entryhi (current_asid);

	conditionally_enable_interrupts (state);
}


/**
 * Clears entries associated with passed ASID in TLB on all CPUs.
 *
 * @param asid
 *	ASID to remove from TLB.
 *
 * @see cpu_send_ipi_cause_all
 * @see tlb_clear_asid_local
 */
void tlb_clear_asid (const uint8_t asid)
{
	cpu_send_ipi_cause_all (CP0_CAUSE_IPI_TLB);		

	tlb_clear_local_asid (asid);
}


/**
 * Clears entries associated with passed ASID in TLB on current CPU.
 *
 * @param asid
 *	ASID to remove from TLB.
 */
void tlb_clear_local_asid (const uint8_t asid)
{
	/* It would be mostly unwise to interrupt this method
	   due to ASID swaping. */
	ipl_t state = query_and_disable_interrupts();
	
	/* Current address space identifier. */
	uint32_t current_asid = read_cp0_entryhi() & CP0_ENTRYHI_ASID_MASK;

	/* Shifted already. */
	uint32_t invalid_page = 0;

	/* Unused address space identifier is the important thing.
	   Used virtual adress is not as long as flags are set invalid. */
	write_cp0_entryhi (invalid_page | UNUSED_ASID);
	
	/* Physical address of the first of the pair of mapped pages. 
	   With valid bit and dirty bit disabled. */
	write_cp0_entrylo0 (0);
	
	/* Physical address of the second of the pair of mapped pages. 
	   With valid bit and dirty bit disabled. */
	write_cp0_entrylo1 (0);
	
	/* Copy this to all TLB entries. */
	for (uint32_t i = 0; i < CP0_INDEX_INDEX_COUNT; i++) 
	{
		if (tlb [cpuid ()] [i].asid == asid)
		{
			tlb [cpuid ()] [i].asid = UNUSED_ASID;
			tlb [cpuid ()] [i].virtual_page = invalid_page;	
			write_cp0_index (i);
			TLBWI ();
		}
	}

	/* Restore current address sapce identifier. */
	write_cp0_entryhi (current_asid);

	conditionally_enable_interrupts (state);
}


/**
 * Clears entries in passed range and associated with passed ASID in TLB on all CPUs.
 *
 * @param low_address
 *	Lowest address to remove from TLB.
 *
 * @param high_address
 *	Highest address to remove from TLB.
 *
 * @param asid
 *	Remove only pages associated to this ASID from TLB.
 *
 * @see cpu_send_ipi_cause_all
 * @see tlb_clear_range_local
 */
void tlb_clear_range (const uint32_t low_address, const uint32_t high_address, uint8_t asid)
{
	cpu_send_ipi_cause_all (CP0_CAUSE_IPI_TLB);		

	tlb_clear_local_range (low_address, high_address, asid);
}


/**
 * Clears entries in passed range and associated with passed ASID in TLB on current CPU.
 *
 * @param low_address
 *	Lowest address to remove from TLB.
 *
 * @param high_address
 *	Highest address to remove from TLB.
 *
 * @param asid
 *	Remove only pages associated to this ASID from TLB.
 */
void tlb_clear_local_range (const uint32_t low_address, const uint32_t high_address, uint8_t asid)
{
	/* It would be mostly unwise to interrupt this method
	 * due to ASID swaping. */
	ipl_t state = query_and_disable_interrupts ();
	
	/* Current address space identifier. */
	uint32_t current_asid = read_cp0_entryhi () & CP0_ENTRYHI_ASID_MASK;

	/* Shifted already. */
	uint32_t invalid_page = 0;

	/* Unused address space identifier is the important thing.
	 * Used virtual adress is not as long as flags are set invalid. */
	write_cp0_entryhi (invalid_page | UNUSED_ASID);
	
	/* Physical address of the first of the pair of mapped pages. 
	 * With valid bit and dirty bit disabled. */
	write_cp0_entrylo0 (0);
	
	/* Physical address of the second of the pair of mapped pages. 
	 * With valid bit and dirty bit disabled. */
	write_cp0_entrylo1 (0);

	/* Convert address to page representation and use even page on 
	   bottom of interval and odd page on top so we don't miss any entry in TLB. */ 
	uint32_t low_page = even_page (page (low_address));
	uint32_t high_page = odd_page (page (high_address));
	
	/* Copy this to all TLB entries. */
	for (uint32_t i = 0; i < CP0_INDEX_INDEX_COUNT; i++) 
	{
		if ( 	tlb [cpuid ()] [i].virtual_page >= low_page &&
			tlb [cpuid ()] [i].virtual_page <= high_page &&
			tlb [cpuid ()] [i].asid == asid)
		{
			tlb [cpuid ()] [i].asid = UNUSED_ASID;
			tlb [cpuid ()] [i].virtual_page = invalid_page;	
			write_cp0_index (i);
			TLBWI ();
		}
	}

	/* Restore current address sapce identifier. */
	write_cp0_entryhi (current_asid);

	conditionally_enable_interrupts (state);
}


/** TLB Refill Exception
 *
 * @param registers 
 */
void wrapped_tlbrefill (context_t *registers)
{
	uint32_t asid = read_cp0_entryhi () & CP0_ENTRYHI_ASID_MASK;
	int32_t tlb_index = read_cp0_random ();
	thread_t thread = asid_thread [cpuid()] [ asid];
	thread_t cthread = thread_get_current (); 
	
	if (thread != cthread)
	{
		printk ("\nTLB: thread %x, cthread %x, asid %d\n", thread, cthread, asid);
		panic ("TLB: asid assigning error.");
 	}	
	
	uint32_t address = read_cp0_badvaddr ();
	uint32_t virtual_page = page (address);
        
	uint32_t even_virtual_page = even_page (virtual_page);
	uint32_t odd_virtual_page = odd_page (virtual_page);

	uint32_t even_page = vm_get_page (thread, even_virtual_page);
	uint32_t odd_page = vm_get_page (thread, odd_virtual_page);

	uint32_t even_flags = CP0_ENTRYLO_D_MASK | CP0_ENTRYLO_V_MASK;
	uint32_t odd_flags = CP0_ENTRYLO_D_MASK | CP0_ENTRYLO_V_MASK;

	if (vm_is_unmapped (thread, virtual_page))
	{
		thread_t this_thread = thread_get_current ();
		if (this_thread->flags & TF_USERLAND)
		{
			printk_not_sync ("Killing process - TLB page not found.\n");
			registers->a0 = SYSCALL_EXIT;
			syscall (registers);
		}
		else
		{
			printk ("\nTLB: page %x not found for address %p.\n", virtual_page, address);
			panic ("TLB: page not found.");
		}
	}

	if (vm_is_unmapped (thread, even_virtual_page)) 
	{
		even_page = 0;
		even_flags = 0;
	}
	if (vm_is_unmapped(thread, odd_virtual_page))
	{
		odd_page = 0;
		odd_flags = 0;
	}

	tlb [cpuid ()] [tlb_index].asid = asid;
	tlb [cpuid ()] [tlb_index].virtual_page = virtual_page;	
	tlb_map (virtual_page, asid, even_page, even_flags, odd_page, odd_flags, tlb_index);
}


/**
 * Chooses and prepares ASID number to be assigned to the just being scheduled thread.
 * @param thread - thread to be scheduled and ASID being picked for
 *
 * @param thread
 *	Thread to choose ASID for.
 *
 * @return
 *	Returns the prepared ASID
 */
uint32_t get_virtual_space (thread_t thread)
{
	if (thread == NULL)
	{	
		panic ("TLB: switching virtual space to thread == NULL");
	}

	/* Determine what ASID to use. */
	int32_t asid = vm_get_asid (thread);

	/* Check if asid assigned in the past is still valid. */
	if (asid != -1 && asid_thread [cpuid ()] [asid] != thread)
	{
		asid = -1;
	}
	
	/* Find out if there's any free ASID to use. 
	 * TODO: There probably should be some better structure for
	 * keeping track of unused ASIDs.*/
	if (asid == -1)
	{
		for (int i = 0; i < CP0_ENTRYHI_ASID_MASK; ++i)
		{
			if (asid_used [cpuid ()][i] == 0)
			{
				asid = i;
				break;
			}
		}
	}	
	
	/* If there's no free ASID, take on from some thread. */
	if (asid == -1) 
	{
		/* Only increment and turn around if needed. */
		last_picked_asid [cpuid ()]++;
		if (last_picked_asid [cpuid ()] == UNUSED_ASID)
		{
			last_picked_asid [cpuid ()]++;
		}
		asid = last_picked_asid [cpuid ()];

		/* Cleaning TLB. */
		tlb_clear_asid (asid);	
	}

	/* Setting asid for new process / thread / space. */
	vm_set_asid (thread, asid);
	asid_used [cpuid ()] [asid] = 1;
	asid_thread [cpuid ()] [asid] = thread;

	return asid;
}

