/**
 * @file utils.c
 *
 * Virtual memory management utils.
 *
 */

#include <mm/utils.h>

#include <drivers/dorder.h>
#include <include/error_codes.h>
#include <mm/tlb.h>


/**
 * Converts full addres to page number.
 *
 * @param address 
 *	Address to convert.
 *
 * @return  
 *	Returns page number in the right most bits
 *	the rest in the left is filled by zeros.
 */
uint32_t page (uint32_t address)
{
	return (address >> UTILS_H_OFFSET_SIZE); 
}


/**
 * Deletes (to zeros) page bits from the address.
 *
 * @param address
 *	Address to clear page bits from.
 *
 * @return 
 * 	Returns only offset of the address.
 */
uint32_t offset (uint32_t address)
{
	return (address & UTILS_H_OFFSET_MASK);
}


/**
 * Combines page number and offset to make full address.
 *
 * @param page
 *	Page to combine with offset.
 *
 * @param offset
 *	Offset to combine with page.
 *
 * @return
 * 	Returns full address combined of page and offset.
 */
uint32_t address (uint32_t page, uint32_t offset)
{
	assert ((page & UTILS_H_PAGE_CHECK) == 0);
	assert ((offset & UTILS_H_OFFSET_CHECK) == 0);
	return ((page << UTILS_H_OFFSET_SIZE) | offset);
}


/** 
 * Gets coresponding even page or the one passed if it's already even.
 *
 * @return
 *	Returns relevant even page.
 */
uint32_t even_page (uint32_t page)
{
	assert ((page & UTILS_H_PAGE_CHECK) == 0);
	return (page & ~1);	
}


/** 
 * Gets coresponding odd page or the one passed if it's already odd.
 *
 * @return
 * 	Returns relevant odd page.
 */
uint32_t odd_page (uint32_t page)
{
	assert ((page & UTILS_H_PAGE_CHECK) == 0);
	return ((page & ~1) + 1);
}


/**
 * Gets following page.
 *
 * @param page
 *	Gets the next page of this one.
 *
 * @return
 *	Returns next page of the passed one.
 */
uint32_t next_page (uint32_t page)
{
	assert ((page & UTILS_H_PAGE_CHECK) == 0);
	assert (page < UTILS_H_PAGE_SIZE);
	return (page + 1);
}


/**
 * Checks if page is even.
 *
 * @param page
 *	Page to check.
 *
 * @return
 *	True - if page is even.
 *	False - else.
 */
uint32_t is_page_even (uint32_t page)
{
	assert ((page & UTILS_H_PAGE_CHECK) == 0);
	return ((page & 1) == 0);
}


/**
 * Checks if odd is even.
 *
 * @param page
 *	Page to check.
 *
 * @return
 *	True - if page is odd.
 *	False - else.
 */
uint32_t is_page_odd (uint32_t page)
{
	assert ((page & UTILS_H_PAGE_CHECK) == 0);
	return ((page & 1) == 1);
}


/**
 * Checks if offset overflowed and reached following page or not.
 *
 * @param offset
 *	Offset to check.
 *
 * @return
 *	True - if it's time to reset offset to zero and to go to the next page.
 *	Flase - if offset is still valid.
 */
uint32_t is_next_page (uint32_t offset)
{
	return (offset > UTILS_H_OFFSET_MASK);
}


/**
 * Gets ASID of currently running thread on current CPU.
 * Reads entryhi register and masks out all but ASIDs bits.
 *
 * @return
 *	ASID of currently running thread set in the entryhi register.
 */
uint32_t get_asid (void)
{
	return (read_cp0_entryhi () & UTILS_H_ASID_MASK);
}


/**
 * Fills passed data into TLB entry of current CPU.
 * Supports writing of 4KB pages only.
 *
 * @param index 
 * 	If index is set to -1, random index is picked from random register.
 *	Otherwise supplied index is used.
 */
void tlb_map( 
	uint32_t virtual_page,
	uint32_t asid,
	uint32_t physical_page0,
	uint32_t flags_page0,
	uint32_t physical_page1,
	uint32_t flags_page1,
	int32_t index)
{
	write_cp0_pagemask (CP0_PAGEMASK_4K);
	
	virtual_page = virtual_page >> 1;
	write_cp0_entryhi ((virtual_page << 13) | asid);
	
	write_cp0_entrylo0 (((physical_page0 << 6) | flags_page0));
	
	write_cp0_entrylo1 (((physical_page1 << 6) | flags_page1));
	
	if (index != -1) 
	{
		write_cp0_index (index);
		TLBWI ();
	}
	else
	{
		TLBWR ();
	}
}


/**
 * Gets random virtual page which is not used (whole pair not used).
 * Algorithm of this function is quite ineffective and more tight 
 * integration with virtual page tables and allocation blocks should
 * should be used for faster run. 
 *
 * @param thread
 *	Thread to find unused pair of virtual pages for.
 *
 * @return 
 *	Unused pair represented by the even page.
 */
uint32_t get_unmapped (thread_t thread);
uint32_t get_unmapped (thread_t thread)
{
	uint32_t epage = 0x00000000;
	uint32_t opage = 0x00000001;

	while (true)
	{
		if (vm_is_unmapped (thread, epage) && vm_is_unmapped (thread, opage))
			break;

		epage += 2;
		opage += 2;
	}
	return epage;
}


/**
 * Copy bytes from one's thread virtual address space to another's one.
 *
 * @param from_thread
 *	Thread to copy bytes from.
 *
 * @param to_thread
 *	Thread to copy bytes to.
 * 
 * @param from_active 
 * 	True - the from_thread is the calling one.
 * 
 * @param to_active
 * 	True - the to_thread is the calling one.
 *
 * @param from_ptr
 *	Source pointer to data in form_thread.
 *
 * @param to_ptr
 *	Destination pointer to data in to_thread.
 *
 * @param length
 *	Number of bytes to copy.
 *
 * @return
 *	EOK - if copied was successfull.
 */
int copy_from_to (thread_t from_thread, thread_t to_thread, bool from_active, bool to_active, void* from_ptr, void* to_ptr, size_t length);
int copy_from_to(
	thread_t from_thread,
	thread_t to_thread,
	bool from_active,
	bool to_active,
	void* from_ptr,
	void* to_ptr,
	size_t length)
{
	/* It's mostly important to disable interrupts, for another thread
	 * using the same virtual address space and the same ASID might be scheduled
	 * and may accidentally use the temporary mapping used in this function. */
	ipl_t state = query_and_disable_interrupts (); 

	uint32_t from_page = page ((uint32_t) from_ptr);
	uint32_t from_offset = offset ((uint32_t) from_ptr);
	uint32_t to_page = page ((uint32_t) to_ptr);
	uint32_t to_offset = offset ((uint32_t) to_ptr);
	
 	/* Virtual pointer of source data. */
	uint8_t* from = (uint8_t*) from_ptr;

	/* Virtual pointer of data destionation */
	uint8_t* to = (uint8_t*) to_ptr;
	
	/* Virtual page of the not calling thread. */
	uint32_t virtual_page;

	/* Physical page of the not calling thread. */
	uint32_t physical_page;
	
	/* ASID of calling thread. */
	int32_t asid;

	/* Index used for temporary mapping in TLB. */
	int32_t index = 0;
	
	/* Valid access flags of physical page in TLB entry. */
	uint32_t flags = CP0_ENTRYLO_D_MASK | CP0_ENTRYLO_V_MASK;
	
	/* Invalid physcal page in TLB. */
	uint32_t no_page = 0;
 
	/* Invalid flags of physical page in TLB which cause physical page
	 * inaccessible. */
	uint32_t no_flags = 0;

	/* This function is used to copy bytes from the calling thread
	 * to another one or vice-versa to copy bytes to the calling thread
	 * from another one. And we must be aware which situation are we handling. */	
	if (from_active == true) {
		asid = vm_get_asid (from_thread);
	} else {
		asid = vm_get_asid (to_thread);
	}

	/* Inner cycle counter. */
	size_t loop;

	/* Inner cycle magic size. */
	size_t loop_max = 40;
	
	size_t i = 0;
	while (i < length)
	{
		/* We need to find unused page to map the physical frame of the
		 * other thread to this thread using TLB only. */
		if (from_active == true)
		{
			virtual_page = get_unmapped (from_thread);
			physical_page = vm_get_page (to_thread, to_page);
			to = (uint8_t*) address (virtual_page, to_offset);
		}
		else
		{
			virtual_page = get_unmapped (to_thread);
			physical_page = vm_get_page (from_thread, from_page);
			from = (uint8_t*) address (virtual_page, from_offset);
		}
		
		/* Temporarily map this page in TLB. */
		if ( is_page_even (virtual_page)) {
			tlb_map (virtual_page, asid, physical_page, flags, no_page, no_page, index);	
		} else {	
			tlb_map (virtual_page, asid, no_page, no_flags, physical_page, flags, index);
		}

		/* This inner loop was meant to copy some bytes and then enable interrupts for
		 * a while so another threads may run and when this thread is scheduled again
		 * continue copying. This would not allow this thread to occupy CPU for too long
		 * if huge amount of data is copied. It is not used right now due to unknown reasons. */
		loop = 0;	
		while (i < length && loop < loop_max)
		{
			/* Actual copying of bytes from one thread to another. */
			*to = *from;

			from++;
			to++;
			from_offset++;
			to_offset++;
			
			/* Determines if just increase offset on the current page or
			 * move to the next page and change mapping in TLB.
			 * This is acutally quite ineffective to check it byte by byte.
			 * TODO: Get how many times it's safe to just increase offset and then 
			 * automatically switch to the next page. */
			if (from_active == true && is_next_page (to_offset))
			{
				to_page = next_page (to_page);
				to_offset = 0;
				to = (uint8_t*) address (virtual_page, to_offset);
				physical_page = vm_get_page (to_thread, to_page);

				if (is_page_even (virtual_page)) {
					tlb_map (virtual_page, asid, physical_page, flags, no_page, no_flags, index);	
				} else {	
					tlb_map (virtual_page, asid, no_page, no_flags, physical_page, flags, index);
				}
			}
			if (to_active == true && is_next_page (from_offset))
			{
				from_page = next_page (from_page);
				from_offset = 0;
				from = (uint8_t*) address (virtual_page, from_offset);
				physical_page = vm_get_page (from_thread, from_page);

				if (is_page_even (virtual_page)) {
					tlb_map (virtual_page, asid, physical_page, flags, no_page, no_flags, index);	
				} else {	
					tlb_map (virtual_page, asid, no_page, no_flags, physical_page, flags, index);
				}
			}
			
			i++;
			loop++;
		}
		/* Clears index used in TLB for temporary mapping. */
		tlb_map (0, asid, no_page, no_flags, no_page, no_flags, index);

	}
	/* Clears index used in TLB for temporary mapping. */
	tlb_map (0, asid, no_page, no_flags, no_page, no_flags, index);
	
	conditionally_enable_interrupts (state);

	return EOK;
}


/**
 * Copy bytes from the calling thread to another one.
 *	
 * @param thread
 *	Thread to copy bytes to.
 *
 * @param destination
 *	Destination pointer to data in the virtual address space of the passed thread.
 *
 * @param source
 *	Source pointer to data in the virtual address space of the calling thread.
 *
 * @param length
 *	Number of bytes to copy.
 *
 * @return
 *	EOK - if copied successfully.
 *
 * @see copy_from_to
 */
int copy_to_thread(
	const thread_t thread,
	void* destination,
	const void* source,
	const size_t length)
{
	uint32_t asid = (read_cp0_entryhi () & UTILS_H_ASID_MASK);

	thread_t from = asid_thread [cpuid ()] [ asid];
	thread_t to = thread;

	bool from_active = true;
	bool to_active = false;

	return copy_from_to (from, to, from_active, to_active, (void*)source, (void*)destination, length);
}


/**
 * Copy bytes to the calling the thread from another one.
 *
 * @param thread
 *	Thread to copy bytes from.
 *
 * @param destination
 *	Destination pointer to data in the virtual address space of the calling thread.
 *
 * @param source
 *	Source pointer to data in the virtual address space of the passed thread.
 *
 * @param length
 *	Number of bytes to copy.
 *
 * @return
 *	EOK - if copied successfully.
 *
 * @see copy_from_to
 */
int copy_from_thread(
	const thread_t thread,
	void* destination,
	const void* source,
	const size_t length)
{
	uint32_t asid = (read_cp0_entryhi () & UTILS_H_ASID_MASK);

	thread_t from = thread;
	thread_t to = asid_thread [cpuid ()] [ asid];

	bool from_active = false;
	bool to_active = true;

	return copy_from_to (from, to, from_active, to_active, (void*)source, (void*)destination, length);
}

