/**
 * @file virtmem.c
 *
 * Help methods for virtual memory manager.
 * These methods are called in the general functions of virtual memory manager. The methods are not synchronized, 
 * the synchronization must be provided by calling method. Every particularly searching in VM space have linear 
 * complexity for VM_MAX_AREAS.
 * @link kernel/mm/vmm.h Virtual memory manager @endlink
 *
 * @see vma_map
 */

#include <include/error_codes.h>
#include <mm/malloc.h>
#include <mm/physmem.h>
#include <mm/tlb.h>
#include <sync/mutex.h>
#include <exc/assert.h>
#include <mm/virtmem.h>


/**
 * Create a virtual mapping for given block of memory.
 * Create a virtual mapping for block of size @size at virtual address @virtual_address. The virtual addresses will 
 * be mapped to @physical_address for the given @size of memory. We suppose, that all the physical addresses is
 * already allocated and the block of virtual memory is unmapped.
 *
 * @param thread The thread.
 * @param virtual_address The starting virtual address of block.
 * @param physical_address The starting physical address of block.
 * @param size The size of block.
 */
void vm_create_mapping (thread_t thread, uint32_t virtual_address, uint32_t physical_address, size_t size)
{
	/* Creates a mapping for each page. */
	uint32_t allocated_size = 0;
	while(allocated_size < size)
	{
		/* Master index. */
		uint32_t mas_i = VM_GET_MAS_INDEX_ADDRESS(virtual_address + allocated_size);

		/* If master index isn't allocated. */
		if (*(thread->vm_map+mas_i) == NULL)
		{
			*(thread->vm_map+mas_i) = malloc (sizeof (vm_seci_t));
			(*(thread->vm_map+mas_i))->pages = malloc ((sizeof (uint32_t**)*VM_SEC_INDEX_SIZE));

			/* Initialize all the pages to NULL. */
			for (uint32_t k=0; k<VM_SEC_INDEX_SIZE; ++k)
			{
				*((*(thread->vm_map+mas_i))->pages+k) = NULL;
			}
		}

		/* Second index. */
		uint32_t sec_i = VM_GET_SEC_INDEX_ADDRESS(virtual_address + allocated_size);

		/* The record in VM table should be empty. */
		if (*((*(thread->vm_map+mas_i))->pages+sec_i) != NULL)
		{
			panic ("Virtual memory: Overwriting the record in virtual memory table, thread %p", thread);
		}

		/* Alloc and assign the new record in VM table. */
		*((*(thread->vm_map+mas_i))->pages+sec_i) = malloc (sizeof (uint32_t));
		*(*((*(thread->vm_map+mas_i))->pages+sec_i)) = (uint32_t)(physical_address + allocated_size);

		allocated_size += VM_PAGE_SIZE;
	}
}


/**
 * Add the new block of allocated virtual memory to internal structures.
 * Creates the new area (block of allocated virtual memory) and add to VM structure of given @thread. The list of 
 * areas for the VM space is sorted by starting addresses. We suppose, that there is a free space for the new area 
 * in the VM space, it means that the number of areas must not be greater then VM_MAX_AREAS. This method doesn't 
 * create a mapping!
 *
 * @param thread The thread.
 * @param from The starting virtual address of area.
 * @param size The size of block.
 * @param mode The access mode (Kernel/Supervisor/User).
 */
void vm_add_area (thread_t thread, uint32_t from, size_t size)
{
	/* Create the new area. */
	vm_area_t* new_area = malloc (sizeof (vm_area_t));
	
	new_area->size = size;
	new_area->from = from;
	new_area->thread_id = (uint32_t) thread;
	
	vm_area_t** areas = thread->vm_areas;
	
	/* Find the suitable position for the new area in the sorted list. There is an assumption, that there is a 
	 * free space in the array of allocated blocks. */
	int i = 0;
	while(((*(areas + i)) != NULL) &&
		(((*(areas + i))->from) < from))
	{
		++i;
	}

	if(*(areas+i) == NULL)
	{
		/* New area is the last item of sorted list.*/
		*(areas + i) = new_area;
	}
	else
	{
		/* Add the new area to the sorted list and tranfer the elements behind this one back. */
		vm_area_t* prev = new_area;
		while((*(areas+i)) != NULL)
		{
			vm_area_t* tmp = *(areas+i);
			*(areas+i) = prev;
			prev = tmp;
			++i;
		}
		*(areas+i) = prev;
	}
}


/** 
 * Choose the right bounds for searching the suitable block of virtual memory. 
 * This is the helper method for searching the available block of virtual memory. The searching starts at the start 
 * address and continues until it finds the free block of given @size or reach the end searching address. The bounds 
 * are set according to the flags that represent information, where we had to allocate the block.
 *
 * (flags & VF_VA_USER) - User chose the virtual address of block in @from.
 *
 * (flags & VF_VA_AUTO) - Choose the appropriate bounds depending on the other flags specifying the segment (VF_AT_KUSEG, 
 *	VF_AT_KSEG0, VF_AT_KSEG1, VF_AT_KSSEG, VF_AT_KSEG3)
 *
 * @param ss_address The start address for block searching.
 * @param se_address The end address for block searching.
 * @param from  The starting adress of the new allocated block (if flag VF_VA_USER).
 * @param size  The size of the new block.
 * @param flags Specific flags for allocating.
 * @see VM_SAFE_RANGE
 */
void set_searching_addresses (uint32_t* ssaddress, uint32_t* seaddress, uint32_t from, const size_t size, const vm_flags_t flags)
{
	
	/* User chose the address for allocating. */
	if (flags & VF_VA_USER)
	{
		(*ssaddress) = from;
		(*seaddress) = from - 1 + size;

		return;
	}

	/* Automatic block selection. We can allocate anywhere (depends on the other flags specifying the segment) where 
	 * is suitable space. */
	if (flags & VF_VA_AUTO)
	{
		if (flags & VF_AT_KUSEG)
		{
			/* See VM_SAFE_RANGE comment. */
			if(flags & VM_SAFE_RANGE)
			{
				(*ssaddress) = VM_USEG_PREFIX_SAFE;
			}
			else
			{
				(*ssaddress) = VM_USEG_PREFIX;
			}
			
			(*seaddress) = VM_USEG_SUFFIX;
		}
		else if (flags & VF_AT_KSEG0)
		{
			(*ssaddress) = VM_KSEG0_PREFIX;
			(*seaddress) = VM_KSEG0_SUFFIX;
		}
		else if (flags & VF_AT_KSEG1)
		{
			(*ssaddress) = VM_KSEG1_PREFIX;
			(*seaddress) = VM_KSEG1_SUFFIX;
		}
		else if (flags & VF_AT_KSSEG)
		{
			(*ssaddress) = VM_SSEG_PREFIX;
			(*seaddress) = VM_SSEG_SUFFIX;
		}
		else if (flags & VF_AT_KSEG3)
		{
			(*ssaddress) = VM_KSEG3_PREFIX;
			(*seaddress) = VM_KSEG3_SUFFIX;
		}
		else
		{
			panic ("An unknown flag for allocating virtual memory: %i", flags);
		}
	}
}


/**
 * Find the first available block of virtual memory for the given @size.
 * Searching is limited by @ss_address and @se_address - these are the bounds, where the block should be. The searching 
 * starts at the @ss_address and goes through the allocated areas until find the block of suitable size or reach the @se_address.
 * The searching complexity is linear for VM_MAX_AREAS.
 *
 * @param areas Allocated areas in VM space.
 * @param from  The address of a available block (if any).
 * @param size  The size of the block.
 * @param ss_address The starting address for searching.
 * @param se_address The ending address for searching.
 * @return EOK if the available block of given size was found, ENOMEM else.
 */
int search_avail_block (vm_area_t** areas, uint32_t* from, const size_t size, uint32_t ss_address, uint32_t se_address)
{

	/* First we find the allocated block directly after the ss_address (lower limit). Then we go through the other 
	 * allocated blocks and check, if there is a suitable free space for a new block between two allocated blocks. 
	 * If we reach the se_address, we know there isn't enough free space in the specified range [ss_address;se_address]. */

	uint32_t search_address = ss_address;
	int i = 0;

	/* sync = Indexed area is directly after the @searching_address. */
	bool sync = false;

	while( (search_address <= se_address) && (*(areas+i) != NULL))
	{
		/* If the indexed area isn't directly after the @search_address. */
		if (!sync)
		{
			if (((*(areas+i))->from + (*(areas+i))->size-1) < search_address)
			{
				/* The whole indexed block is before the @search_address. We can get on the next allocated block. */
				++i;
				continue;
			}
			else
			{
				/* Is the @search_address inside the indexed block? */
				if (((*(areas+i))->from <= search_address) &&
					(((*(areas+i))->from + (*(areas+i))->size-1) >= search_address))
				{
					/* Move the @search_address to the first free address, it means after the allocated block. */
					search_address = (*(areas+i))->from +
					(*(areas+i))->size;
					++i;
				}

				/* Now the allocated block is directly after the @search_address. */
				sync = true;
				continue;
			}
		}

		/* Now the indexed block is directly after the @search_address */
		size_t free_space = (*(areas+i))->from - search_address;
		if (free_space >= size)
		{
			/* Is the end of block after the upper limit for searching? */
			if ((search_address + size-1) <= se_address)
			{
				(*from) = search_address;
				return EOK;
			}
			else
			{
				return ENOMEM;
			}
		}
		else
		{
			/* There is not enough space for allocating, set the @search_address to the first free address after 
			 * indexed block, get on the next allocated block and repeat the cycle. */
			search_address = (*(areas+i))->from + (*(areas+i))->size;
			++i;
		}
	}

	/* Now we reach the upper limit (se_address) or there isn't any more areas to search.  */
 
	if((*(areas+i) == NULL) &&
		(se_address > search_address) &&
		((se_address - search_address+1) >= size))
	{
		/* There is no other areas to search, the rest of the virtual memory since the @search_address to @seaddress 
		 * is free. Both condition for @search_address and @se_address must be tested, as it might overflow unsigned 
		 * long... */
		(*from) = search_address;
		return EOK;
	}
	else
	{
		/* Upper limit (@se_address) was reached or there isn't enough space to allocate. */
		return ENOMEM;
	}
}


/**
 * Checks for accessability for the given block of memory.
 * Checks, if the block starting at @start_address with size @size is not at the border of two segments and if it 
 * is accessible for the current mode.
 *
 * Access limitation for separate modes: 
 * USEG  - kernel/supervisor/user;
 * KSEG0 - kernel;
 * KSEG1 - kernel;
 * SSEG  - kernel/supervisor;
 * KSEG2 - kernel;
 *
 * Other parameteres is the searching limits, which are modified when the need for a new block searching with the new 
 * limits. It can occur, if the block is at the border of USEG and KSEG0 and the current mode is SUPERVISOR. Then we 
 * modify the lower and upper limit to the range of SSEG, because this is the only next space for searching. Other cases 
 * we need to repeat the searching with modifyed searching limits occur, when the start and end addresses aren't in the 
 * same segment. Then we set the searching limits to the segment, where is the end address, and start the searching again.
 *
 * @param start_address The starting address of the block.
 * @param size The size of the block.
 * @param ss_address The lower limit for searching.
 * @param se_address The upper limit for searching.
 * @param flags The specifying information for allocation.
 * @return EOK if the whole block is in one segment and accessible in current mode, EINVAL if is not, but the limits was 
 *	changed, ENOMEM if it is useless to search again - there is no suitable block.
 * @see vma_map
 */
int check_access (const uint32_t start_address, const size_t size, uint32_t* ss_address, uint32_t* se_address, const vm_flags_t flags)
{

	vm_segment s_seg = get_vm_segment (start_address);
	vm_segment e_seg = get_vm_segment ((start_address+size-1));
	vm_mode mode = get_vm_mode ();

	/* Is the start and end addresses of the block in the same segment? */
	if (s_seg == e_seg)
	{
		/* Check the access limitations. */
		if ((mode == KERNEL) ||
			((e_seg == SSEG) && (mode == SUPERVISOR)) ||
			((e_seg == USEG) && ((mode = SUPERVISOR) || (mode = USER))))
		{
			return EOK;
		}
		else
		{
			/* If the mode is SUPERVISOR, modify the limits for searching. See the function comment. */
			if ((mode == SUPERVISOR) && ((e_seg == KSEG0) || (e_seg == KSEG1)))
			{
				(*ss_address) = VM_SSEG_PREFIX;
				(*se_address) = VM_SSEG_SUFFIX;
				return EINVAL;
			}
			else
			{
				/* Not accessible block of memory. */
				return ENOMEM;
			}
		}
	}
	else
	{
		/* Start and end addresses of the block aren't in the same segment. */
		if ((mode == KERNEL))
		{
			/* Set the lower limit for searching to the prefix of segment where is the end address of block. */
			(*ss_address) = get_seg_prefix (e_seg);
			return EINVAL;
		}
		else if (mode == SUPERVISOR)
		{
			/* If the mode is SUPERVISOR, modify the limits for searching. See the function comment. */
			(*ss_address) = VM_SSEG_PREFIX;
			(*se_address) = VM_SSEG_SUFFIX;
			return EINVAL;
		}
		else
		{
			/* Not accessible block of memory. */
			return ENOMEM;
		}
	}
}


/**
 * Create a virtual memory area at the given @address of size @size. 
 * Add the area to the sorted list of areas and create mapping between virtual and allocated physical memory.
 * We suppose, that @size is aligned to the size of page (VM_PAGE_SIZE).
 *
 * @param thread The thread.
 * @param address The start address of the block.
 * @param size The size of the block.
 * @param flags The specifying information for allocation.
 * @return EOK if successful, other error code generates physical memory allocation
 * @see vm_add_area
 * @see frame_alloc
 * @see vm_create_mapping
 * @see vma_map
 */
int add_mapping (thread_t thread, uint32_t address, const size_t size, const vm_flags_t flags)
{

	/* Create and add the new area to the VM structure. */
	vm_add_area (thread, address, size);

	/* Physical memory allocation. */
	void* phys_address = 0;

	/* Number of frames - We suppose, that the size of page is equal to size of frame and that @size is aligned 
	 * to the page_size (VM_PAGE_SIZE). */
	size_t falloc_frames = (size/VM_PAGE_SIZE);
	int falloc_res = frame_alloc (&phys_address, falloc_frames, flags);
	if (falloc_res != EOK)
	{
		/* Some error in frame_alloc. */
		return falloc_res;
	}

	/* Create mapping - update virtual table for new data. */
	vm_create_mapping (thread, address, (uint32_t) phys_address, size);

	/* Allocation successful. */
	return EOK;
}


/**
 * Checks if the given block of virtual memory is unmapped. 
 * Linear complexity for VM_MAX_AREAS.
 *
 * @param thread The thread.
 * @param address The start address of block.
 * @param size The block size.
 * @return Is the block unmapped? (bool)
 */
bool vm_block_is_unmapped (thread_t thread, uint32_t address, size_t size)
{
	/* Find the first allocated block after the given @address. */
	size_t index = 0;
	while((*(thread->vm_areas+index) != NULL) && (address > ((*(thread->vm_areas+index))->from +
			(*(thread->vm_areas+index))->size-1)) && (index < VM_MAX_AREAS))
	{
		index++;
	}

	vm_area_t* area = *(thread->vm_areas + index);

	/* If the @address are after all the allocated blocks and the @address + @size is in the range of virtual memory, 
	 * we can return true. */
	if ((area == NULL) && ((address+size-1) <= VM_KSEG3_SUFFIX))
	{
		return true;
	}

	/* If there is a maximum allocated blocks for the VM space. */
	if(index == VM_MAX_AREAS)
	{
		return false;
	}

	/* Is the @address before the allocated block? */
	if(address < area->from)
	{
		/* Is the whole block before the allocated block? */
		if((address+size-1) < area->from)
		{
			return true;
		}
		else
		{
			return false;
		}
	}
	else
	{
		/* @address is inside the allocated block. */
		return false;
	}
}


/**
 * Removes mapping from virtual to physical memory for the given block. 
 * The block is identified by the index @index in the array of VM blocks. Removes the block from the VM structures.
 *
 * @param areas The array of VM blocks.
 * @param index The index of removed block.
 * @return EOK if the memory is successfully unallocated and unmapped, other error code generates the physical memory freeing
 * @see tlb_clear_range
 * @see frame_free
 */
int vma_unmap_block (thread_t thread, const int index)
{
	
	vm_area_t** areas = thread->vm_areas;
	
	size_t size = (*(areas+index))->size;
	uint32_t from = (*(thread->vm_areas+index))->from;
	
	/* We need to delete the old records in TLB for all the CPUs. */
	tlb_clear_range (from, from + size-1, vm_get_asid (thread));
	
	/* Free the area and move the following items in the list one forward. */
	free ((*(areas+index)));
	int j = index + 1;
	while (j < VM_MAX_AREAS)
	{
		*(areas + j - 1) = *(areas + j);
		++j;
	}
	*(areas + VM_MAX_AREAS-1) = NULL;


	/* Free the allocated physical memory associated to the block. */
	uint32_t mas_i = VM_GET_MAS_INDEX_ADDRESS(from);
	uint32_t sec_i = VM_GET_SEC_INDEX_ADDRESS(from);

	/* Number of frames - We suppose, that the size of page is equal to size of frame and that @size is aligned to 
	 * the page_size (VM_PAGE_SIZE). */
	int res_ffree = frame_free ((void*)(*(*((*(thread->vm_map+mas_i))->pages+sec_i))), size/VM_PAGE_SIZE);
	if (res_ffree != EOK)
	{
		/* Some error in physical memory freeing. */
		return res_ffree;
	}

	/* Free all the pages in the block. */
	size_t deallocated_size = 0;
	while(deallocated_size < size)
	{
		mas_i = VM_GET_MAS_INDEX_ADDRESS((from + deallocated_size));
		sec_i = VM_GET_SEC_INDEX_ADDRESS((from + deallocated_size));

		/* The records for the pages shouldn't be NULL. */
		if((*(thread->vm_map+mas_i) == NULL) ||
			(*((*(thread->vm_map+mas_i))->pages+sec_i) == NULL))
		{
			panic ("Virtual memory: Unmapping the unallocated pages, thread %p", thread);
		}

		/* Free the page and set to NULL (initialization). */
		free ((*((*(thread->vm_map+mas_i))->pages+sec_i)));
		*((*(thread->vm_map+mas_i))->pages+sec_i) = NULL;

		deallocated_size += VM_PAGE_SIZE;
	}

	return EOK;
}
