/**
 * @file malloc.c
 *
 * Kernel memory allocator.
 *
 * The allocator expects the heap to reside in a continuous physical memory range.
 * Heap blocks have headers and footers and are kept next to each other so that
 * traversing can be done by adding or subtracting the appropriate size.
 *
 * The allocation policy is first fit, blocks are merged on free.
 *
 * Kalisto
 *
 * Copyright (c) 2001-2010
 *   Department of Distributed and Dependable Systems
 *   Faculty of Mathematics and Physics
 *   Charles University, Czech Republic
 *
 */

#include <include/shared.h>
#include <include/c.h>

#include <drivers/printer.h>
#include <exc/assert.h>
#include <mm/memtest.h>
#include <sync/mutex.h>
#include <mm/physmem.h>
#include <mm/vmm.h>
#include <include/error_codes.h>

#include <mm/vmalloc.h>

/** Some maximum heap block size to avoid overflow. */
#define HEAP_BLOCK_SIZE_MAX    0x10000000

/** Generic memory alignment */
#define ALIGNMENT  4


/** heap block structure
 *
 */
typedef struct
{
	// true for free block, false for used block
	bool   is_free;
	// size of whole block including header structure
	size_t block_size;
	// pointer to previous block
	void * previous_block;
	// pointer to next block
	void * next_block;
} heap_block_t;


/** Initialize the heap allocator
 *
 * Finds how much physical memory we have and creates
 * the heap management structures that mark the whole
 * physical memory as a single free block.
 *
 */
void init_virtual_memory (void)
{
}


/** Initialize a heap block
 *
 * Fills in the structures related to a heap block.
 *
 * @param addr Address of the block.
 * @param size Size of the block including the header and the footer.
 * @param free Indication of a free block.
 * @param prev pointer to previous block
 * @param next pointer to next block
 *
 */
static void virtual_block_init (void * addr, size_t size, bool free, void * prev, void * next);
static void virtual_block_init (void * addr, size_t size, bool free, void * prev, void * next)
{
	heap_block_t * heap_block = (heap_block_t *)(addr);

	heap_block->block_size     = size;
	heap_block->is_free        = free;
	heap_block->previous_block = prev;
	heap_block->next_block     = next;
}


/** finds the next block
 *
 * @param this_block actual block for finding the next
 *
 * @return returns the next block
 *
 */
heap_block_t * virtual_get_next_block (void * this_block);
heap_block_t * virtual_get_next_block (void * this_block)
{
	return (heap_block_t *)(((heap_block_t *)(this_block))->next_block);
}


/** finds previous block
 *
 * @param this_block actual block for finding the next
 *
 * @return returns previous block
 *
 */
heap_block_t * virtual_get_previous_block (void * this_block);
heap_block_t * virtual_get_previous_block (void * this_block)
{
	return (heap_block_t *)(((heap_block_t *)(this_block))->previous_block);
}


/** splits a free block to two blocks according to size of the first one which is set NOT free
 *
 * @param address address of the block to divide
 * @param real_size size of the first block (including header structure)
 *
 */
void virtual_block_split (void * address, size_t real_size);
void virtual_block_split (void * address, size_t real_size)
{
	heap_block_t * old_block = (heap_block_t *)(address);
	size_t         old_size  = old_block->block_size;

	assert(old_block->is_free);

	if (old_block->block_size < real_size + sizeof(heap_block_t))
	{
		old_block->is_free = 0;
	}
	else
	{
		void * prev = old_block->previous_block;
		void * next = old_block->next_block;
		void * scnd = address + real_size;
		virtual_block_init(address, real_size, 0, prev, scnd);
		virtual_block_init(scnd, old_size - real_size, 1, address, next);
		if (((heap_block_t *)(scnd))->next_block)
		{
			((heap_block_t *)(((heap_block_t *)(scnd))->next_block))->previous_block = scnd;
		}
	}
}


/** joins block with its free neighbours and returns that huge block
 *
 * @param this_block address of block for joining neighbours blocks
 *
 * @return returns a new huge free block
 *
 */
heap_block_t* virtual_join_blocks (heap_block_t * this_block);
heap_block_t* virtual_join_blocks (heap_block_t * this_block)
{
	assert (this_block);
	assert (this_block->is_free);

	heap_block_t * next_block = virtual_get_next_block((void *)this_block);
	if ((next_block) && (next_block->is_free) && ((void *)this_block + this_block->block_size == (void *)next_block))
	{
		virtual_block_init((void *)this_block, this_block->block_size + next_block->block_size, 1, this_block->previous_block, next_block->next_block);
		if (this_block->next_block)
		{
			((heap_block_t *)(this_block->next_block))->previous_block = (void *)this_block;
		}
	}

	heap_block_t * prev_block = virtual_get_previous_block((void *)this_block);
	if ((prev_block) && (prev_block->is_free) && ((void *)prev_block + prev_block->block_size == (void *)this_block))
	{
		virtual_block_init((void *)prev_block, prev_block->block_size + this_block->block_size, 1, prev_block->previous_block, this_block->next_block);
		if (prev_block->next_block)
		{
			((heap_block_t *)(prev_block->next_block))->previous_block = (void *)prev_block;
		}
		this_block = prev_block;
	}

	return this_block;
}


/** prints a kernel heap for debugging purposes
 *
 */
void virtual_heap_print (void)
{
	thread_t thread = thread_get_current();

	printk_not_sync("\nVIRTUAL HEAP (starts at %p)\n", thread->heap_start);
	heap_block_t * block = (heap_block_t *)(thread->heap_start);
	while (block)
	{
		printk_not_sync("[address %p; size %u; free %u; prev %p; next %p]\n",
		                 block,
		                 block->block_size,
		                 block->is_free,
		                 block->previous_block,
		                 block->next_block);
		block = virtual_get_next_block((void *)block);
	}

}


/** 
 *
 * @param size The size of the block to allocate.
 * @return The address of the block or NULL when not enough memory.
 *
 */
void * virtual_malloc( const size_t size)
{
	thread_t thread = thread_get_current ();	
	spinlock_lock (&(thread->heap_lock));

	/* Checking for maximum size avoids errors due to
	   overflow, which would be hard to debug. */
	assert (size <= HEAP_BLOCK_SIZE_MAX);
	
	/* We have to allocate a bit more to have room for
	   header and footer. The size of the memory block
	   also has to be 4 bytes aligned to avoid unaligned
	   memory access exception while accessing the
	   footer structure. */
	size_t real_size = ALIGN_UP (size, ALIGNMENT) + sizeof (heap_block_t);
	
	void * result = NULL;

	heap_block_t * pos        = (heap_block_t *)thread->heap_start;
	heap_block_t * last_block = NULL;
	
	while ((!result) && (pos)) 
	{
		/* Try to find a block that is free and large enough. */
		if ((pos->is_free) && (pos->block_size >= real_size)) 
		{
			/* We have found a suitable block.
			   See if we should split it. */
			virtual_block_split((void *)pos, real_size);
			
			/* Either way we have our result. */
			result = ((void *) pos) + sizeof (heap_block_t);
		}
		/* Advance to the next block. */
		last_block = pos;
		pos        = virtual_get_next_block((void *)pos);
	}
	if (!result)
	{
		void * new_frame = NULL;
		size_t frames_count = (real_size - 1) / PHYSMEM_PAGE_SIZE + 1;
		if (EOK != vma_map( &new_frame, frames_count * PHYSMEM_PAGE_SIZE, VF_VA_AUTO | VF_AT_KUSEG | VM_SAFE_RANGE))
		{
			result = NULL;
		}
		else
		{
			if (last_block)
			{
				virtual_block_init(new_frame, frames_count * PHYSMEM_PAGE_SIZE, 1, (void *)last_block, last_block->next_block);
				last_block->next_block = new_frame;
			}
			else
			{
				thread->heap_start = new_frame;
				virtual_block_init(new_frame, frames_count * PHYSMEM_PAGE_SIZE, 1, NULL, NULL);			
			}
			virtual_block_split(new_frame, real_size);
			result = new_frame + sizeof(heap_block_t);
		}
	}

	spinlock_unlock (&(thread->heap_lock));

	return result;
}


/** 
 *
 * @param free_block free block to release something inside
 *
 */
void virtual_try_to_physical_free (heap_block_t * free_block);
void virtual_try_to_physical_free (heap_block_t * free_block)
{
	assert (free_block);
	assert (free_block->is_free);
	
	void* vm_block_start = free_block;
	void* vm_block_end = (void*) ((uint32_t)free_block + free_block->block_size);
	
	void* heap_block_start = free_block; 
	void* heap_block_end = (void*) ((uint32_t)free_block + free_block->block_size);
	
	vma_find_block (&vm_block_start, &vm_block_end);
	
	/* No free vm block found. */
	if (vm_block_start == NULL && vm_block_end == NULL)
	{
		return;
	}
	
	heap_block_t* prev = free_block->previous_block;
	heap_block_t* next = free_block->next_block;

	if ((uint32_t) vm_block_start == (uint32_t) heap_block_start && (uint32_t) vm_block_end == (uint32_t) heap_block_end)
	{
		if (prev)
			prev->next_block     = (void*)next;
		if (next)
			next->previous_block = (void*)prev;
	
		vma_unmap (vm_block_start);
	}
	else if ((uint32_t) vm_block_start > (uint32_t) heap_block_start && (uint32_t) vm_block_end == (uint32_t) heap_block_end)
	{
		virtual_block_init (heap_block_start, (uint32_t) vm_block_start - (uint32_t) heap_block_start, 1, (void*) prev, (void*) next);
	
		if (prev)
			prev->next_block     = heap_block_start;
		if (next)
			next->previous_block = heap_block_start;
	
		vma_unmap( vm_block_start);
	}
	else if ((uint32_t) vm_block_start == (uint32_t) heap_block_start && (uint32_t) vm_block_end < (uint32_t) heap_block_end)
	{
		virtual_block_init (vm_block_end, (uint32_t) heap_block_end - (uint32_t) vm_block_end, 1, (void*) prev, (void*) next);
		
		if (prev)
			prev->next_block     = vm_block_end;
		if (next)
			next->previous_block = vm_block_end;
		
		vma_unmap (vm_block_start);
	}
	else if ((uint32_t) vm_block_start != (uint32_t) heap_block_start && (uint32_t) vm_block_end != (uint32_t) heap_block_end)
	{	
		virtual_block_init (heap_block_start, (uint32_t) vm_block_start - (uint32_t) heap_block_start, 1, (void*) prev, vm_block_end);
		virtual_block_init (vm_block_end, (uint32_t) heap_block_end - (uint32_t) vm_block_end, 1, heap_block_start, (void*) next);
		
		if (prev)
			prev->next_block     = heap_block_start;
		if (next)
			next->previous_block = vm_block_end;
		
		vma_unmap (vm_block_start);
	}	
	else
	{
		panic ("Unreachable code.");
	}
}


/** 
 *
 * @param addr The address of the block.
 */
void virtual_free( const void * addr)
{
	thread_t thread = thread_get_current ();
	spinlock_lock (&(thread->heap_lock));

	/* Calculate the position of the header. */
	heap_block_t * heap_block = (heap_block_t *) (addr - sizeof (heap_block_t));

	assert (!(heap_block->is_free));

	/* Mark the block itself as free. */
	heap_block->is_free = true;

	heap_block_t* free_block = virtual_join_blocks (heap_block);
	
	virtual_try_to_physical_free (free_block);
	
	spinlock_unlock (&(thread->heap_lock));
}
