/**
 * @file vmm.c
 *
 * Simple virtual memory manager.
 *
 * Virtual memory manager provides virtual blocks allocation and virtual addresses translation. The information about 
 * every virtual block (or virtual page) for given process is stored in a virtual area table (or virtual translation 
 * table). In this context we define the process as a group of threads with the same virtual space (it means the same 
 * virtual translation table and area table). The number of virtual areas in one virtual space is limited by the constant 
 * VM_MAX_AREAS.
 *
 * To decrease memory complexity, the virtual translation table is designed to 3-level translation at the cost of the 
 * more memory accesses (higher execution speed complexity). The first 10 bits from the address is used as index to 
 * master index table to get a pointer to second index table. Then next 10 bits is used as index in the second index 
 * table to get a pointer to the physical address of page (frame in the physical memory context) for the given virtual 
 * address. The remaining 12 bits of the virtual address is joined to this physical address to get the final physical 
 * address.
 * 
 * For more information about virtual memory manager, please see the virtual memory manager introduction...
 * @link doc/doc/vmm.h Virtual memory manager introduction @endlink
 *
 * @see thread_create
 * @see wrapped_tlbrefill
 * @see frame_alloc
 * @see frame_free
 * @see vma_map
 * @see vma_unmap
 */

#include <include/error_codes.h>
#include <sync/mutex.h>
#include <mm/malloc.h>
#include <mm/physmem.h>
#include <mm/tlb.h>
#include <exc/assert.h>
#include <mm/virtmem.h>


/**
 * Initialize virtual memory space for the given @thread.
 * Allocate all the structures for virtual mapping. All the master indexes and blocks in the VM space are set to NULL. 
 *
 * @param thread The current thread.
 * @see vm_area_t
 * @see vm_seci_t
 */
void vm_init_thread (thread_t thread)
{
	/* Init the mutex for current VM space. */
	thread->vm_lock = malloc (sizeof (struct mutex));
	mutex_init (thread->vm_lock);
	
	/* Create structure for virtual mapping (master indexes). */
	thread->vm_map = malloc ((sizeof (vm_seci_t**) * VM_MAS_INDEX_SIZE));

	/* Create structure for areas (allocated blocks of memory). */
	thread->vm_areas = malloc ((sizeof (vm_area_t**) * VM_MAX_AREAS));

	/* This is the first thread shared this VM space. */
	thread->vm_share_counter = malloc (sizeof (unsigned int));
	*(thread->vm_share_counter) = 1;
	
	/* ASID record for this virtual memory space. */
	thread->asid = malloc (sizeof (unsigned int));
	/* Unassigned ASID*/
	*(thread->asid) = -1;
	
	
	/* Initialize the state of mapping. All pages is unmapped and master-level indexes are empty. */
	for (unsigned long i=0; i<VM_MAS_INDEX_SIZE; ++i)
	{
		*(thread->vm_map+i) = NULL;
	}

	/* Initialize the state of areas. All the areas are free. */
	for (int j=0; j<VM_MAX_AREAS; ++j)
	{
		*(thread->vm_areas+j) = NULL;
	}
}


/**
 * Free the virtual memory space and related structures for the given @thread.
 *
 * @param thread The current thread.
 * @see vm_area_t
 * @see vm_seci_t
 */
void vm_free_thread (thread_t thread)
{
	/* Free all the areas. There shouldn't be any allocated block. */
	for(int i=0; i<VM_MAX_AREAS; ++i)
	{
		if(*(thread->vm_areas+i) != NULL)
		{
			free ((*(thread->vm_areas+i)));
		}
	}
	/* Free the VM areas structure. */
	free (thread->vm_areas);

	/* Free all the master indexes. There shouldn't be any allocated pages. */
	for(uint32_t k=0; k<VM_MAS_INDEX_SIZE; ++k)
	{
		if (*(thread->vm_map+k) != NULL)
		{
			for(uint32_t l=0; l<VM_SEC_INDEX_SIZE; ++l)
			{
				if (*((*(thread->vm_map+k))->pages+l) != NULL)
				{
					/* Free all the allocated pages. */
					free ((*((*(thread->vm_map+k))->pages+l)));
				}
			}
			/* Free the master index. */
			free ((*(thread->vm_map+k))->pages);
			free ((*(thread->vm_map+k)));
		}
	}
	/* Free the VM table structure. */
	free (thread->vm_map);

	/* Free the VM space share counter. */
	*(thread->vm_share_counter) = 0;
	free (thread->vm_share_counter);
	
	/* Free the ASID structure. */
	free (thread->asid);
	
	/* Destroy and free the mutex for current VM space. */
	mutex_destroy (thread->vm_lock);
	free (thread->vm_lock);
}


/**
 * Returns the structure for allocated areas in VM space for given @thread.
 *
 * @param thread The thread.
 * @see vm_area_t
 */
vm_area_t** vm_get_areas (thread_t thread)
{
	return thread->vm_areas;
}


/**
 * Returns the structure for virtual mapping for given @thread. 
 * The structure contains pointer to second-level virtual table.
 *
 * @param thread The thread.
 * @see vm_seci_t
 */
vm_seci_t** vm_get_map (thread_t thread)
{
	return thread->vm_map;
}


/**
 * Return any unmapped page in virtual space for the given @thread. 
 * Linear complexity for VM_MAX_AREAS.
 *
 * @param thread The thread.
 * @param page Unmapped page (for return).
 * @return ENOMEM if there isn't any unmapped page, EOK else
 */
int vm_get_unmapped (thread_t thread, uint32_t* page)
{
	/* We go through the areas until we find at least one free address between the allocated blocks. Since 
	 * the allocated blocks are aligned on the page size, then we know that when we find a free address, then 
	 * it is free the whole page. */
	
	vm_area_t** areas = thread->vm_areas;
	uint32_t unmapped_address = VM_USEG_PREFIX;
	int index = 0;
	
	while((index < VM_MAX_AREAS) && (*(areas+index) != NULL) && ((*(areas+index))->from == unmapped_address))
	{
		unmapped_address = (*(areas+index))->from + (*(areas+index))->size;
		index++;
	}
	
	/* If the founded address is in the range of virtual memory, we know that there is a free page. */
	if (CHECK_RANGE_ADDRESS(unmapped_address, VM_USEG_PREFIX, VM_KSEG3_SUFFIX))
	{
		*page = VM_GET_PAGE_ADDRESS(unmapped_address);
		return EOK;
	}
	
	/* Overflow of uint32 - there isn't any free page. */
	return ENOMEM;
}


/**
 * Checks if the given page address is unmapped.
 *
 * @param thread The thread.
 * @param page_address The page address for checking.
 * @return Is the current page unmapped? (boolean)
 */
bool vm_is_unmapped (thread_t thread, uint32_t page_address)
{
	/* Indexes in VM map */
	uint32_t mas_i = VM_GET_MAS_INDEX_PAGE(page_address);
	uint32_t sec_i = VM_GET_SEC_INDEX_PAGE(page_address);

	/* Checks, if the master index or the current page is empty. */
	if((*(thread->vm_map+mas_i) == NULL) || (*((*(thread->vm_map+mas_i))->pages+sec_i) == NULL))
	{
		return true;
	}
	else
	{
		return false;
	}
}


/**
 * Create a virtual mapping for given block of physical memory.
 * Create a virtual mapping for block of size @size at virtual address @virtual_address and add the new allocated block 
 * to VM structure of the @thread. The virtual addresses will be mapped to @physical_address for the given @size of memory. 
 * We suppose, that all the physical addresses is already allocated.
 * 
 * @param thread The thread.
 * @param virtual_address The starting virtual address of block.
 * @param physical_address The starting physical address of block.
 * @param size The size of block.
 * @return EINVAL if the block of virtual memory isn't unmapped, ENOMEM if there are allocated maximum of blocks, EOK else
 * @see vm_add_area
 * @see vm_create_mapping
 */
int vm_virt_phys_link (thread_t thread, uint32_t virtual_address, uint32_t physical_address, size_t size)
{
	ipl_t state = query_and_disable_interrupts ();
	mutex_lock (thread->vm_lock);

	/* There are maximum number of allocated blocks. */
	if(*(thread->vm_areas+VM_MAX_AREAS-1) != NULL)
	{
		mutex_unlock (thread->vm_lock);
		conditionally_enable_interrupts (state);
		return ENOMEM;
	}

	/* The given block of virtual memory isn't unmapped. */
	if (!vm_block_is_unmapped (thread, virtual_address, size))
	{
		mutex_unlock (thread->vm_lock);
		conditionally_enable_interrupts (state);
		return EINVAL;
	}

	vm_add_area (thread, virtual_address, size);
	vm_create_mapping (thread, virtual_address, physical_address, size);

	mutex_unlock (thread->vm_lock);
	conditionally_enable_interrupts ( state);
	return EOK;
}


/**
 * Translate the given virtual address of page to physical frame (page). 
 * If the virtual page isn't mapped, returns 0 (NULL). 
 *
 * @param thread The thread.
 * @param address The virtual address.
 * @return The physical address of frame or NULL if unmapped.
 */
uint32_t vm_get_page (thread_t thread, uint32_t page_address)
{
	/* Indexes in VM table. */
	uint32_t mas_i = VM_GET_MAS_INDEX_PAGE(page_address);
	uint32_t sec_i = VM_GET_SEC_INDEX_PAGE(page_address);

	if ((*(thread->vm_map + mas_i) == NULL) || (*((*(thread->vm_map + mas_i))->pages+sec_i) == NULL))
	{
		/* There is no record in the VM table. */
		return NULL;
	}

	uint32_t page = VM_GET_PAGE_ADDRESS(*(*((*(thread->vm_map+mas_i))->pages+sec_i)));
	
	return page;
}


/**
 * Get the asid of the thread's VM space.
 *
 * @param thread The thread.
 *
 * @return The asid for the given @thread.
 */
int32_t vm_get_asid (thread_t thread)
{
	return *(thread->asid);
}


/**
 * Set the asid of the thread's VM space.
 *
 * @param thread The thread.
 * @param asid The asid.
 */
void vm_set_asid (thread_t thread, int32_t asid)
{
	*(thread->asid) = asid;
}


/**
 * Create mapping from virtual to physical memory for the block of given @size. 
 * Flags (@flags) can specify area, where should be the new block allocated:
 *
 * (flags & VF_VA_USER) - User chose the start address of block in @from;
 * (flags & VF_VA_AUTO) - The place for allocation depends on the other flags specifying the segment (VF_AT_KUSEG, 
 * VF_AT_KSEG0, VF_AT_KSEG1, VF_AT_KSSEG, VF_AT_KSEG3);
 * 
 * If the block is successfully allocated, return the address in @from. Works only for calling thread.
 * @link doc/doc/vmm.h Virtual memory manager introduction @endlink
 *
 * @param from The start address of the new allocated block.
 * @param size The size of the new block.
 * @param flags The specifying information for allocation (see vm_map comment).
 * @return EOK if the memory is successfully allocated and mapped, ENOMEM if there is not enough memory to block 
 *	allocation or if the number of allocated areas is equal to VM_MAX_AREAS (maximum), EINVAL if it is not possible 
 * 	to allocate block at the given address @from or if @from or @size are not aligned to the page size (VM_PAGE_SIZE).
 * @see tlb_clear_range
 * @see set_searching_addresses
 * @see search_avail_block
 * @see check_access
 * @see add_mapping
 * @see VM_SAFE_RANGE
 */
int vma_map (void** from, const size_t size, const vm_flags_t flags)
{
	/* Works only for the calling thread. */
	thread_t thread = thread_get_current ();

	/* Now we need to work with VM structures - must synchronize. */
	ipl_t state = query_and_disable_interrupts ();
	mutex_lock (thread->vm_lock);

	/* There is a limit for the size of allocated block. */
	if (size >= VM_MAX_SIZE_BLOCK)
	{
		mutex_unlock (thread->vm_lock);
		conditionally_enable_interrupts (state);
		return ENOMEM;
	}

	/* @size  must be aligned. */
	if (!(VM_ALIGNED (size)))
	{
		mutex_unlock (thread->vm_lock);
		conditionally_enable_interrupts (state);
		return EINVAL;
	}
	
	/* If it is given start address of block from user, it must be aligned and in the range of VM space. See 
	 * VM_SAFE_RANGE comment. */
	if (((flags & VF_VA_USER) && (!VM_ALIGNED((uint32_t)(*from)))) || ((flags & VM_SAFE_RANGE) && (flags & VF_VA_USER) && 
			(!CHECK_RANGE_ADDRESS((uint32_t)(*from), VM_USEG_PREFIX_SAFE, VM_KSEG3_SUFFIX))))
	{
		mutex_unlock (thread->vm_lock);
		conditionally_enable_interrupts (state);
		return EINVAL;
	}

	/* Start address (lower limit) for searching. */
	uint32_t ss_address;
	/* End address (upper limit) for searching. */
	uint32_t se_address;

	/* Set the bounds (limits) for searching according to the flags. */
	set_searching_addresses (&ss_address, &se_address, (uint32_t)(*from), size, flags);

	vm_area_t** areas = thread->vm_areas;

	/* Checks, if there is any free area. */
	if (*(areas+VM_MAX_AREAS-1) != NULL)
	{
		mutex_unlock (thread->vm_lock);
		conditionally_enable_interrupts (state);
		return ENOMEM;
	}

	/* Finds the first available block of the given size that matches the choosen limits (lower and upper limit). 
	 * Then checks, if it is acessible for the current mode (kernel/supervisor/user) and if it isn't at the border of 
	 * two segments. */

	uint32_t result_address = 0;
	int res_search = search_avail_block (areas, &result_address, size, ss_address, se_address);
	int res_check = NULL;
	
	/* This cycle can be done repeatly only if the searching will be started again for the modified limits. (in the 
	 * worst case 4 iterations). */
	while(res_search == EOK)
	{
		/* An available block of the size @size was found. */
		res_check = check_access (result_address, size, &ss_address, &se_address, flags);

		/* If res_check == EINVAL, it means, that the limits for searching was changed. However continuous searching 
		 * is useless if the @from address was given by the user.*/
		if ((res_check == EINVAL) && (!(flags & VF_VA_USER)))
		{
			/* Repeat the searching with new limits. */
			res_search = search_avail_block (areas, &result_address, size, ss_address, se_address);
		}
		else
		{
			/* Continuous searching is useless, either the searching and checking was successful, or there 
			 * isn't any suitable block of memory. */
			break;
		}
	}

	/* If allocating and checking was successfull. */
	if ((res_search == EOK) && (res_check == EOK))
	{
 		int res_map = add_mapping (thread, result_address, size, flags);
		if (res_map != EOK)
 		{
 			/* Allocating physical memory error. */
 			mutex_unlock (thread->vm_lock);
			conditionally_enable_interrupts (state);
			return res_map;
 		}

 		if (!(flags & VF_VA_USER))
 		{
 			*from = (void*)result_address;
 		}
	
		/* If mapping adds a paired page to one already mapped and present in TLB without clearing TLB it may 
		 * cause a TLB loop. */
		tlb_clear_range ((uint32_t)(*from), (uint32_t)(*from) + size-1, vm_get_asid (thread));	
		
 		mutex_unlock (thread->vm_lock);
		conditionally_enable_interrupts (state);
		return EOK;
	}
	else
	{
		/* Some error in searching. */
		if (flags & VF_VA_USER)
		{
			/* User specified the address where should be the block allocated, but there is not enough space. */
			mutex_unlock (thread->vm_lock);
			conditionally_enable_interrupts (state);
			return EINVAL;
		}
		else
		{
			/* There isn't any suitable block of virtual memory. */
			mutex_unlock (thread->vm_lock);
			conditionally_enable_interrupts (state);
			return ENOMEM;
		}
	}
}


/**
 * Remove mapping from virtual to physical memory.
 * Works for the block starting at address @from and the calling thread.
 * @link doc/doc/vmm.h Virtual memory manager introduction @endlink
 *
 * @param from The start address of the allocated block.
 * @return EOK if the memory is successfully unallocated and unmapped, EINVAL if the given @from address is not a starting 
 *	address of a previously mapped block or if the calling thread isn't the same thread that mapped it.
 * @see vma_unmap_block
 * @see tlb_clear_range
 */
int vma_unmap (const void* from)
{	
	/* Works only for calling thread. */
	thread_t thread = thread_get_current ();

	/* Now we need to work with VM structures - must synchronize. */
	ipl_t state = query_and_disable_interrupts ();
	mutex_lock (thread->vm_lock);
	
	vm_area_t ** areas = thread->vm_areas;
	
	/* Searching the block belongs to the given address @from. The list of areas is sorted. */
	int i = 0;
	vm_area_t * area;
	while((i<VM_MAX_AREAS) && (*(areas+i) != NULL))
	{
		/* Is the current thread from the same process as the owner? */
		area = *(areas + i);
		if (area->from == (uint32_t)from)
		{
			int res = vma_unmap_block (thread, i);
			mutex_unlock (thread->vm_lock);
			conditionally_enable_interrupts (state);
			return res;
		}
		++i;
	}

	/* There is no area with @from as a starting address. */
	mutex_unlock (thread->vm_lock);
	conditionally_enable_interrupts (state);
	return EINVAL;
}


/**
 * Find the first suitable block in the range of address [@from;@to].
 * Works only for the calling thread.
 *
 * @param from The start of the range of addresses.
 * @param to The end of the range of addresses.
 * @return Return EOK if block found or EINVAL if no such block exists.
 */
int vma_find_block (void** from, void** to)
{
	thread_t current_thread = thread_get_current (); 
	
	/* Now we need to work with VM structures - must synchronize. */
	ipl_t state = query_and_disable_interrupts ();
	mutex_lock (current_thread->vm_lock);
	
	/* Works only for calling thread. */
	vm_area_t** areas = current_thread->vm_areas;
	
	/* Find the first block in the given address range. */
	int index = 0;
	while((index < VM_MAX_AREAS) && (*(areas+index) != NULL))
	{
		if( (uint32_t)(*from) <= (*(areas+index))->from)
		{
			if ( (uint32_t)(*to) >= (*(areas+index))->from + (*(areas+index))->size)
			{
				/* We have just found our block. */
				*from = (void*)(*(areas+index))->from;
				*to = (void*)((*(areas+index))->from + (*(areas+index))->size);
				mutex_unlock (current_thread->vm_lock);
				conditionally_enable_interrupts (state);
				return EOK;	
			}
			else
			{
				/* There isn't any block in this range of addresses. */
				*from = NULL;
				*to = NULL;
				mutex_unlock (current_thread->vm_lock);
				conditionally_enable_interrupts (state);
				return EINVAL;
			}
		}	
		++index;
	}
	
	if ((index == VM_MAX_AREAS) || (*(areas+index) == NULL))
	{
		/* There isn't any block in this range of addresses. */
		*from = NULL;
		*to = NULL;
		mutex_unlock (current_thread->vm_lock);
		conditionally_enable_interrupts (state);
		return EINVAL;
	}
	
	panic ("VMM: shouldn't reach this code");
}

