/**
 * @file vmm.h
 *
 * Simple virtual memory manager.
 *
 * Virtual memory manager provides virtual blocks allocation and virtual addresses translation. The information about 
 * every virtual block (or virtual page) for given process is stored in a virtual area table (or virtual translation 
 * table). In this context we define the process as a group of threads with the same virtual space (it means the same 
 * virtual translation table and area table). The number of virtual areas in one virtual space is limited by the constant 
 * VM_MAX_AREAS.
 *
 * To decrease memory complexity, the virtual translation table is designed to 3-level translation at the cost of the 
 * more memory accesses (higher execution speed complexity). The first 10 bits from the address is used as index to 
 * master index table to get a pointer to second index table. Then next 10 bits is used as index in the second index 
 * table to get a pointer to the physical address of page (frame in the physical memory context) for the given virtual 
 * address. The remaining 12 bits of the virtual address is joined to this physical address to get the final physical 
 * address.
 * 
 * For more information about virtual memory manager, please see the virtual memory manager introduction...
 * @link doc/doc/vmm.h Virtual memory manager introduction @endlink
 *
 * @see thread_create
 * @see wrapped_tlbrefill
 * @see frame_alloc
 * @see frame_free
 * @see vma_map
 * @see vma_unmap
 */

#ifndef VMM_H_
#define VMM_H_

#include <include/c.h>
#include <mm/physmem.h>
#include <proc/thread.h>


#define VM_PAGE_BITS  		(12)
/** Page size in bytes (2^12). */
#define VM_PAGE_SIZE  		(1 << VM_PAGE_BITS)

/** Macro checks, if the given address @x is aligned to page size.
 * @param x The address for checking.
 */
#define VM_ALIGNED(x) 		(!((x) % (VM_PAGE_SIZE)))

#define VM_MAS_INDEX_BITS  	(10)
/** Maximum size of master index level paging (2^10). */
#define VM_MAS_INDEX_SIZE  	(1 << VM_MAS_INDEX_BITS)


#define VM_SEC_INDEX_BITS  	(10)
/** Maximum size of second index level paging (2^10). */
#define VM_SEC_INDEX_SIZE  	(1 << VM_SEC_INDEX_BITS)
 
/** The maximum number of used areas. */
#define VM_MAX_AREAS 		(100)

/** Enumeration of modes. */
typedef enum 
{
	USER,
	SUPERVISOR,
	KERNEL,
	MODE_UNK
} vm_mode;


/** 
 * Virtual memory - second level mapping index. 
 */
typedef struct vm_seci
{
	/** Array of mapped pages. */
	uint32_t** pages;
} vm_seci_t;


/** 
 * Virtual memory area.
 * Virtual memory area is an unbroken block of mapped virtual memory.  
 */
typedef struct vm_area
{
	/** Size of the area. */
	size_t size;
	
	/** Start address of the area. */
	uint32_t from;
	
	/** Identifier of the thread. */
	uint32_t thread_id;
} vm_area_t;


/** 
 * Return the VM master index of the given page address.
 * @param page_address The address of page.
 * @return The VM master index of the given address of page.
 */
#define VM_GET_MAS_INDEX_PAGE(page_address) ((page_address) >> VM_SEC_INDEX_BITS)

/** 
 * Return the VM second index of the given page address.
 * @param page_address The address of page.
 * @return The VM msecond index of the given page address.
 */
#define VM_GET_SEC_INDEX_PAGE(page_address) (((page_address) << (VM_MAS_INDEX_BITS + VM_PAGE_BITS)) \
	>> (VM_MAS_INDEX_BITS + VM_PAGE_BITS))

/** 
 * Return the VM master index of the given address.
 * @param address The address.
 * @return The VM master index of the given address.
 */
#define VM_GET_MAS_INDEX_ADDRESS(address) ((address) >> (VM_SEC_INDEX_BITS + VM_PAGE_BITS))

/** 
 * Return the VM second index of the given address.
 * @param address The address.
 * @return The VM second index of the given address.
 */
#define VM_GET_SEC_INDEX_ADDRESS(address) (((address) << VM_MAS_INDEX_BITS) \
	>> (VM_MAS_INDEX_BITS + VM_PAGE_BITS))

/** 
 * Return the page address of the given address.
 * @param address The address.
 * @return The page address of the given address.
 */
#define VM_GET_PAGE_ADDRESS(address) ((address) >> VM_PAGE_BITS)

/** The maximum size of virtual memory block. */
#define VM_MAX_SIZE_BLOCK   		(0x80000000)

/* ADDRESS RANGES OF SEGMENTS */
 
/* USEG == SUSEG == KUSEG */

/** Address prefix of USEG segment. */
#define VM_USEG_PREFIX   		(0x00000000)

/**
 * Virtual memory safe range notation.
 * For virtual malloc the first virtual page is skipped, then the virtual malloc cannot return the address 0x00000000 
 * and the evaluation, e.g. [if (address)], is correct. Virtual malloc calls the vma_map with the spec. flag 
 * VM_SAFE_RANGE in order to use this 'safe' range without the first page.
 * @see VM_SAFE_RANGE
 * @see vma_map
 */
#define VM_USEG_PREFIX_SAFE   		(0x00001000)

/** Address suffix of USEG segment. */
#define VM_USEG_SUFFIX   		(0x7FFFFFFF)

/* SSEG == KSSEG == KSEG2 */

/** Address prefix of SSEG segment. */
#define VM_SSEG_PREFIX 			(0xC0000000)

/** Address suffix of SSEG segment. */
#define VM_SSEG_SUFFIX  		(0xDFFFFFFF)

/* KSEG0, KSEG1 and KSEG3 */

/** Address prefix of KSEG0 segment. */
#define VM_KSEG0_PREFIX  		(0x80000000)

/** Address suffix of KSEG0 segment. */
#define VM_KSEG0_SUFFIX  		(0x9FFFFFFF)

/** Address prefix of KSEG1 segment. */
#define VM_KSEG1_PREFIX  		(0xA0000000)

/** Address suffix of KSEG1 segment. */
#define VM_KSEG1_SUFFIX  		(0xBFFFFFFF)

/** Address prefix of KSEG3 segment. */
#define VM_KSEG3_PREFIX  		(0xE0000000)

/** Address suffix of KSEG3 segment. */
#define VM_KSEG3_SUFFIX  		(0xFFFFFFFF)


/**
  * Check if the given address is in the given range (prefix, suffix).
  *
  * @param x The address.
  * @param prefix The prefix of block of memory.
  * @param suffix The suffix of block of memory.
  * @return True if the address is in the given address range, false else.
  */
#define CHECK_RANGE_ADDRESS(x, prefix, suffix) ( \
	(((prefix) < (x)) || ((prefix) == (x))) && \
	(((x) < (suffix)) || ((suffix) == (x))))


/** Enumeration of segments. */
typedef enum {
	USEG,
	KSEG0,
	KSEG1,
	SSEG,
	KSEG3,
	SEG_UNK
} vm_segment;


/**
  * Check the segment of the given address.
  *
  * @return The suitable segment for the address.
  */
static inline vm_segment get_vm_segment (unsigned long add)
{
	if (CHECK_RANGE_ADDRESS(add, VM_USEG_PREFIX, VM_USEG_SUFFIX))
	{
		return USEG;
	}
	else if (CHECK_RANGE_ADDRESS(add, VM_KSEG0_PREFIX, VM_KSEG0_SUFFIX))
	{
		return KSEG0;
	}
	else if (CHECK_RANGE_ADDRESS(add, VM_KSEG1_PREFIX, VM_KSEG1_SUFFIX))
	{
		return KSEG1;
	}
	else if (CHECK_RANGE_ADDRESS(add, VM_SSEG_PREFIX, VM_SSEG_SUFFIX))
	{
		return SSEG;
	}
	else if (CHECK_RANGE_ADDRESS(add, VM_KSEG3_PREFIX, VM_KSEG3_SUFFIX))
	{
		return KSEG3;
	}
	else
	{
		return SEG_UNK;
	}
}


/**
  * Returns the segment prefix (the starting address) for the given segment.
  *
  * @param segment The given segment.
  * @return The segment prefix.
  */
static inline uint32_t get_seg_prefix (vm_segment segment)
{
	switch (segment)
	{
		case USEG:
			return VM_USEG_PREFIX;
		case KSEG0:
			return VM_KSEG0_PREFIX;
		case KSEG1:
			return VM_KSEG1_PREFIX;
		case SSEG:
			return VM_SSEG_PREFIX;
		case KSEG3:
			return VM_KSEG3_PREFIX;
		default:
			return NULL;
	}
}


/**
 * Checks the current mode.
 * Kernel-Supervisor-User (KSU - Status register).
 *	00 for kernel mode;
 *	01 for supervisor mode;
 *	10 for user mode;
 * If EXL bit or ERL bit --> Kernel mode.
 *
 * @return The current mode.
 */
static inline vm_mode get_vm_mode (void)
{
	unative_t status = read_cp0_status ();
	if ((!(CP0_STATUS_EXL(status))) && (!(CP0_STATUS_ERL(status))))
	{
		if (CP0_STATUS_KSU1(status))
		{
			if (CP0_STATUS_KSU2(status))
			{
				return MODE_UNK;
			}
			else
			{
				return SUPERVISOR;
			}
		}
		else
		{
			if (CP0_STATUS_KSU2(status))
			{
				return USER;
			}
			else
			{
				return KERNEL;
			}
		}
	}
	else
	{
		return KERNEL;
	}
}


/* Commented with implementation */

void 		vm_init_thread		(thread_t thread);
void 		vm_free_thread		(thread_t thread);
vm_area_t**	vm_get_areas		(thread_t thread);
vm_seci_t**	vm_get_map		(thread_t thread);
int32_t		vm_get_asid		(thread_t thread);
void		vm_set_asid		(thread_t thread, int32_t asid);
int 		vm_get_unmapped		(thread_t thread, uint32_t* page);
bool 		vm_is_unmapped		(thread_t thread, uint32_t page);
uint32_t 	vm_get_page		(thread_t thread, uint32_t vm_add);
int 		vm_virt_phys_link	(thread_t thread, uint32_t virtual_address, uint32_t physical_address, size_t size);
int		vma_find_block		(void** from, void** to);

int	 	vma_map			(void** from, const size_t size, const vm_flags_t flags);
int 		vma_unmap		(const void* from); 

#endif
