/**
 * @file thread.c
 *
 * Thread management.
 *
 */

#include <adt/list.h>
#include <adt/hashmap.h>
#include <drivers/dorder.h>
#include <exc/assert.h>
#include <exc/syscall.h>
#include <include/c.h>
#include <include/error_codes.h>
#include <include/shared.h>
#include <mm/malloc.h>
#include <mm/tlb.h>
#include <mm/vmalloc.h>
#include <mm/vmm.h>
#include <sync/mutex.h>
#include <sched/sched.h>


#include <proc/thread.h>


/** Currently running threads on each CPU */
thread_t current_thread[MAX_CPU];

/** Hashmap - stores initialized user mutexes. */
map_t u_threads = NULL;

/** Internal mutex for synchronized access to hashmap. */
struct mutex* hashmap_u_threads_mutex = NULL;

/** Synchronize picking CPUs for thread creating;
 */
struct spinlock cpu_pick_lock;

/** CPU to start next thread on. 
 */
int32_t cpu_pick = 0;


/** Initialize threads management
 *
 * Set current threads on all CPUs
 * to be NULL.
 *
 */
void init_threads (void)
{
	spinlock_init (&cpu_pick_lock);

	/* Initialize the state of current threads. */
	unsigned int i;
	for (i = 0; i < MAX_CPU; i++)
		current_thread[i] = NULL;
}

/** 
 * Initialize the structure for user threads. 
 */
void init_u_threads (void)
{
	u_threads = hashmap_new ();
	hashmap_u_threads_mutex = malloc (sizeof (struct mutex));
	mutex_init (hashmap_u_threads_mutex);
}


/** frees given thread, deallocate everything etc.
 *
 * @param thr target thread for free
 *
 */
void thread_free (thread_t * thr);
void thread_free (thread_t * thr)
{
	disabled (*thr);
	spinlock_destroy (&((*thr)->thread_lock));
	free ((*thr)->stack_block);
	
	/* The last thread uses this VM space? */
	if (*((*thr)->vm_share_counter) == 1)
	{
		vm_free_thread (*thr);
	}
	else
	{
		(*((*thr)->vm_share_counter))--;
	}
	
	/* If it is user thread, remove from hashmap. */
	hashmap_remove (u_threads, (unsigned long) *thr);
	
	free (*thr);
	(*thr) = NULL;
}


/** get current thread identification
 *
 * @return Returns current executing thread.
 *
 */
thread_t thread_get_current (void)
{
	ipl_t status = query_and_disable_interrupts ();
	thread_t result = current_thread[cpuid ()];
	conditionally_enable_interrupts (status);
	return result;
}


/** finishing thread according to attached/detached etc.
 *
 * @thread thread to finish
 *
 */
void thread_finish_over (thread_t thread);
void thread_finish_over (thread_t thread)
{
	spinlock_lock (&(thread->thread_lock));
	ipl_t status = query_and_disable_interrupts ();
	/* if thread is attached, joining thread will free it */
	if (thread->is_attached)
	{
		thread->is_zombie = 1;
		if (NULL != thread->superior_ancestor)
		{
			runnable (thread->superior_ancestor);
		}
		spinlock_unlock (&(thread->thread_lock));
	}
	else
	{
		spinlock_unlock (&(thread->thread_lock));
		thread_free (&thread);
	}
	scheduled ();
	conditionally_enable_interrupts (status);
	/* a thread finish its life in endless loop */
	while (1);
}


/** envelope for real thread function; wrapper function
 *
 * @param thread owner of this function
 *
 * @return returns the same as inner function
 *
 */
void envelope_function (thread_t thread);
void envelope_function (thread_t thread)
{
	/* alloc user stack if this is userland thread */
	if (TF_USERLAND & thread->flags)
	{
		if (TF_USERINIT & thread->flags)
		{
			if (EOK != vm_virt_phys_link(thread, 0, USER_BIN_START, USER_BIN_LEN))
			{
				panic ("User thread init fault!");
			}
		}

		thread->user_stack_block = virtual_malloc (USER_THREAD_STACK_SIZE);
		thread->user_stack_top   = ((char *) (thread->user_stack_block)) - sizeof (context_t) + USER_THREAD_STACK_SIZE;

		context_t * context = (context_t *) (thread->user_stack_top);
        	context->ra = (unative_t) 0x00000000;
        	context->gp = ADDR_IN_KSEG0 (0);
	        context->status = CP0_STATUS_IE_MASK | CP0_STATUS_IM_MASK | CP0_STATUS_KSU2_MASK;

		ipl_t status = query_and_disable_interrupts (); 
		thread->is_in_kernel_mode = 0;
		switch_cpu_context (&(thread->stack_top     ),
		                    &(thread->user_stack_top),
				    read_cp0_entryhi());
		conditionally_enable_interrupts (status);

		syscall_function ();
	}

	/* kernel threads have no userland stack */
	thread->thread_function (thread->data);
	thread_finish_over (thread);
}


/** creates, allocates and runs new thread
 *
 * @param thread_ptr pointer to a new thread (inserted pointer is empty, function allocates it)
 * @param thread_start pointer to a function that thread should execute
 * @param data pointer to parameter of thread_start function
 * @param flags flags for new thread
 *
 * @return returns EOK or other error code
 *
 */
int thread_create (thread_t * thread_ptr, void * (*thread_start)(void *), void * data, const thread_flags_t flags)
{
	spinlock_lock( &cpu_pick_lock);
	
	uint32_t cpu = cpu_pick;
	cpu_pick++;
	if (cpu_pick >= atomic_get(&cpu_ready))
		cpu_pick = 0;
	
	spinlock_unlock (&cpu_pick_lock);

	int ret = thread_create_on_cpu (thread_ptr, thread_start, data, flags, cpu);

	return ret;
}
int thread_create_on_cpu (thread_t * thread_ptr, void * (*thread_start)(void *), void * data, const thread_flags_t flags, uint32_t cpu)
{
	
	*thread_ptr = malloc (sizeof(thread_struct_t));
        if (!(*thread_ptr))
        {
        	return ENOMEM;
        }
	
	/* scheduler takes strict handling of these structures */
	(*thread_ptr)->cpu = cpu;
	(*thread_ptr)->queued_on_cpu = false;

	/* every thread starts kernel way; user threads will switch after creating user stack */
	(*thread_ptr)->is_in_kernel_mode = 1;

	/* flags are holden for future use */
	(*thread_ptr)->flags = flags;

        /* Initialize the list membership structure. */
        item_init (&((*thread_ptr)->item));
        
        /* Flag TF_NEW_VMM - allocate new virtual memory space. */
        if ((flags & TF_NEW_VMM) || (thread_get_current () == NULL))
	{
		/* Init VM space. */
		vm_init_thread ((*thread_ptr));
	}
	else
	{
		/* Share the VM space with the calling thread (parent). */
		thread_t parent = thread_get_current ();
		(*thread_ptr)->vm_map = parent->vm_map;
		(*thread_ptr)->vm_areas = parent->vm_areas;
		(*thread_ptr)->vm_lock = parent->vm_lock;
		(*thread_ptr)->vm_share_counter = parent->vm_share_counter;
		(*((*thread_ptr)->vm_share_counter))++;
		(*thread_ptr)->asid = parent->asid;
	}


        (*thread_ptr)->stack_block = safe_malloc (THREAD_STACK_SIZE);

        /* Stack grows downwards inside the allocated block. Its top
           contains the thread context structure, which is restored
           on context switch. Of the entire structure, only the
           starting address of the thread and some vital
           processor registers are set. */
        (*thread_ptr)->stack_top = ((char *) ((*thread_ptr)->stack_block)) - sizeof(context_t) + THREAD_STACK_SIZE;

	(*thread_ptr)->user_stack_block = 0;
	(*thread_ptr)->user_stack_top   = 0;

        /* saves thread function */
        (*thread_ptr)->thread_function = thread_start;
	
        (*thread_ptr)->heap_start = NULL;

        spinlock_init (&((*thread_ptr)->heap_lock));
        
	/* every thread is by default stoppable */
        (*thread_ptr)->is_unstoppable = false;

        /* saves thread function data */
        (*thread_ptr)->data = data;

        /* thread doesnt start as zombie */
        (*thread_ptr)->is_zombie = 0;

        /* a thread is basicaly count as attached by null thread */
        (*thread_ptr)->is_attached = 1;

        /* thread is not owned by anyone when created */
        (*thread_ptr)->superior_ancestor = NULL;

        /* joining is basicaly EOK */
        (*thread_ptr)->join_result = EOK;

        /* return value for thread function is unknown - 0 */
        (*thread_ptr)->return_value = 0;
        
        /* thread start NOT suspended */
        (*thread_ptr)->is_suspended = 0;

        /* sleeping counter starts as zero */
        (*thread_ptr)->sleeping_counter = 0;

        /* init thread lock */
        spinlock_init (&((*thread_ptr)->thread_lock));

	/* initializing context for scheduler */
        context_t * context = (context_t *) ((*thread_ptr)->stack_top);
       	context->ra = (unative_t) envelope_function;
        context->a0 = (unative_t) *thread_ptr;
       	context->gp = ADDR_IN_KSEG0 (0);
        context->status = CP0_STATUS_IE_MASK | CP0_STATUS_IM_MASK | CP0_STATUS_CU0_MASK;

        runnable (*thread_ptr);
        
	return EOK;
}


/** function for joining timer routine
 *
 * @param tmr calling timer (quite useless)
 * @param data thread for join
 *
 */
void joining_timer_function (struct timer * tmr, void * data);
void joining_timer_function (struct timer * tmr, void * data)
{
	thread_t target_thread           = (thread_t) data;
	thread_t joining_thread          = target_thread->superior_ancestor;
	target_thread->superior_ancestor = NULL;
	runnable (joining_thread);
}


/** this thread waits for given thread
 *
 * @param thr thread for waiting for
 * @param usec max waiting time in microseconds
 * @param timed true = version with time limit; false = version without timer
 *
 * @return returns EINVAL for not valid thread etc., EKILLED for killed thread or EOK for thread done
 *
 */
int thread_join_internal (thread_t thr, const unsigned int usec, bool timed);
int thread_join_internal (thread_t thr, const unsigned int usec, bool timed)
{
	struct timer tmr;
	thread_t this_thread = thread_get_current ();
	int result;
	
	/* we need to test if it is correct operation to join the thread */
	spinlock_lock (&(thr->thread_lock));
	if (NULL == thr)
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}
	if (!(thr->is_attached))
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}
	if (NULL != thr->superior_ancestor)
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}
	if (this_thread == thr)
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}

	ipl_t state = query_and_disable_interrupts ();
	if ((timed) && (0 < usec))
	{
		timer_init (&tmr, usec, joining_timer_function, (void *)thr);
		timer_start (&tmr);
	}

	/* if joining thread is still running, this thread blocks itself, otherwise not */
	if ((!(thr->is_zombie)) && ((!timed) || (0 < usec)))
	{
		thr->superior_ancestor = this_thread;
		spinlock_unlock (&(thr->thread_lock));
		disabled (this_thread);
		scheduled ();
		spinlock_lock (&(thr->thread_lock));
		result = thr->join_result;
	}
	else
	{
		if (thr->is_zombie)
		{
			result = thr->join_result;
		}
		else
		{
			result = ETIMEDOUT;
		}
	}

	if ((timed) && (0 < usec))
	{
		timer_destroy (&tmr);
	}
	conditionally_enable_interrupts (state);
	spinlock_unlock (&(thr->thread_lock));
	if (ETIMEDOUT != result)
	{
		thread_free (&thr);
	}
	return result;
}


/** this thread waits for given thread
 *
 * @param thr thread for waiting for
 *
 * @return returns EINVAL for not valid thread etc., EKILLED for killed thread or EOK for thread done
 *
 */
int thread_join (thread_t thr)
{
	return thread_join_internal (thr, 0, 0);
}


/** this thread waits for given thread
 *
 * @param thr thread for waiting for
 * @param usec max waiting time in microseconds
 *
 * @return returns EINVAL for not valid thread etc., EKILLED for killed thread or EOK for thread done
 *
 */
int thread_join_timeout (thread_t thr, const unsigned int usec)
{
	return thread_join_internal (thr, usec, 1);
}


/** this thread detach given thread
 *
 * @param thr thread for detaching
 *
 * @return returns EINVAL for not valid thread etc., EOK otherwise
 *
 */
int thread_detach (thread_t thr)
{
	spinlock_lock (&(thr->thread_lock));

	if (NULL == thr)
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}
	if (!(thr->is_attached))
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}
	if (NULL != thr->superior_ancestor)
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}
	if (thr->is_zombie)
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}

	thr->is_attached = 0;
	spinlock_unlock (&(thr->thread_lock));
	return EOK;
}


/** internal function for wake up thread or reset timer
 *
 * @param tmr waking timer
 * @param data sleeping thread
 *
 */
void conditionally_wake_up (struct timer * tmr, void * data);
void conditionally_wake_up (struct timer * tmr, void * data)
{
	thread_t thread = (thread_t) data;
	thread->sleeping_counter--;
	/* sleeping_counter watches how many times the timer has been reset */
	if (0 < thread->sleeping_counter)
	{
		/* in this case timer is reset and the thread remains sleeping */
		timer_start (tmr);
	}
	else
	{
		/* in this case a thread is awaken */
		if (EKILLED != thread->join_result)
		{
			runnable (thread);
		}
	}
}


/** sleeps a thread for some time
 *
 * @param sec time in seconds
 * @param usec time in microseconds
 *
 */
void thread_sleep_internal (const unsigned int sec, const unsigned int usec);
void thread_sleep_internal (const unsigned int sec, const unsigned int usec)
{
	struct timer tmr;
	thread_t this_thread = thread_get_current ();

	/* when a time is less than 1 second, only one timer cycle is needed,
	 * otherwise the timer is reset every second */
	if (0 < sec)
	{
		timer_init (&tmr, 1000000, conditionally_wake_up, this_thread);
		this_thread->sleeping_counter = sec;
	}
	else
	{
		timer_init (&tmr, usec, conditionally_wake_up, this_thread);
		this_thread->sleeping_counter = 1;
	}

	ipl_t state = query_and_disable_interrupts ();
	disabled (this_thread);
	timer_start (&tmr);
	conditionally_enable_interrupts (state); 
	scheduled ();
	timer_destroy (&tmr);
}


/** thread sleeps itself for some time
 *
 * @param sec time in seconds
 *
 */
extern void thread_sleep (const unsigned int sec)
{
	if (0 < sec)
	{
		thread_sleep_internal (sec, 0);
	}
}


/** thread sleeps itself for some time
 *
 * @param usec time in microseconds
 *
 */
extern void thread_usleep (const unsigned int usec)
{
	if (0 < usec)
	{
		thread_sleep_internal (0, usec);
	}
}


/** schedule out target thread */
void thread_yield (void)
{
	scheduled ();
}


/** thread sleeps itself */
void thread_suspend (void)
{
	thread_t this_thread = thread_get_current ();
	this_thread->is_suspended = 1;
	disabled (this_thread);
	scheduled ();
}


/** thread wakes up some suspended thread
 *
 * @param thr thread to wake up
 *
 */
int thread_wakeup (thread_t thr)
{
	if (NULL == thr)
	{
		return EINVAL;
	}

	spinlock_lock (&(thr->thread_lock));
	if ((thr->is_suspended) && (EKILLED != thr->join_result))
	{
		thr->is_suspended = 0;
		runnable (thr);
	}
	spinlock_unlock (&(thr->thread_lock));

	return EOK;
}


/** kills target thread
 *
 * @param thr thread to kill
 *
 * @return returns EINVAL for invalid thread or EOK otherwise
 *
 */
int thread_kill (thread_t thr)
{
	spinlock_lock (&(thr->thread_lock));
	if (NULL == thr)
	{
		spinlock_unlock (&(thr->thread_lock));
		return EINVAL;
	}

	disabled (thr);
	/* it is not allowed to kill threads waiting on synchronization primitives */
	if (NULL != thr->item.list)
	{
		panic ("killing thread in crit queue");
	}
	thr->join_result = EKILLED;
	thr->is_zombie = 1;
	if (thr->is_attached)
	{
		if (NULL != thr->superior_ancestor)
		{
			runnable (thr->superior_ancestor);
		}
		spinlock_unlock (&(thr->thread_lock));
	}
	else
	{
		spinlock_unlock (&(thr->thread_lock));
		thread_free (&thr);
	}
	return EOK;
}


/** Start executing given thread
 *
 * The function saves the context of the current thread and
 * restores the context of the thread given as argument.
 *
 * @param thread Thread to start executing.
 *
 */
void switch_to_thread (thread_t thread)
{
	/* The only thing to do besides switching the processor context
	   is to maintain the pointer to the current thread. We disable
	   interrupts to do that so that the pointer to the current
	   thread is always in sync with the current thread. */
	
	ipl_t state = query_and_disable_interrupts ();
	
	thread_t old_thread;
	old_thread = current_thread[cpuid ()];
	current_thread[cpuid ()] = thread;

	/* One special case to consider here is when switching context
	   for the very first time. At that time, we are running without
	   a thread structure and with a temporary stack, and therefore
	   we do not care about the current stack top that gets saved. */

	void *  dummy_stack_top;
	void ** old_stack;
	void ** new_stack;

	/* userland threads has two stacks so we need to switch to the right stack */

	if (NULL == old_thread)
	{
		old_stack = &dummy_stack_top;
	}
	else
	{
		if (old_thread->is_in_kernel_mode)
		{
			old_stack = &(old_thread->stack_top);
		}
		else
		{
			old_stack = &(old_thread->user_stack_top);
		}
	}

	if (thread->is_in_kernel_mode)
	{
		new_stack = &(thread->stack_top);
	}
	else
	{
		new_stack = &(thread->user_stack_top);
	}

	uint32_t asid = get_virtual_space (thread);
	switch_cpu_context (old_stack, new_stack, asid);

	/* This is subtle. The switch_cpu_context function will execute
	   another thread, which might switch to yet another thread,
	   and so on, eventually switching back to this thread.
	   At that time, it will look like we have just
	   returned here from the switch_cpu_context
	   function. */
	conditionally_enable_interrupts (state);
}


/**
 * Creates a new user thread. 
 * User threads are stored in the hashmap, so that for the threads operation we can control, if the given thread was 
 * successfully created as a user thread and its authorization.
 * @link kernel/adt/hashmap.h Hashmap implementation @endlink
 * 
 * @param thr_ptr Pointer to kernel address, where will be stored the kernel thread related to the new user thread.
 * @param thread_start User thread function.
 * @param data The data for user thread function.
 * @param flags The specifying information for thread creation (see thread_create comment).
 * @return ENOMEM if there isn't enough memory for new user thread, EOK else
 * @see thread_create
 * @see u_threads
 */
int u_thread_create (void** thr_ptr, void* (*thread_start)(void*), void* data, const thread_flags_t flags)
{
	
	/* We store all the user threads in hashmap. There is an unsigned long integer as a key (address of kernel 
	 * thread related to the user thread) and an unsigned long integer as a record (address of owner thread). */

	thread_t* thread_ptr = (thread_t*) thr_ptr; 	
	int result = thread_create (thread_ptr, thread_start, data, flags);
	if (result != EOK)
	{
		/* Not enough memory for new thread. */
		return result;
	}

	/* Add the thread creation record to the hashmap. */
	uint32_t thread_owner = (uint32_t) thread_get_current ();
	mutex_lock (hashmap_u_threads_mutex);
	if (hashmap_put (u_threads, (unsigned long) *thread_ptr, (void*) thread_owner) != MAP_OK)
	{
		/* Not enough memory. */
		mutex_unlock (hashmap_u_threads_mutex);
		return ENOMEM;
	}
	
	/* Thread creation successful. */
	mutex_unlock (hashmap_u_threads_mutex);
	return EOK;
}


/**
 * Checks, if the given @kernel_thread was created as an user thread. 
 * Every user thread is stored in hashmap, so we can check, if the given @kernel_thread is related to some existing 
 * user thread. Only user threads from the same VM space can handle with the other user threads.
 * @link kernel/adt/hashmap.h Hashmap implementation @endlink
 *
 * @param kernel_thread The thread.
 * @return EINVAL if there isn't any record for the given @kernel_thread in the hashmap or if the current thread 
 *	can't handle with this thread, EOK else 
 * @see u_threads
 */
int u_thread_valid (void* kernel_thread)
{
	/* Is the @kernel_thread a correct pointer to kernel thread related to user thread? */
	uint32_t thread_owner = NULL;
	mutex_lock (hashmap_u_threads_mutex);
	if ((hashmap_get (u_threads, (unsigned long) kernel_thread, (void**) &thread_owner) != MAP_OK)
		|| ((thread_t)thread_owner)->vm_map != (thread_get_current ())->vm_map)
	{
		/* There is no record for this thread or the current thread can't handle with the thread. */
		mutex_unlock (hashmap_u_threads_mutex);
		return EINVAL;
	}
	
	mutex_unlock (hashmap_u_threads_mutex);
	return EOK;
}


/**
 * Kills the current process.
 * All the threads sharing the VM space with the given @thread will be killed.
 * 
 * @param thread The thread.
 * @see thread_kill
 * @see thread_finish_over
 */
void u_process_kill (thread_t thread)
{
	vm_area_t** areas = thread->vm_areas;
	
	int index = 0;
	while (*(areas+index) != NULL)
	{
		/* Kills all the threads axcept the given @thread. */
		if (thread != (thread_t)(*(areas+index))->thread_id)
		{
			thread_t killing_thr = (thread_t)(*(areas+index))->thread_id; 
			thread_kill (killing_thr);
		}
		index++;
	}
	thread_finish_over (thread);
}


/**
 * Kills the current process.
 * If there isn't any other user process, then stop the core execution.
 *
 * @see u_process_kill
 */
void u_exit (void)
{
	u_process_kill (thread_get_current ());
	
	/* If there isn't any other user process, then stop the core execution. */
	if (!(hashmap_length (u_threads)))
	{
		msim_halt ();
	}
	
	scheduled ();
	return;
}
