/**
 * @file thread.c
 *
 * Thread management.
 *
 */


#ifndef THREAD_H_
#define THREAD_H_

#include <include/shared.h>
#include <include/c.h>
#include <adt/list.h>
#include <sync/spinlock.h>

typedef uint32_t thread_flags_t;

#define TF_NEW_VMM  (1 << 0)
#define TF_USERLAND (1 << 1)
#define TF_USERINIT (1 << 2)

/* Forward declaration */
struct vm_seci;
struct vm_area;


/** Thread stack size
 *
 * The size of the thread stack. This should be set liberally,
 * since stack overflow typically has obscure syndroms and
 * is notoriously difficult to debug.
 *
 */
#define THREAD_STACK_SIZE      4096
#define USER_THREAD_STACK_SIZE 4096

struct thread_struct;


/** Thread control structure
 *
 * The structure represents a kernel thread. Since the thread context
 * is kept on the thread stack, the only thing that the structure
 * contains is a reference to the stack. The structure can also
 * be an item on a list.
 *
 */
typedef struct thread_struct {
	/** A thread can be an item on a list */
	item_t item;

	/** CPU */
	uint32_t cpu;
	bool queued_on_cpu;

	/** TLB asid identifier */
	int32_t* asid;

	/** Virtual heap start */
	void* heap_start;
	
	/** Lock for virtual heap operations. */
	struct spinlock heap_lock;

	/** true = thread cannot be disabled */
	bool is_unstoppable;

	/** flag from thread create function */
	thread_flags_t flags;
	/** true = kernel mode, false = user mode */
	bool is_in_kernel_mode;
	/** pointer to context where syscall can read arguments and write there whatever */
	context_t * registers;
	
	/** pointer to stack memory */
	uint8_t * stack_block;
	/** Stack top is where the thread context is */
	void * stack_top;

	/** pointer to user stack memory */
	uint8_t * user_stack_block;
	/** user stack top is where the user thread context is */
	void * user_stack_top;

	/** thread executing function */
	void * (* thread_function) (void *);
	/** data for executing function */
	void * data;

	/** is zombie, i.e. not detached && not joined */
	bool is_zombie;
	/** true = attached; false = detached */
	bool is_attached;
	/** a thread waiting for join */
	struct thread_struct * superior_ancestor;
	/** return value for joining function */
	int join_result;
	
	/** return value for thread function */
	int return_value;

	/** true = suspended; false = enabled */
	bool is_suspended;
	
	/** Which blocks are allocated in the virtual memory space, where belongs 
	    the current thread. */
	struct vm_area** vm_areas;
	
	/** How many threads share the virtual memory space. */
	unsigned int* vm_share_counter;
	
	/** Virtual memory translation table. */
	struct vm_seci** vm_map;
	
	/** Virtual memory space mutex. */
	struct mutex* vm_lock;
	
	/** sleeping counter for awaking timers */
	unsigned int sleeping_counter;

	/** lock for thread operations */
	struct spinlock thread_lock;
} thread_struct_t;


typedef thread_struct_t * thread_t;


/** Currently running threads on each CPU */
extern thread_t current_thread[MAX_CPU];


/* Externals are commented with implementation */
extern void     init_threads (void);
extern void 	init_u_threads (void);

extern int      thread_create       (thread_t * thread_ptr, void * (*thread_start)(void *), void * data, const thread_flags_t flags);
extern int      thread_create_on_cpu(thread_t * thread_ptr, void * (*thread_start)(void *), void * data, const thread_flags_t flags, uint32_t cpu);
extern thread_t thread_get_current  (void);
extern int      thread_join         (thread_t thr);
extern int      thread_join_timeout (thread_t thr, const unsigned int usec);
extern int      thread_detach       (thread_t thr);
extern void     thread_sleep        (const unsigned int sec);
extern void     thread_usleep       (const unsigned int usec);
extern void     thread_yield        (void);
extern void     thread_suspend      (void);
extern int      thread_wakeup       (thread_t thr);
extern int      thread_kill         (thread_t thr);
extern void 	thread_finish_over  (thread_t thread);

extern int 	u_thread_create     (void** thread_ptr, void* (*thread_start)(void*), void* data, const thread_flags_t flags);
extern int 	u_thread_valid      (void* kernel_thread);
extern void 	u_process_kill      (thread_t thread);
extern void 	u_exit              (void);

extern void     switch_to_thread (thread_t thread);


#endif
