/**
 * @file sched.c
 *
 * Round-robin multiprocessor kernel thread scheduler.
 *
 */

#include <sched/sched.h>

#include <adt/list.h>
#include <drivers/dorder.h>
#include <drivers/printer.h>
#include <drivers/timer.h>
#include <exc/assert.h>
#include <include/c.h>
#include <include/shared.h>
#include <mm/malloc.h>
#include <mm/tlb.h>
#include <proc/thread.h>
#include <sync/mutex.h>


/**
 * Synchronize the startup of CPUs.
 */
ATOMIC_DECLARE (cpu_ready, 0);

/** 
 * Random seed for the scheduler. 
 */
static unsigned int seed = 0;

/** 
 * Minimum number of ticks a thread is allowed to run. 
 */
#define THREAD_QUANTUM_MIN  10000 

/** 
 * Maximum number of ticks a thread is allowed to run.
 */
#define THREAD_QUANTUM_MAX  14000

/** 
 * Actual number of ticks is randomized. 
 */
#define THREAD_QUANTUM  (THREAD_QUANTUM_MIN + random (&seed) % (THREAD_QUANTUM_MAX - THREAD_QUANTUM_MIN))


/**  
 * This thread serves two purposes.
 * It's a timer worker thread it runs timer handlers.
 * And it remain always in the scheduler queue so there's
 * always a thread to run. 
 */
static void* timer_worker (void * data)
{
	struct timer* tmr = NULL;

	while (true)
	{
		while ((tmr = timer_ready ()) != NULL)
		{
			tmr->handler (tmr, tmr->data);
			if (tmr != NULL && tmr->free_me)
				free (tmr);
		}
		thread_yield ();
	}

	return 0;
}


/** 
 * List of schedulable threads for each CPU. 
 *
 * @see list_t
 */
list_t sched_threads [MAX_CPU];

/** 
 * Timer worker thread for each CPU. 
 */
thread_t sched_timer [MAX_CPU];

/** 
 * Big lock for the kernel scheduler shared by by all CPUs. 
 * We wanted to implement scheduler with per CPU locking or 
 * nonblocking lists but didn't have enough time for testing.
 */
struct spinlock sched_big_lock;
struct spinlock sched_lock [MAX_CPU];


/**
 * Initialize lock for schedulers.
 */
void init_scheduler ()
{
	spinlock_init (&sched_big_lock);
}


/** 
 * Initializec scheduler on current CPU.
 * Threads can be enqueued.
 */
thread_t init_scheduler_cpu ()
{
	spinlock_init (&sched_lock [cpuid ()]);

	list_init (&sched_threads [cpuid ()]);
	
	thread_create_on_cpu (&sched_timer [cpuid ()], timer_worker, NULL, TF_NEW_VMM, cpuid ());
	sched_timer [cpuid ()]->is_unstoppable = true;

	return sched_timer[ cpuid ()];
}


/**
 * Starts hardware timer on current CPU and thread switching.
 * It's important to run this function just before 
 * switching context in the main function to the example thread.
 * If it's started to early an interrupt might fire before 
 * main function switched which results in panic. 
 */
void start_scheduler_cpu ()
{
	timer_setup (THREAD_QUANTUM); 
}


/** 
 * Enqueues thread on current CPU. Fails if thread 
 * has been already started on different CPU.
 *
 * @param thread
 *	Thread to enqueue.
 *
 * @see thread_t
 * @see sched_threads
 */
void enqueue (thread_t thread)
{
	assert (thread != NULL);

	ipl_t state = query_and_disable_interrupts ();
	spinlock_lock (&sched_lock [thread->cpu]);	
//	spinlock_lock (&sched_big_lock);	
	
	if (thread->queued_on_cpu == true)
	{
		panic ("Scheduler: scheduling already queued thread.");
	}
	else
	{
		list_append (&sched_threads [thread->cpu], &thread->item);
		thread->queued_on_cpu = true;
	}

	spinlock_unlock (&sched_lock [thread->cpu]);	
//	spinlock_unlock (&sched_big_lock);	
	conditionally_enable_interrupts (state);
}


/** 
 * Dequeues thread on current CPU. Some threads 
 * cannot be deqeueud.
 *
 * @param thread
 *	Thread to dequeue.
 *
 * @see thread_t
 * @see sched_threads
 */
void dequeue (thread_t thread)
{
	assert (thread != NULL);
	
	ipl_t state = query_and_disable_interrupts ();
	spinlock_lock (&sched_lock [thread->cpu]);
//	spinlock_lock (&sched_big_lock);	

	if (thread->is_unstoppable == true)
		goto finalize;	
	
	if (thread->queued_on_cpu == false)
		goto finalize;

	if (thread->item.list != &sched_threads [thread->cpu] )
	{
		panic ("Scheduler: wrong list by thread.");
	}
	
	list_remove (&sched_threads [thread->cpu], &thread->item);
	thread->queued_on_cpu = false;

	finalize:	
//		spinlock_unlock (&sched_big_lock);	
		spinlock_unlock (&sched_lock [thread->cpu]);
		conditionally_enable_interrupts (state);
}


/** 
 * Scheduled another thread on CPU.
 * This function can be called explicitly or from
 * hardware timer interrupt. 
 *
 * @param cpu
 * 	CPU to schedule thread on.
 *
 * @see thread_t
 * @see sched_threads
 * @see switch_to_thread
 */
void schedule (uint32_t cpu)
{
	ipl_t state = query_and_disable_interrupts ();
	spinlock_lock (&sched_lock [cpu]);
//	spinlock_lock (&sched_big_lock);	

	/* Adjust timers. */
	timer_ticks (read_cp0_count ());
	
	/* Prepare next timer interrutp. */
	timer_setup (THREAD_QUANTUM); 
	
	thread_t next_thread = (thread_t)list_rotate (&sched_threads [cpu]);

//	spinlock_unlock (&sched_big_lock);	
	spinlock_unlock (&sched_lock [cpu]);
	
	switch_to_thread (next_thread);
	
	conditionally_enable_interrupts (state);
}


/**
 * Just alias to keep old names.
 * Same semantics as called function.
 *
 * @see enqueue
 */
void runnable (thread_t thread)
{
	enqueue (thread);
}


/**
 * Just alias to keep old names.
 * Same semantics as called function.
 *
 * @see dequeue
 */
void disabled (thread_t thread)
{
	dequeue (thread);
}


/**
 * Just alias to keep old names.
 * Same semantics as called function.
 *
 * @see schedule
 */
void scheduled (void)
{
	schedule (cpuid ());
}
