/**
 * @file cvar.c
 *
 * Conditional variables are made of enhanced semaphores.
 *
 */


#include <sync/cvar.h>

#include <sync/mutex.h>


/** initialization of cond structure
 *
 * @param cvar pointer to condition variable structure
 *
 */
void cond_init (struct cond * cvar)
{
	/* internal semaphore initialization */
	sem_init (&(cvar->internal_semaphore), 0);
}


/** freeing of cond structure
 *
 * @param cvar pointer to condition variable structure
 *
 */
void cond_destroy (struct cond * cvar)
{
	/* internal semaphore freeing */
	sem_destroy (&(cvar->internal_semaphore));
}


/** set one thread free
 *
 * @param cvar pointer to condition variable structure
 *
 */
void cond_signal (struct cond * cvar)
{
	/* raise the semaphore by limited version, i.e. just unblock the thread, not raise indeed */
	sem_up_limited (&(cvar->internal_semaphore), 0);
}


/** set all threads free
 *
 * @param cvar pointer to condition variable structure
 *
 */
void cond_broadcast (struct cond * cvar)
{
	/* free all blocked threads on internal semaphore */
	sem_lay_down_all_burdens (&(cvar->internal_semaphore));
}


/** thread will wait on condition variable for freeing
 *
 * @param cvar pointer to condition variable structure
 * @param mtx pointer to mutex to work with
 *
 */
void cond_wait (struct cond * cvar, struct mutex * mtx)
{
	/* just call parametrized enhanced semaphore function */
	sem_down_with_revert_mutex (&(cvar->internal_semaphore), mtx, 0, 0);
}


/* thread will wait on condition variable for freeing or time limit
 *
 * @param cvar pointer to condition variable structure
 * @param mtx pointer to mutex to work with
 * @param usec time limit in microseconds
 *
 * @return returns ETIMEDOUT if time limit expires or EOK otherwise
 *
 */
int cond_wait_timeout (struct cond * cvar, struct mutex * mtx, const unsigned int usec)
{
	/* just call parametrized enhanced semaphore function */
	return sem_down_with_revert_mutex (&(cvar->internal_semaphore), mtx, 1, usec);
}