/**
 * @file mutex.c
 *
 * Mutes is implemented by semaphore with enhanced functions.
 *
 */

#include <sync/mutex.h>
#include <exc/assert.h>
#include <mm/malloc.h>
#include <adt/hashmap.h>
#include <include/error_codes.h>

/** Hashmap - stores initialized user mutexes. */
map_t u_mutexes = NULL;

/** Internal mutex for synchronized access to hashmap. */
struct mutex* hashmap_u_mutexes_mutex = NULL;

/** Initialize the structure for user synchronization. */
void init_u_sync (void)
{
	u_mutexes = hashmap_new ();
	hashmap_u_mutexes_mutex = malloc (sizeof (struct mutex));
	mutex_init (hashmap_u_mutexes_mutex);
}


/** initialize mutex structure
 *
 * @param mtx pointer to mutex structure
 *
 */
void mutex_init (struct mutex * mtx)
{
	/* initialize internal semaphore */
	sem_init (&(mtx->sem), 1);
}


/** freeing mutex structure
 *
 * @param mtx pointer to mutex structure
 *
 */
void mutex_destroy (struct mutex * mtx)
{
	/* destroying internal semaphore */
	int res = sem_destroy (&(mtx->sem));
	if (res != EOK)
	{
		/* call panic */
		panic ("Annihilation of blocking semaphore!");
	}
}


/** mutex lock down
 *
 * @param mtx pointer to mutex structure
 *
 */
void mutex_lock (struct mutex * mtx)
{
	/* down on one count semaphore is mutex lock */
	sem_down (&(mtx->sem));
}


/** mutex lock down for some maximum time
 *
 * @param mtx pointer to mutex structure
 * @param usec time limit in microseconds
 *
 * @return returns EOK if unlocked or ETIMEDOUT if expires time limit
 *
 */
int mutex_lock_timeout (struct mutex * mtx, const unsigned int usec)
{
	/* down on one count semaphore is mutex lock */
	return sem_down_timeout (&(mtx->sem), usec);
}


/** 
 * Initialize the user mutex structure to the unlocked state. 
 * User mutexes are stored in the hashmap, so that for the mutex operation we can control, if the given mutex was 
 * successfully created as a user mutex and its authorization.
 * @link kernel/adt/hashmap.h Hashmap implementation @endlink
 *
 * @param mtx Pointer to kernel address, where will be stored the kernel mutex related to the new user mutex.
 * @return ENOMEM if isn't enough memory for new user mutex, EOK else 
 * @see sem_init
 */
int u_mutex_init (void** mtx) 
{

	/* We store all the user mutexes in the hashmap. There is an unsigned long integer as a key (address of kernel 
	 * mutex related to the user mutex) and an unsigned long integer as a record (address of owner thread). */

	struct mutex** mutex = (struct mutex**) mtx;
	*mutex = malloc (sizeof(struct mutex));
	
	/* Add the record of the new user thread to the hashmap. */
	uint32_t owner_thread = (uint32_t) thread_get_current ();
	mutex_lock (hashmap_u_mutexes_mutex);
	if ((!(*mutex)) || (hashmap_put (u_mutexes, (unsigned long) *mutex, (void*) owner_thread) != MAP_OK))
	{
		/* Not enough memory. */
		free (*mutex);
		mutex_unlock (hashmap_u_mutexes_mutex);
		return ENOMEM;
	}
	mutex_unlock (hashmap_u_mutexes_mutex);
	
	/* Initialization of internal structure. */	
	sem_init (&((*mutex)->sem), 1);
	(*mutex)->owner = thread_get_current ();

	return EOK;
}


/** 
 * Function performs cleaning the mutex structure. 
 * If there are still some threads blocked on destroyed mutex, cancel the current thread. Checks, if the user mutex 
 * was previously allocated (is in the hashmap), and then destroys and removes it from the hashmap.
 * @link kernel/adt/hashmap.h Hashmap implementation @endlink
 *
 * @param mtx Pointer to kernel mutex structure.
 * @return EINVAL if given @mtx wasn't initialized, EOK else
 * @see sem_destroy
 */
int u_mutex_destroy (void* mtx)
{
	/* Is @mtx a correct pointer to kernel mutex related to user mutex?. */
	uint32_t owner_thread = NULL;
	mutex_lock (hashmap_u_mutexes_mutex);
	if (hashmap_get (u_mutexes, (uint32_t) mtx, (void**) &owner_thread) != MAP_OK)
	{
		mutex_unlock (hashmap_u_mutexes_mutex);
		return EINVAL;
	}
	
	/* Only threads from the same process as the owner can handle with the mutex. */
	struct mutex* mutex = (struct mutex*) mtx;
	if (mutex->owner->vm_map != (thread_get_current ())->vm_map)
	{
		mutex_unlock (hashmap_u_mutexes_mutex);
		return EINVAL;
	}
	
	/* Destroying the internal structure. */	
	int res = sem_destroy (&(mutex->sem));
	if (res != EOK)
	{
		/* Some threads are blocked on this mutex -> cancel the current thread. */
		mutex_unlock (hashmap_u_mutexes_mutex);
		thread_finish_over (thread_get_current ());
	}
	
	/* Free the mutex structure and remove the record in the hashmap. */
	hashmap_remove (u_mutexes, (uint32_t) mtx);
	mutex_unlock (hashmap_u_mutexes_mutex);
	
	free (mutex);
	return res;
}


/**
 * Fuction attempts to lock the mutex.
 * If not possible, it blocks the current thread. Checks, if the user mutex was previously allocated (is in the 
 * hashmap). Mutex locking is allowed only for the threads from the same process as the owner of the mutex.
 * @link kernel/adt/hashmap.h Hashmap implementation @endlink
 *
 * @param mtx Pointer to kernel mutex structure.
 * @return EINVAL if given @mtx wasn't initialized or if the mutex isn't accessible for the current thread, EOK else
 * @see sem_down
 */
int u_mutex_lock (void* mtx)
{
	/* Is @mtx a correct pointer to kernel mutex related to user mutex?. */
	uint32_t owner_thread = NULL;
	mutex_lock (hashmap_u_mutexes_mutex);
	if (hashmap_get (u_mutexes, (uint32_t) mtx, (void**) &owner_thread) != MAP_OK)
	{
		mutex_unlock (hashmap_u_mutexes_mutex);
		return EINVAL;
	}
	mutex_unlock (hashmap_u_mutexes_mutex);
	
	/* Internal semaphor structure - locking. */	
	struct mutex* mutex = (struct mutex*) mtx;
	sem_down (&(mutex->sem));
	
	return EOK;
}


/**
 * See u_mutex_lock comment.
 * Variant _timeout waits for locking mutex @usec microseconds long as possible. If timeout 0, no threads to block.
 * @link kernel/adt/hashmap.h Hashmap implementation @endlink
 * 
 * @param mtx Pointer to kernel mutex structure.
 * @param usec Time limit in microseconds.
 * @return EINVAL if given @mtx wasn't initialized or if the mutex isn't accessible for the current thread, ETIMEDOUT 
 *	if failed to lock in time limit, EOK else
 * @see sem_down_timeout
 */
int u_mutex_lock_timeout (void* mtx, const unsigned int usec)
{
	/* Is @mtx a correct pointer to kernel mutex related to user mutex?. */
	uint32_t owner_thread = NULL;
	mutex_lock (hashmap_u_mutexes_mutex);
	if (hashmap_get (u_mutexes, (uint32_t) mtx, (void**) &owner_thread) != MAP_OK)
	{
		mutex_unlock (hashmap_u_mutexes_mutex);
		return EINVAL;
	}
	mutex_unlock (hashmap_u_mutexes_mutex);
	
	/* Only threads from the same process as the owner can handle with the mutex. */
	struct mutex* mutex = (struct mutex*) mtx;
	if (mutex->owner->vm_map != (thread_get_current ())->vm_map)
	{
		return EINVAL;
	}
	
	/* Internal semaphor structure - locking with time limit. */	
	int result = sem_down_timeout (&(mutex->sem), usec);
	
	return result;
}


/** 
 * The function unlocks the given mutex. 
 * Wakes up threads that were blocked to lock the mutex. Checks, if the user mutex was previously allocated 
 * (is in the hashmap). If the symbol DEBUG_MUTEX >= 1 defined, the mutex implementation makes sure that the 
 * unlocking thread that it originally locked. If the mutex unlock another thread, cancel the current thread.
 * @link kernel/adt/hashmap.h Hashmap implementation @endlink
 * 
 * @param mtx Pointer to kernel mutex structure.
 * @return EINVAL if given @mtx wasn't initialized, EOK else
 * @see sem_up_limited
 */
int u_mutex_unlock (void* mtx)
{
	/* Is @mtx a correct pointer to kernel mutex related to user mutex?. */
	uint32_t owner_thread = NULL;
	mutex_lock (hashmap_u_mutexes_mutex);
	if (hashmap_get (u_mutexes, (uint32_t) mtx, (void**) &owner_thread) != MAP_OK)
	{
		mutex_unlock (hashmap_u_mutexes_mutex);
		return EINVAL;
	}
	mutex_unlock (hashmap_u_mutexes_mutex);
	
	/* Only threads from the same process as the owner can handle with the mutex. */
	struct mutex* mutex = (struct mutex*) mtx;
	if (mutex->owner->vm_map != (thread_get_current ())->vm_map)
	{
		return EINVAL;
	}
	
	/* Control of unlocking by another thread. */
	#ifdef DEBUG_MUTEX
		if ((0 < DEBUG_MUTEX) && (thread_get_current () != (mutex->sem).last_owner))
		{
			/* Unlocking by another thread - cancel the current thread. */
			thread_finish_over (thread_get_current ());
		}
	#endif
	
	/* Internal semaphor structure - unlocking. */	
	sem_up_limited (&(mutex->sem), 1);
	
	return EOK;
}
