/**
 * @file rwlock.c
 *
 * Read-write lock is based on read-lock semaphore and write-lock mutex.
 *
 */


#include <sync/rwlock.h>
#include <sync/semaphore.h>
#include <include/error_codes.h>


/** initialize rwlock data structure 
 *
 * @param rwl pointer to target lock structure
 *
 */
void rwlock_init (struct rwlock * rwl)
{
	/* initialize mutex lock */
	mutex_init (&(rwl->write_lock));
	/* initialize semaphore for value 1 */
	sem_init (&(rwl->read_lock), 1);
}


/** clean up and free rwlock data structure
 *
 * @param rwl pointer to target lock structure
 *
 */
void rwlock_destroy (struct rwlock * rwl)
{
	/* destroy mutex for write lock */
	mutex_destroy (&(rwl->write_lock));
	/* destroy semaphore for read lock */
	sem_destroy (&(rwl->read_lock));
}


/** locks target lock for reading
 *
 * @param rwl pointer to target lock structure
 *
 */
void rwlock_read_lock (struct rwlock * rwl)
{
	/* if locked for writting, cant read */
	mutex_lock (&(rwl->write_lock));
	/* notice reader in crit section */
	sem_down_unblockable (&(rwl->read_lock));
	/* read lock is not real write lock */
	mutex_unlock (&(rwl->write_lock));
}


/** locks target lock for writting
 *
 * @param rwl pointer to target lock structure
 *
 */
void rwlock_write_lock (struct rwlock * rwl)
{
	/* lock writting lock */
	mutex_lock (&(rwl->write_lock));
	/* write lock contains read lock */
	sem_down (&(rwl->read_lock));
}


/** locks target lock for reading until time limit expires
 *
 * @param rwl pointer to target lock structure
 * @param usec time limit
 *
 */
int rwlock_read_lock_timeout (struct rwlock * rwl, const unsigned int usec)
{
	/* if locked for writting, cant read */
	int result = mutex_lock_timeout (&(rwl->write_lock), usec);
	/* if time limit expired, thread dont lower the semaphore */
	if (ETIMEDOUT != result)
	{
		/* notice reader in crit section */
		sem_down_unblockable (&(rwl->read_lock));
		/* read lock is not real write lock */
		mutex_unlock (&(rwl->write_lock));
	}
	/* returns what returned writting lock */
	return result;
}


/** locks target lock for writting until time limit expires
 *
 * @param rwl pointer to target lock structure
 * @param usec time limit
 *
 */
int rwlock_write_lock_timeout (struct rwlock * rwl, const unsigned int usec)
{
	/* lock writting lock */
	int result = mutex_lock_timeout (&(rwl->write_lock), usec);
	/* if time limit expired, thread dont lower the semaphore */
	if (ETIMEDOUT != result)
	{
		/* if time limit didnt expired, try another lock */
		result = sem_down_timeout (&(rwl->read_lock), usec);
		/* if write lock was locked properly and read lock was not, unlock the write lock */
		if (ETIMEDOUT == result)
		{
			/* unlock the locked write lock */
			mutex_unlock (&(rwl->write_lock));
		}
	}
	/* returns what returned writting lock */
	return result;
}


/** unlocks target lock for writting
 *
 * @param rwl pointer to target lock structure
 *
 */
void rwlock_write_unlock (struct rwlock * rwl)
{
	/* unlock writting lock */
	mutex_unlock (&(rwl->write_lock));
	/* write lock contains read lock */
	sem_up (&(rwl->read_lock));
}


/** unlocks target lock for reading
 *
 * @param rwl pointer to target lock structure
 *
 */
void rwlock_read_unlock (struct rwlock * rwl)
{
	/* notice reader leaves crit section */
	sem_up (&(rwl->read_lock));
}