/**
 * @file semaphore.c
 *
 * Semaphore is the core synchronization primitive. Others are derived from it.
 *
 */


#include <sync/semaphore.h>

#include <drivers/printer.h>
#include <exc/assert.h>
#include <proc/thread.h>
#include <sched/sched.h>
#include <mm/malloc.h>
#include <drivers/dorder.h>
#include <sync/mutex.h>


/** structure passed to timer handler */
struct semaphore_timer_data
{
	/* blocking semaphore */
	struct semaphore * sem;
	/* thread blocked by semaphore */
	thread_t blocked_thread;
	/* true = timer executed; false = timer not executed */
	bool timed_out;
};

/**
 * Creating semaphore.
 *
 * @param sem Pointer to semaphore structure.
 * @param value The value of semaphore.
 *
 * @return ENOMEM if not enough memory, EOK else
 *
 */
int sem_create (struct semaphore** sem, const int value)
{
	/* Allocating the structure. */
	*sem = malloc (sizeof (struct semaphore));
	/* Not enough memory. */
	if (!(*sem))
	{
		return ENOMEM;
	}
	/* Initialization. */
	sem_init (*sem, value);
	return EOK;
}
  

/** initialize semaphore
 *
 * @param sem pointer to semaphore structure
 *
 */
void sem_init (struct semaphore * sem, const int value)
{
	/* init max possible threads value */
	sem->free_space = value;
	/* initializing queue */
	list_init (&(sem->waiting_queue));
	/* initialize critical section spinlock */
	spinlock_init (&(sem->crit_lock));
	/* semaphore wasnt lowered yet, there is no last owner */
	sem->last_owner = NULL;
}


/** freeing semaphore
 *
 * @param sem pointer to semaphore structure
 *
 * @return EOK if successful, EINVAL if some threads are waiting on the semaphore
 *
 */
int sem_destroy (struct semaphore * sem)
{
	/* blocking semaphore annihilation causes kernel panic */
	if (sem->waiting_queue.head)
	{
		/* semaphore freeing will fail */
		return EINVAL;
	}
	/* critical section spinlock freeing */
	spinlock_destroy (&(sem->crit_lock));
	/* we succesfully destroyed the semaphore */
	return EOK;
}


/** returns semaphore actual value
 *
 * @param sem pointer to semaphore structure
 *
 */
int sem_get_value (struct semaphore * sem)
{
	/* returns count of threads which can be blocked */
	return sem->free_space;
}


/** raise the semaphore
 *
 * @param sem pointer to semaphore structure
 * @param limit raising limit
 * @param isLimited true = is limited; false = unlimited
 *
 */

void sem_up_internal (struct semaphore * sem, const int limit, bool isLimited);
void sem_up_internal (struct semaphore * sem, const int limit, bool isLimited)
{
	/* get first blocked thread */
	item_t * first_item = sem->waiting_queue.head;
	/* get item as thread */
	thread_t thr = (thread_t)first_item;
	/* if there is some blocked thread, do not touch counter and let thread run if it should run */
	if ((thr) && (!(0 > sem->free_space)))
	{
		/* remove blocked thread from semaphore */
		list_remove (&(sem->waiting_queue), first_item);
		/* plan that thread to run */
		runnable (thr);
		/* change the owner */
		sem->last_owner = thr;
	}
	/* if there is no blocked thread, just increment counter */
	else
	{
		/* limited semaphores wont be raised */
		if ((sem->free_space < limit) || (!isLimited))
		{
			/* counter increment */
			sem->free_space++;
		}
		/* reset the last owner */
		sem->last_owner = NULL;
	}
}



/** raise the semaphore but not over given value
 *
 * @param sem pointer to semaphore structure
 * @param limit limit of semaphore raising
 *
 */
void sem_up_limited (struct semaphore * sem, const int limit)
{
	/* disable interrupts to run critical section faster */
	ipl_t state = query_and_disable_interrupts ();
	/* critical section atomicity */
	spinlock_lock (&(sem->crit_lock));

	/* call limited version of sem_up */
	sem_up_internal (sem, limit, 1);

	/* end of critical section */
	spinlock_unlock (&(sem->crit_lock));
	/* enable interrupts again */
	conditionally_enable_interrupts (state);
}


/** raise the semaphore
 *
 * @param sem pointer to semaphore structure
 *
 */
void sem_up (struct semaphore * sem)
{
	/* disable interrupts to make things faster */
	ipl_t state = query_and_disable_interrupts ();
	/* critical section atomicity */
	spinlock_lock (&(sem->crit_lock));

	/* call unlimited version of sem_up */
	sem_up_internal (sem, 0, 0);

	/* end of critical section */
	spinlock_unlock (&(sem->crit_lock));
	/* enable interrupts again */
	conditionally_enable_interrupts (state);
}


/** semaphore passed anyway due to time limit
 *
 * @param tmr executing timer
 * @param data semaphore timer data with thread_t and flags
 *
 */
void pass_anyway (struct timer * tmr, void * data);
void pass_anyway (struct timer * tmr, void * data)
{
	/* blocked thread and flag for timeout */
	struct semaphore_timer_data * std = (struct semaphore_timer_data *) data;

	/* disable interrupts to make things faster */
	ipl_t state = query_and_disable_interrupts ();
	/* critical section atomicity */
	spinlock_lock (&(std->sem->crit_lock));

	/* this condition ensures that thread is still blocked */
	if (&(std->sem->waiting_queue) == (std->blocked_thread->item.list))
	{
		/* set flag to timeout */
		std->timed_out = 1;
		/* unblock thread */
		list_remove (&(std->sem->waiting_queue), &(std->blocked_thread->item));
		/* schedule that thread */
		runnable (std->blocked_thread);
	}

	/* end of critical section */
	spinlock_unlock (&(std->sem->crit_lock));
	/* enable interrupts again */
	conditionally_enable_interrupts (state);
}


/** decrease or lock semaphore
 *
 * @param sem pointer to semaphore structure
 * @param usec time limit for pass anyway
 * @param timed true = time limit; false = time unlimited
 *
 * @return returns EOK if passed or ETIMEDOUT if time limit fades away
 *
 */
int sem_down_internal (struct semaphore * sem, const unsigned int usec, bool timed, bool unblockable, struct mutex * cvar_mutex);
int sem_down_internal (struct semaphore * sem, const unsigned int usec, bool timed, bool unblockable, struct mutex * cvar_mutex)
{
	/* timer for time limited version */
	struct timer tmr;
	/* data for timer handler routine */
	struct semaphore_timer_data std;
	
	/* disable interrupts to make things faster */
	ipl_t state = query_and_disable_interrupts ();
	/* critical section atomicity */
	spinlock_lock (&(sem->crit_lock));

	/* semaphore can be lowered */
	if ((0 < sem->free_space) || (unblockable))
	{
		/* just lower the semaphore and nothing more */
		sem->free_space--;
		/* set the last owner */
		sem->last_owner = thread_get_current ();
		/* end of critical section */
		spinlock_unlock (&(sem->crit_lock));
		/* enable interrupts again */
		conditionally_enable_interrupts (state);
		/* thread was not blocked */
		return EOK;
	}
	/* semaphore can not be lowered, thread has to be blocked */
	else
	{
		/* get current thread */
		thread_t thr = thread_get_current ();
		/* init timer only if it will be needed */
		if (timed)
		{
			/* initialize timer structure */
			timer_init (&tmr, usec, pass_anyway, &std);
			/* add semaphore to the structure */
			std.sem = sem;
			/* add current thread to the structure */
			std.blocked_thread = thr;
			/* set flag to not timeout yet */
			std.timed_out = 0;
		}
		/* zero timeout is not blocked */
		if ((timed) && (!(0 < usec)))
		{
			/* end of critical section */
			spinlock_unlock (&(sem->crit_lock));
			/* enable interrupts again */
			conditionally_enable_interrupts (state);
			/* if there is mutex for condition variable, work with it */
			if (cvar_mutex)
			{
				/* unlock mutex */
				mutex_unlock (cvar_mutex);
				/* with unlocked mutex give a chance to other threads */
				thread_yield ();
				/* lock mutex again */
				mutex_lock (cvar_mutex);
			}
			/* thread would be blocked if time was gt 0 */
			return ETIMEDOUT;
		}
		/* plan out actual thread */
		disabled (thr);
		/* block actual thread on this semaphore */
		list_append (&(sem->waiting_queue), &(thr->item));
		/* if there is mutex for condition variable, work with it */
		if (cvar_mutex)
		{
			/* unlock mutex */
			mutex_unlock (cvar_mutex);
		}
		/* if there is timeout, start timer */
		if (timed)
		{
			/* timer start */
			timer_start (&tmr);
		}
		/* end of critical section */
		spinlock_unlock (&(sem->crit_lock));
		/* enable interrupts again */
		conditionally_enable_interrupts (state);
		/* schedule out of this thread */
		scheduled ();
		/* if there is mutex for condition variable, work with it */
		if (cvar_mutex)
		{
			/* lock mutex */
			mutex_lock (cvar_mutex);
		}
		/* timer is initialized if timed */
		if (timed)
		{
			/* clean up timer */
			timer_destroy (&tmr);
		}
		/* now thread could be planned because of timeout or not */
		if ((timed) && (std.timed_out))
		{
			/* thread is unblocked due to timeout */
			return ETIMEDOUT;
		}
		else
		{
			/* thread is unblocked by normal way */
			return EOK;
		}
	}	
}


/** decrease or lock semaphore if it is possible in given time
 *
 * @param sem pointer to semaphore structure
 * @param usec time limit for passing semaphore anyway
 * 
 * @return returns EOK if locked or ETIMEDOUT if time limit fades away
 *
 */
int sem_down_timeout (struct semaphore * sem, const unsigned int usec)
{
	/* call internal function with proper parameters */
	return sem_down_internal (sem, usec, 1, 0, 0);
}


/** decrease or lock semaphore
 *
 * @param sem pointer to semaphore structure
 *
 */
void sem_down (struct semaphore * sem)
{
	/* call internal function with proper parameters */
	sem_down_internal (sem, 0, 0, 0, 0);
}


/** decrease the semaphore but never block a thread
 *
 * @param sem pointer to semaphore structure
 *
 */
void sem_down_unblockable (struct semaphore * sem)
{
	/* call internal function with proper parameters */
	sem_down_internal (sem, 0, 0, 1, 0);
}


/* lower the semaphore; if thread blocked, given mutex unlocked for the time of blocking
 *
 * @param sem pointer to semaphore structure
 * @param mtx pointer to mutex structure (retyped for easier declaration)
 * @param timed true = time limit; false = without time limit
 * @param usec time limit in microseconds
 *
 */
int sem_down_with_revert_mutex (struct semaphore * sem, void * mtx, bool timed, const unsigned int usec)
{
	/* call internal function with proper parameters */
	return sem_down_internal (sem, usec, timed, 0, (struct mutex *)mtx);
}


/** free all threads blocked by this semaphore; do not touch the counter
 *
 * @param sem pointer to semaphore structure
 *
 */
void sem_lay_down_all_burdens (struct semaphore * sem)
{
	/* disable interrupts to make things faster */
	ipl_t state = query_and_disable_interrupts ();
	/* critical section atomicity */
	spinlock_lock (&(sem->crit_lock));

	/* get all blocked threads, one by one */
	while (sem->waiting_queue.head)
	{
		/* call unlimited version of sem_up */
		sem_up_internal (sem, 0, 1);
	}

	/* end of critical section */
	spinlock_unlock (&(sem->crit_lock));
	/* enable interrupts again */
	conditionally_enable_interrupts (state);
}
