/**
 * @file semaphore.h
 *
 * Semaphore is the core synchronization primitive. Others are derived from it.
 *
 */


#ifndef SEMAPHORE_H_
#define SEMAPHORE_H_


#include <include/shared.h>
#include <include/c.h>
#include <adt/list.h>
#include <sync/spinlock.h>
#include <proc/thread.h>


/** structure for semaphore */
struct semaphore
{
	/* count of threads that this semaphore can accept */
	int32_t free_space;
	/* queue of threads waiting for this semaphore (removeds from scheduler) */
	list_t waiting_queue;
	/* spinlock to protect critical sections in this semaphore */
	struct spinlock crit_lock;
	/* last thread which lower the semaphore - just for mutex owner purposes */
	thread_t last_owner;
};


/* creating of semaphore structure */
int sem_create (struct semaphore ** sem, const int value);


/* initialization of semaphore structure */
void sem_init (struct semaphore * sem, const int value);


/* freeing of semaphore structure */
int sem_destroy (struct semaphore * sem);


/* return actual semaphore value */
int sem_get_value (struct semaphore * sem);


/* raise the semaphore */
void sem_up (struct semaphore * sem);


/* raise the semaphore but not over given value; this is here because of mutex implementation */
void sem_up_limited (struct semaphore * sem, const int limit);


/* decrease-lock the semaphore */
void sem_down (struct semaphore * sem);


/* decrease the semaphore but never block a thread */
void sem_down_unblockable (struct semaphore * sem);


/* decrease-lock the semaphore if it is possible in given time */
int sem_down_timeout (struct semaphore * sem, const unsigned int usec);


/* lower the semaphore; if thread blocked, given mutex unlocked for the time of blocking */
int sem_down_with_revert_mutex (struct semaphore * sem, void * mtx, bool timed, const unsigned int usec);


/* free all threads blocked by this semaphore; do not touch the counter */
void sem_lay_down_all_burdens (struct semaphore * sem);


#endif