/**
 * @file spinlock.c
 *
 * Active synchronization primitive ensures atomicity for higher primitives.
 *
 */


#include <sync/spinlock.h>

#include <drivers/printer.h>
#include <exc/assert.h>


/** initialize spinlock
 *
 * @param splk pointer to spinlock structure
 *
 */
void spinlock_init (struct spinlock * splk)
{
	/* spinlock starts unlocked */
	splk->lock = 0;
}


/** free spinlock
 *
 * @param splk pointer to spinlock structure
 *
 */
void spinlock_destroy (struct spinlock * splk)
{
	/* annihilation of locked spinlock causes kernel panic */
	if (0 < splk->lock)
	{
		/* call panic */
		panic ("Locked spinlock annihilation!");
	}
}


/** lock spinlock
 *
 * @param splk pointer to spinlock structure
 *
 */
void spinlock_lock (struct spinlock * splk)
{
	/* just for associate with register */
	native_t spinreg;

	asm volatile
	(
		".set push\n"
		".set noat\n"
		".set noreorder\n"

		"1: ll %[spinreg], %[spinlock]\n"
		"   bnez %[spinreg], 1b\n"
		"   li %[spinreg], 1\n"
		"   sc %[spinreg], %[spinlock]\n"
		"   beqz %[spinreg], 1b\n"

		"   sync\n"

		".set pop\n"
		:
		[spinreg] "=&r" (spinreg),
		[spinlock] "+m" (splk->lock)
		:
		/* no input */
		:
		"memory"
	);
}


/** unlock spinlock
 *
 * @param splk pointer to spinlock structure
 *
 */
void spinlock_unlock (struct spinlock * splk)
{
	/* unlock is atomic due to ll-sc */
	splk->lock = 0;
} 