/**
 * @file test_nblock.c
 *
 * Nonblocking list benchmark.
 *
 * The benchmark itself consists of six parts. First three of them test the speed of operations for elements with random 
 * generated keys. The other three parts test operations only for fixed keys. In the first and fourth part are the elements 
 * inserted to the list, the second and fifth is used for their research and finally the third and sixth for their removing. 
 * In every test is first of all made the set number of threads and each of them makes the constant number of operations on 
 * the set shared list. The important condition is the fact, that all the threads start the test code at the same time. The 
 * time that runs from the start of the test to the final end of all threads is measured and recorded. By using the 
 * modification of macros of the source files of benchmark it is possible to modify the number of operations for every 
 * thread, scale of generating of random numbers etc.
 * @link doc/files/nblock_study.pdf Speed comparison of non-blocking algorithm @endlink
 * 
 * @see nblock_list
 * @see nblock_list_remove
 * @see nblock_list_insert
 * @see nblock_list_find
 */

static char * desc =      
	"Nonblocking list benchmark.\n\n"
	"The benchmark itself consists of six parts.\n"
	"First three of them test the speed of operations for elements with\n"
	"random generated keys. The other three parts test operations only for\n"
	"fixed keys. In the first and fourth part are the elements inserted to\n"
	"the list, the second and fifth is used for their research and finally\n"
	"the third and sixth for their removing.\n\n"
	"In every test is first of all made the set number of threads and each\n"
	"of them makes the constant number of operations on the set shared list.\n\n"
	"The important condition is the fact, that all the threads start the\n"
	"test code at the same time. The time that runs from the start of the\n" 
	"test to the final end of all threads is measured and recorded.\n\n"
	"By using the modification of macros of the source files of benchmark it\n"
	"is possible to modify the number of operations for every thread, the \n"
	"number of threads, scale of generating of random numbers etc.\n\n";

#include <api.h>
#include "../include/defs.h"

/** The number of threads of each kind to run concurrently. */
#define THREAD_COUNT 5


/** 
 * The number of tests. 
 * Recommended at least 30 (for statistic comparison). 
 */
#define TEST_COUNT 5

/** The key range (there is inconsiderable the search operation for the large key ranges). */
#define KEY_RANGE 256
 
/**
 * The number of test parts. 
 * Random insert test, random find test, random remove test, direct insert test, direct find test, direct remove test - 6 parts. 
 */
#define TEST_PARTS 6

/** The number of operate cycles for every thread - for random key operations tests. */
#define RANDOM_OPERATE_CYCLES 10000

/** The number of operate cycles for every thread - for direct key operations tests. */
#define DIRECT_OPERATE_CYCLES 10000

/** The concurrent list for benchmark. */
static nblock_list_t list;


/** The array for storing random generated keys for the first three parts of test. */
volatile unsigned long rand_keys[THREAD_COUNT][RANDOM_OPERATE_CYCLES];

/**
 * The read-write lock.
 * In every part first we need to create all the threads (the critical section is still locked), then the section is unlocked 
 * and we begin to count the execute time.
 */
static struct rwlock rwlock;


/**
 * The function for random key inserting. 
 * The generated random keys is stored in the @rand_keys array.
 *
 * @param data The ID of the thread.
 * @see get_random
 * @see nblock_list_insert
 * @see rand_keys
 */
static void* thread_item_random_inserting (void* data)
{
	int thread_ID = (int)data;
	
	rwlock_read_lock (&rwlock);
	
	for (int i=0; i<RANDOM_OPERATE_CYCLES; ++i)
	{
		unsigned long rand = get_random () % KEY_RANGE;
		rand_keys[thread_ID][i] = rand;
		nblock_list_insert (&list, rand, DATA_MAGIC);
	}
	
	rwlock_read_unlock (&rwlock);
	
	return NULL;
}


/**
 * The function for random key finding. 
 * The keys for finding is stored in the @rand_keys array.
 *
 * @param data The ID of the thread.
 * @see nblock_list_find
 * @see rand_keys
 */
static void* thread_item_random_finding (void* data)
{
	void* item_data;
	int thread_ID = (int)data;
	
	rwlock_read_lock (&rwlock);
	
	for (int i=0; i<RANDOM_OPERATE_CYCLES; ++i)
	{
		nblock_list_find (&list, rand_keys[thread_ID][i], &item_data);
	}
	
	rwlock_read_unlock (&rwlock);
	
	return NULL;
}


/**
 * The function for random key removing. 
 * The keys for removing is stored in the @rand_keys array.
 *
 * @param data The ID of the thread.
 * @see nblock_list_remove
 * @see rand_keys
 */
static void* thread_item_random_removing (void* data)
{
	void* item_data;
	int thread_ID = (int)data;
	
	rwlock_read_lock (&rwlock);
	
	for (int i=0; i<RANDOM_OPERATE_CYCLES; ++i)
	{
		nblock_list_remove (&list, rand_keys[thread_ID][i], &item_data);
	}
	
	rwlock_read_unlock (&rwlock);
	
	return NULL;
}


/**
 * The function for direct key inserting.
 * Try to repeatly insert the one item with magic key.
 *
 * @param data The ID of the thread.
 * @see nblock_list_insert
 */
static void* thread_item_direct_inserting (void* data)
{
	rwlock_read_lock (&rwlock);
	
	for (int i=0; i<DIRECT_OPERATE_CYCLES; ++i)
	{
		nblock_list_insert (&list, (unsigned long)DIRECT_KEY_MAGIC, DATA_MAGIC);
	}
	
	rwlock_read_unlock (&rwlock);
	
	return NULL;
}


/**
 * The function for direct key finding. 
 * Try to repeatly find the one item with magic key.
 *
 * @param data The ID of the thread.
 * @see nblock_list_find
 */
static void* thread_item_direct_finding (void* data)
{
	void* item_data;
	
	rwlock_read_lock (&rwlock);
	
	for (int i=0; i<DIRECT_OPERATE_CYCLES; ++i)
	{
		nblock_list_find (&list, (unsigned long)DIRECT_KEY_MAGIC, &item_data);
	}
	
	rwlock_read_unlock (&rwlock);
	
	return NULL;
}


/**
 * The function for direct key removing. 
 * Try to repeatly remove the one item with magic key.
 *
 * @param data The ID of the thread.
 * @see nblock_list_remove
 */
static void* thread_item_direct_removing (void* data)
{
	void * item_data;
	
	rwlock_read_lock (&rwlock);
	
	for (int i=0; i<DIRECT_OPERATE_CYCLES; ++i)
	{
		nblock_list_remove (&list, (unsigned long)DIRECT_KEY_MAGIC, &item_data);
	}
	
	rwlock_read_unlock (&rwlock);
	
	return NULL;
}


void run_test (void)
{
	
	/* Init. */
	rwlock_init (&rwlock);
	nblock_list_init (&list);
	int cnt;
	unsigned long curr_time;
	int test_nmb = 0;
	
	/* The array of threads. */
	thread_t threads [THREAD_COUNT];
	
	/* The array for storing the execute time for every part of test. */
	unsigned long time_array [TEST_COUNT][TEST_PARTS]; 

	printk (desc);
	
	printk ("Benchmark started.\n\n");
	
	while(test_nmb < TEST_COUNT)
	{
		
		
		/* ******** */
		/*  PART 1  */
		/* ******** */
	
		rwlock_write_lock (&rwlock);
	
		for (cnt=0; cnt<THREAD_COUNT; ++cnt)
		{
			threads [cnt] = robust_thread_create (thread_item_random_inserting, (void*)cnt);
		}
	
		curr_time = time ();
		rwlock_write_unlock (&rwlock);
	
		/* Wait for all the threads to finish. */
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			robust_thread_join (threads [cnt]);
		}
	
		time_array [test_nmb][0] = time () - curr_time;
		printk ("\tPart 1/6 finished.\n");
	
	
		/* ******** */
		/*  PART 2  */
		/* ******** */
	
		rwlock_write_lock (&rwlock);
	
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			threads [cnt] = robust_thread_create (thread_item_random_finding, (void*)cnt);
		}
	
		curr_time = time ();
		rwlock_write_unlock (&rwlock);
	
		/* Wait for all the threads to finish. */
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			robust_thread_join (threads [cnt]);
		}
	
		time_array [test_nmb][1] = time () - curr_time;
		printk ("\tPart 2/6 finished.\n");
		
		
		/* ******** */
		/*  PART 3  */
		/* ******** */
	
		rwlock_write_lock (&rwlock);
		
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			threads [cnt] = robust_thread_create (thread_item_random_removing, (void*)cnt);
		}
	
		curr_time = time ();
		rwlock_write_unlock (&rwlock);
	
		/* Wait for all the threads to finish. */
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			robust_thread_join (threads [cnt]);
		}
	
		time_array [test_nmb][2] = time () - curr_time;
		printk ("\tPart 3/6 finished.\n");
		
		
		/* ******** */
		/*  PART 4  */
		/* ******** */
	
		rwlock_write_lock (&rwlock);
	
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			threads [cnt] = robust_thread_create (thread_item_direct_inserting, (void*)cnt);
		}
	
		curr_time = time ();
		rwlock_write_unlock (&rwlock);
	
		/* Wait for all the threads to finish. */
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			robust_thread_join (threads [cnt]);
		}
	
		time_array [test_nmb][3] = time () - curr_time;
		printk ("\tPart 4/6 finished.\n");
		
	
		/* ******** */
		/*  PART 5  */
		/* ******** */
	
		rwlock_write_lock (&rwlock);
	
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			threads [cnt] = robust_thread_create (thread_item_direct_finding, (void*)cnt);
		}
	
		curr_time = time ();
		rwlock_write_unlock (&rwlock);
	
		/* Wait for all the threads to finish. */
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			robust_thread_join (threads [cnt]);
		}
	
		time_array [test_nmb][4] = time () - curr_time;
		printk ("\tPart 5/6 finished.\n");
		
		
		/* ******** */
		/*  PART 6  */
		/* ******** */
	
		rwlock_write_lock (&rwlock);
	
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			threads [cnt] = robust_thread_create (thread_item_direct_removing, (void*)cnt);
		}
	
		curr_time = time ();
		rwlock_write_unlock (&rwlock);
	
		/* Wait for all the threads to finish. */
		for (cnt=0; cnt<THREAD_COUNT; ++cnt) 
		{
			robust_thread_join (threads [cnt]);
		}
	
		time_array [test_nmb][5] = time () - curr_time;
		
		printk ("\tPart 6/6 finished.\n");
		printk ("\nTest #%u finished.\n\n", test_nmb);
		++test_nmb;
	}
	
	/* Calculate the sum of time of each part of test. */
	unsigned long sum_time[TEST_PARTS];
	unsigned long sum;
	
	for(int part_id=0; part_id<TEST_PARTS; ++part_id)
	{
		sum = 0;
		for(int test_id=0; test_id<TEST_COUNT; ++test_id)
		{
			sum += time_array [test_id][part_id];
		}
		sum_time[part_id] = sum;
	}
	
	printk ("Random key cycles %u, direct key cycles %u.\n", RANDOM_OPERATE_CYCLES, DIRECT_OPERATE_CYCLES);
	printk ("The test counter %u.\n", TEST_COUNT);
	printk ("\nThe sum of execute time in each part of test: (in seconds)\n\n");
	
	printk ("Random insert test: %u\n", sum_time[0]);
	printk ("Random find test: %u\n", sum_time[1]);
	printk ("Random remove test: %u\n", sum_time[2]);
	printk ("Direct insert test: %u\n", sum_time[3]);
	printk ("Direct find test: %u\n", sum_time[4]);
	printk ("Direct remove test: %u\n", sum_time[5]);
	
	rwlock_destroy (&rwlock);
}
