/**
 * Non-blocking list test #1
 */

static char * desc =
    "Non-blocking list test #1\n"
    "Spawns a number of threads and makes them all remove the one\n"
    "item from a non-blocking list. Only one thread can remove it. Test\n"
    "of remove and find operation. \n\n";

#include <api.h>
#include "../include/defs.h"


/*
 * The number of threads of each kind to run concurrently.
 */
#define THREAD_COUNT 100

/*
 * The number of remove cycles for every thread.
 */
#define REMOVE_CYCLES 100

/*
 * The concurrent list.
 */
static nblock_list_t list;

/*
 * The read-write lock.
 * First we need to create all the threads, then unlock the section.
 */
static struct rwlock rwlock;

/*
 * Array of records, which thread removed the item.
 */
static volatile bool removed[THREAD_COUNT];

/*
 * Was the one item in the list removed?
 */
static volatile bool item_removed = false;


/*
 * The function that repeatly tries to remove the one item from the non-blocking linked list.
 *
 * @param data The ID of the thread.
 */
static void *
thread_item_removing (void * data)
{
	
	int thread_id = (int) data;
	void * list_data;
	bool result;
	removed[thread_id] = false;
	
	rwlock_read_lock (&rwlock);
	
	for (int i=0; i<REMOVE_CYCLES; ++i)
	{
		result = nblock_list_remove (&list, (unsigned long)KEY_MAGIC, &list_data);
		if (result)
		{
			removed[thread_id] = true;
			item_removed = true;
			assert(list_data == DATA_MAGIC);
		}
		
		if(item_removed)
		{
			assert(!nblock_list_find (&list, (unsigned long)KEY_MAGIC, &list_data));
		}
	}
	
	rwlock_read_unlock (&rwlock);
	
	return NULL;
}


void
run_test (void)
{
	rwlock_init (&rwlock);
	
	/* Init. */
	nblock_list_init (&list);
	
	/* Insert the one item to the list. */
	nblock_list_insert (&list, (unsigned long)KEY_MAGIC, DATA_MAGIC);
	
	int cnt;
	thread_t threads [THREAD_COUNT];
	
	printk (desc);
	
	rwlock_write_lock (&rwlock);
	
	for (cnt= 0; cnt < THREAD_COUNT; ++cnt) 
	{
		threads [cnt] = robust_thread_create (
		    thread_item_removing, (void*)cnt);
	}
	
	rwlock_write_unlock (&rwlock);
	
	/*
	 * Wait for all the threads to finish.
	 */
	printk ("Waiting for all the threads to finish...\n");
	
	/*
	 * Reap the threads.
	 */
	for (cnt = 0; cnt < THREAD_COUNT; ++cnt) 
	{
		robust_thread_join (threads [cnt]);
	}
	
	rwlock_destroy (&rwlock);
	
	printk ("Items in the list: (should be empty)\n");
	nblock_list_dump (&list);
	
	/* How many threads removed the item. */
	int counter = 0;
	for (int i=0; i<THREAD_COUNT; ++i)
	{
		if(removed[i])
		{
			++counter;
		}
	}
	
	if (counter == 1) {
		printk ("Test passed...\n");
	} else {
		printk ("Test failed...\n");
	}
}
