/**
 * @file assert.h
 *
 * Simple assert and debug tools support for userspace.
 *
 * There are implemented some macros and debug tools, mainly the macro assert and dprintf. The assert macro is useful 
 * for checking some assumption (if the evaluation fails, the current process is killed). dprintf macro prints the 
 * message given as the argument to the console. It can be used as simple log trace debug tool. The implemented macros 
 * are used only if the standart C-assertions is disabled (!(NDEBUG)).
 * @link kernel/exc/assert.h Kernel debug tools implementation @endlink
 */
 
#ifndef U_ASSERT_H_
#define U_ASSERT_H_

/* NDEBUG disables standard C-assertions */
#if !(NDEBUG)

/* GCC Compiler provides variable __FUNCTION__, which contains the name of current function. Unfortunatelly it is 
 * only GCC Compiler support, so we have to check if we work with GCC Compiler. There was a variable __func__ in 
 * older versions of GCC. These variables are not macros! */
#if __STDC_VERSION__ < 199901L
#if __GNUC__ >= 2
#define __func__ __FUNCTION__
#else /*__GNUC__*/
#define __func__ "<unknown>"
#endif /*__GNUC__*/
#endif /*__STDC_VERSION__*/

#include <librt/io.h>
#include <librt/thread.h>


/**
 * Standard macro for assert function.
 * If evaluation fails, the current file, fuction and line is printed to the console and the current process is killed.
 *
 * @param expr The expression for evaluation.
 * @see printf
 * @see exit 
 */
#define assert(expr) \
{ \
	if (!(expr)) \
	{ \
		printf ("Assert error in file %s, function %s, line %i.\n", __FILE__, __func__, __LINE__); \
		exit (); \
	} \
	else \
	{ \
		(void) 0; \
	} \
}


/**
 * Simple debug tool for userspace.
 * dprintf macro will print the name of file, the name of current function, the line number and the formatted text entered 
 * as the argument of DPRINTF macro to the console. Its only debug tool, so that isn't synchronized...
 *
 * @param msg The message.
 * @see printf
 * @see printf_variadic_list
 */
#define dprintf(msg, ...) \
{ \
 	printf ("Dprintf message in file %s, function %s, line %i:\n",__FILE__, __func__, __LINE__); \
	__builtin_va_list args;	\
	__builtin_va_start (args, msg);	\
	printf_variadic_list (msg, args); \
	__builtin_va_end (args);	\
}
 
#else /* NDEBUG */

#define assert(expr)            (void) 0
#define dprintf(msg, ...)       (void) 0

#endif /* NDEBUG */

#endif /* U_ASSERT_H */
