/**
 * @file io.c
 *
 * Userland IO.
 *
 */


#include <librt/io.h>
#include <librt.h>


char* err = "\nMALLOC FAILED, NOT ENOUGH MEMORY FOR PRINTF\n";


/**
 * Prints single character.
 *
 * @param chr
 *	Character to print.
 *
 * @return
 *	Number of printed characters.
 */
size_t putc (const char chr)
{
	return syscall (SYSCALL_PUTC, (unsigned long)&chr, 0, 0);	
}

/**
 * Prints character array - string.
 *
 * @param chr
 *	String to print.
 *
 * @return
 *	Number of printed characters.
 */
size_t puts (const char* str)
{
	return syscall (SYSCALL_PUTS, (unsigned long)str, 0, 0);
}


/**
 * Reads single character from keyboard.
 *
 * @return
 *	Read character.
 */
char getc (void)
{
	return syscall (SYSCALL_GETC, 0, 0, 0);
}


/**
 * Reads character array - string.
 *
 * @param str
 *	Buffer to fill with characters.
 *
 * @param len
 *	Number of characters to read.
 *
 * @return
 *	Number of read characters.
 */
ssize_t gets (char* str, const size_t len)
{
	if (len == 0)
		return EINVAL;

	for (size_t i = 0; i < len; ++i)
	{
		*(str+i) = getc();
		*(str+i+1) = '\0';
		if ( (*(str+i)) == '\n')
		{
			return i+1;
		}
	}
	return len;
}


/**
 * Convert unsigned integer to string.
 *
 * @param uint
 *	Unsigned integer to convert.
 *
 * @param str
 *	Buffer to fill with characters.
 *
 * @return
 *	Number of converted characters.
 */
size_t copy_u (uint32_t uint, char** str);
size_t copy_u (uint32_t uint, char** str)
{
	// printed chars
	size_t result = 0;

	// 32bit unsigned integer has max 10 chars as string
	char output [10];

	// index of array
	uint32_t loop1 = 0;

	// if zero, number seems to be consumed
	if (0 == uint)
	{
		// print zero
		(*(*str)++) = '0';
		
		// zero is one char width
		return 1;
	}

	// go until number is consumed
	while (uint)
	{
		// move last cipher to array
		output[loop1++] = uint % 10;
		
		// cut last cipher away
		uint /= 10;
	}

	// index means count of printed chars
	result = loop1--;

	// the number is reverted in array
	do
	{
		// print number as char
		(*(*str)++) = ('0' + output[loop1]);
	}
	// repeat until array printed
	while (0 != loop1--);

	// returns count of printed chars
	return result;
}


/**
 * Convert signed integer to string.
 *
 * @param sint
 *	Signed integer to convert.
 *
 * @param str
 *	Buffer to fill with characters.
 *
 * @return
 *	Number of converted characters.
 */
size_t copy_d (int32_t sint, char** str);
size_t copy_d (int32_t sint, char** str)
{
	// printed chars count
	uint32_t result = 0;

	// lt zero, print '-' and signed integer
	if (0 > sint)
	{
		// print '-' and count it
		(*(*str)++) = '-';
		result += 1;
		
		// revert negative integer to positive
		sint = -1 * sint;
		
		// print number and count it
		result += copy_u (sint, str);
	}
	else
	{
		// gt zero, print it as unsigned
		result += copy_u (sint, str);
	}

	// return count of printed chars
	return result;
}


/**
 * Convert hexadecimal integer to string.
 *
 * @param hint
 *	Hexadecimal integer to convert.
 *
 * @param str
 *	Buffer to fill with characters.
 *
 * @return
 *	Number of converted characters.
 */
size_t copy_h (uint32_t hint, char** str);
size_t copy_h (uint32_t hint, char** str)
{
	// printed chars
	size_t result = 0;

	// 32bit unsigned integer has max 8 chars as hexa string
	char output[8];

	// index of array
	uint32_t loop1 = 0;

	// if zero, number seems to be consumed
	if (0 == hint)
	{
		// print zero
		(*(*str)++) = '0';
	
		// zero is one char width
		return 1;
	}

	// go until number is consumed
	while (hint)
	{
		// move last cipher to array
		output[loop1++] = hint % 16;
		// cut last cipher away
		hint /= 16;
	}

	// index means count of printed chars
	result = loop1--;

	// the number is reverted in array
	do
	{
		// if lt 0xa, print as digit
		if (10 > output[loop1])
		{
			// print number as char
			(*(*str)++) = ('0' + output[loop1]);
		}
		else
		{
			// print hexachar
			(*(*str)++) = ('a' + output[loop1] - 10);
		}
	}
	// repeat until array printed
	while (0 != loop1--);

	// returns count of printed chars
	return result;
}



/**
 * Copy character to string.
 *
 * @param c
 *	Character to copy.
 *
 * @param str
 *	Buffer to fill with characters.
 *
 * @return
 *	Number of copied characters.
 */
size_t copy_c (const char c, char** str); 
size_t copy_c (const char c, char** str)
{
	(*(*str)++) = c;
	return 1;
} 



/**
 * Copy string to string.
 *
 * @param s
 *	String to copy.
 *
 * @param str
 *	Buffer to fill with characters.
 *
 * @return
 *	Number of printed characters.
 */
size_t copy_s (const char* s, char** str); 
size_t copy_s (const char* s, char** str)
{
	size_t result = 0;

	while (*s != '\0')
	{
		(*(*str)++) = *(s++);
		
		result++;
	}

	return result;
} 



/**
 * Computes length of string.
 *
 * @param str
 *	String to measure.
 *
 * @return
 *	Length of string.
 */
size_t length_s (const char* str);
size_t length_s (const char* str)
{
	size_t size = 0;
	
	while ('\0' != (*str++))
	{
		size++;
	}

	return size;
}



/**
 * Functions prints formated string. At first it counts number of characters the whole
 * formated string will need, later it allocates enough memory for the whole string 
 * and writes into it the formated string. It's the send via syscall to kernel.
 * This is necessary to ensure that the string is printed in one piece only.
 *
 * @param chr
 *	Character to print.
 *
 * @return
 *	Number of printed characters.
 */
size_t printf_variadic_list (const char* fst, __builtin_va_list args, const char* fst2, __builtin_va_list args2);
size_t printf_variadic_list (const char* fst, __builtin_va_list args, const char* fst2, __builtin_va_list args2)
{
	/* Computation of the result string size is not as correct as it could be
	 * because it uses only maximum estimates. */

	/* Number of printed characters. */ 
	uint32_t result = 0;
	
	/* Size of character array we will need. */
	size_t length = 1;
	
	/* Size of raw formated string. */
	size_t size = 0;
	
	while ('\0' != (*fst))
	{		
		if ('%' == *fst)
		{			
			fst++; size++;
			switch(*fst)
			{				
				case 'c':
					__builtin_va_arg (args, char *);
					length += 1;
					break;
				case 's':
					length += length_s (__builtin_va_arg (args, char *));
					break;
				case 'd':
					__builtin_va_arg (args, int32_t);
					length += 10;
					break;
				case 'u':
					__builtin_va_arg (args, uint32_t);
					length += 10;
					break;
				case 'i':
					__builtin_va_arg (args, int32_t);
					length += 11;
					break;
				case 'x':
					__builtin_va_arg (args, uint32_t);
					length += 8;
					break;
				case 'p':
					__builtin_va_arg (args, uint32_t);
					length += 10;
					break;
				case '%':
					length += 1;
					break;
				default:
					break;
			}
			fst++; size++;
		}
		else
		{
			fst++; size++;
			length += 1;
		}
	}
	fst = (fst-size);

	char* str = malloc (length);
	char* str_start = str;

	/* If there is not enough space print the string print at least warning. */
	if (str == NULL)
	{
		puts (err);
		return 0;
	}

        /* '\0' is not always 0 (depends on platform). */
	while ('\0' != (*fst))
	{
		/* If there is %, special char will follow. */
		if ('%' == *fst)
		{
			/* Skip that %. */
			fst++;
			/* Look at special char type. */
			switch (*fst)
			{
				/* Print char. */
				case 'c':
					result += copy_c (__builtin_va_arg (args2, int), &str);
					break;
				/* Print string. */
				case 's':
					result += copy_s (__builtin_va_arg (args2, char *), &str);
					break;
				/* Print signed integer. */
				case 'd':
					result += copy_d (__builtin_va_arg (args2, int32_t), &str);
					break;
				/* Print unsigned integer. */
				case 'u':
					result += copy_u (__builtin_va_arg (args2, uint32_t), &str);
					break;
				/* Print signed integer. */
				case 'i':
					result += copy_d (__builtin_va_arg (args2, int32_t), &str);
					break;
				/* Print unsigned integer as hexa. */
				case 'x':
					result += copy_h (__builtin_va_arg (args2, uint32_t), &str);
					break;
				/* Print pointer. */
				case 'p':
					(*str++) = '0'; (*str++) = 'x';
					result += 2  + copy_h (__builtin_va_arg (args2, uint32_t), &str);
					break;
				/* Print real %. */
				case '%':
					(*str++) = '%';
					result += 1; 
					break;
				/* Ignore other cases. */
				default:
					break;
			}
			/* Skip special char. */
			fst++;
		}
		/* Standard char. */
		else
		{
			/* Print char and count it. */
			*(str++) = *(fst++);
			result += 1;
			
		}
	}

	*str = '\0';
	puts (str_start);
	free (str_start);
	
	/* Returns count of printed chars. */
	return result;
}



/**
 * Prints formated string,
 *
 * @param fst
 *	String format.
 *
 * @param ...
 *	Arguments.
 *
 * @return
 *	Number of printed characters.
 */
size_t printf (const char * fst, ...)
{
	/* Count of printed chars. */
	size_t result;

	/* Take args to pass them into internal function. */
	__builtin_va_list args;
	__builtin_va_list args2;
	
	/* Initialize args list. */
	__builtin_va_start (args, fst);
	__builtin_va_start (args2, fst);

	/* Call internal function with args for this function. */
	result = printf_variadic_list (fst, args, fst, args2);

	/* clean up variadic args. */
	__builtin_va_end (args);
	__builtin_va_end (args2);
	
	/* Return count of printed chars. */
	return result;
}
