/**
 * @file mutex.c
 *
 * Mutex implementation for user-threads.
 *
 * The mutex operations in userspace is provided by syscalls. Every user mutex has his own internal kernel mutex structure 
 * managed by kernel. For every user mutex operation is checked, if the internal kernel mutex really exists (it means if it 
 * is valid operation) and then is called the related operation to its internal kernel mutex. In the other words, the user 
 * mutexes have the same funcionality as the kernel mutexes, but user mutexes are checked, if really exists.
 * @link kernel/sync/mutex.h Kernel mutex implementation @endlink
 */

#include <exc/syscall.h>
#include <include/error_codes.h>
#include <librt/malloc.h>
#include <librt/mutex.h>


/* 
 * Initialize user mutex structure to the unlocked state.
 */
int mutex_init (struct mutex* mtx)
{	
	/* If pointer isn't valid. */
	if(!mtx)
	{
		return EINVAL;
	}
	
	return syscall (SYSCALL_MUTEX_INIT, 
		(unsigned long) &(mtx->kernel_mutex), 
		(unsigned long) 0, 
		(unsigned long) 0);
}


/* 
 * Function performs cleaning the user mutex structure. 
 */
int mutex_destroy (struct mutex* mtx)
{
	/* If pointer isn't valid. */
	if(!mtx)
	{
		return EINVAL;
	}
	
	return syscall (SYSCALL_MUTEX_DESTROY, 
		(unsigned long) mtx->kernel_mutex, 
		(unsigned long) 0, 
		(unsigned long) 0);
}


/*
 * Fuction attempts to lock the user mutex. 
 */
int mutex_lock (struct mutex* mtx)
{
	/* If pointer isn't valid. */
	if(!mtx)
	{
		return EINVAL;
	}
	
	return syscall (SYSCALL_MUTEX_LOCK, 
		(unsigned long) mtx->kernel_mutex, 
		(unsigned long) 0, 
		(unsigned long) 0);
}


/*
 * Fuction attempts to lock the user mutex. 
 */
int mutex_lock_timeout (struct mutex* mtx, const unsigned int usec)
{
	/* If pointer isn't valid. */
	if(!mtx)
	{
		return EINVAL;
	}
	
	int result = syscall (SYSCALL_MUTEX_LOCK_TIMEOUT, 
		(unsigned long) mtx->kernel_mutex, 
		(unsigned long) usec, 
		(unsigned long) 0);
	
	return result;
}


/* 
 * The function unlocks the given user mutex.
 */
int mutex_unlock (struct mutex* mtx)
{
	/* If pointer isn't valid. */
	if(!mtx)
	{
		return EINVAL;
	}
	
	return syscall (SYSCALL_MUTEX_UNLOCK, 
		(unsigned long) mtx->kernel_mutex, 
		(unsigned long) 0, 
		(unsigned long) 0);
}
