/**
 * @file mutex.h
 *
 * Mutex implementation for user-threads.
 *
 * The mutex operations in userspace is provided by syscalls. Every user mutex has his own internal kernel mutex structure 
 * managed by kernel. For every user mutex operation is checked, if the internal kernel mutex really exists (it means if it 
 * is valid operation) and then is called the related operation to its internal kernel mutex. In the other words, the user 
 * mutexes have the same funcionality as the kernel mutexes, but user mutexes are checked, if really exists.
 * @link kernel/sync/mutex.h Kernel mutex implementation @endlink
 */


#ifndef U_MUTEX_H_
#define U_MUTEX_H_

#include <librt/thread.h>

/** Structure for user mutex. */
struct mutex
{
	/** Internal kernel mutex. */
	void* kernel_mutex;
};

/* All these functions are only the users handlers for syscalls commented only in this header file. */

/**
 * Initialize user mutex structure to the unlocked state.
 *
 * @param mtx Pointer to user mutex structure.
 * @return EINVAL if @mtx isn't valid pointer or was already initialized, ENOMEM if isn't enough memory, EOK else
 * @see mutex_destroy 
 */
int mutex_init (struct mutex* mtx);


/** 
 * Function performs cleaning the user mutex structure. 
 * If there are still some threads blocked on destroyed user mutex, cancel the current user thread.
 *
 * @param mtx Pointer to user mutex structure.
 * @return EINVAL if given @mtx wasn't initialized or if it isn't valid pointer, EOK else
 * @see mutex_init
 */
int mutex_destroy (struct mutex* mtx);


/**
 * Fuction attempts to lock the user mutex. 
 * If it is not possible, blocks the current thread. Mutex locking is allowed only for the threads from the same 
 * process as the owner of the mutex.
 *
 * @param mtx Pointer to user mutex structure.
 * @return EINVAL if given @mtx wasn't initialized or if it isn't valid pointer, or if the user mutex isn't accessible 
 *	for the current user thread, EOK else
 * @see mutex_lock_timeout
 * @see mutex_unlock 
 */
int mutex_lock (struct mutex* mtx);


/**
 * Fuction attempts to lock the user mutex. 
 * If it is not possible, blocks the current thread. Mutex locking is allowed only for the threads from the same process 
 * as the owner of the mutex. Variant _timeout waiting for locking user mutex usec microseconds long as possible. If 
 * timeout is 0, no threads to block.
 *
 * @param mtx Pointer to user mutex structure.
 * @param usec Time limit in microseconds.
 * @return EINVAL if given @mtx wasn't initialized or if it isn't valid pointer, or if the user mutex isn't accessible for 
 *	the current user thread, ETIMEDOUT if failed to lock in time limit, EOK else
 * @see mutex_lock
 * @see mutex_unlock
 */
int mutex_lock_timeout (struct mutex* mtx, const unsigned int usec);


/** 
 * The function unlocks the given user mutex.
 * If success, wakes up threads that were blocked to lock the user mutex. If the symbol DEBUG_MUTEX >= 1 defined, the mutex 
 * implementation makes sure that the unlocking user thread that it originally locked. If the mutex unlock another thread, 
 * cancel the current user thread.
 * 
 * @param mtx Pointer to mutex structure.
 * @return EINVAL if given @mtx wasn't initialized or if it isn't valid pointer, EOK else
 * @see mutex_lock
 * @see mutex_lock_timeout
 */
int mutex_unlock (struct mutex* mtx);


#endif