/**
 * @file thread.c
 *
 * User thread management.
 *
 * The thread operations in userspace is provided by syscalls. Every user thread has his own internal kernel thread 
 * structure managed by kernel. For every user thread operation is checked, if the internal kernel thread really exists
 * (it means if it is valid operation) and then is called the related operation to its internal kernel thread. 
 * In the other words, the user threads have the same funcionality as the kernel threads, but user threads are checked,
 * if really exists.
 * @link kernel/proc/thread.h Kernel thread management @endlink
 */

#include <exc/syscall.h>
#include <include/error_codes.h>
#include <librt/malloc.h>
#include <librt/thread.h>
#include <librt/io.h>


/*
 * Creates a new user thread. 
 */
int thread_create (thread_t* thread_ptr, void* (*thread_start)(void*), void* data)
{
	/* Pointer isn't valid. */
	if (!thread_ptr)
	{
		return EINVAL;
	}
	
	/* Allocation of structure. */
	*thread_ptr = malloc (sizeof (user_thread_struct_t));
	if((!(*thread_ptr)))
	{
		/* Not enough memory. */
		free ((*thread_ptr));
		return ENOMEM;
	}
	
	(*thread_ptr)->thread_function = thread_start;
	(*thread_ptr)->data = data;
	
	int result = syscall (SYSCALL_THREAD_CREATE, 
		(unsigned long) &((*thread_ptr)->kernel_thread), 
		(unsigned long) thread_start, 
		(unsigned long) data);
	
	/* Creation failed. */
	if(result != EOK)
	{
		free (*thread_ptr);
		*thread_ptr = NULL;
	}
	
	return result;
}


/*
 * Returns the current running user thread.
 */
thread_t thread_self (void)
{
	thread_t self_thread = malloc (sizeof (user_thread_struct_t)); 
	
	syscall (SYSCALL_THREAD_SELF, 
		(unsigned long) &(self_thread->kernel_thread), 
		(unsigned long) &(self_thread->thread_function), 
		(unsigned long) &(self_thread->data));
	
	return self_thread;
}


/*
 * The current thread is blocked until the thread @thr runs. 
 */
int thread_join (thread_t thr, void** thread_retval)
{
	/* Pointer isn't valid. */
	if(!thr)
	{
		return EINVAL;
	}
	
	/* Return value for the thread function of @thr. */
	int thr_retval = 0;
	
	int result = syscall (SYSCALL_THREAD_JOIN, 
		(unsigned long) thr->kernel_thread, 
		(unsigned long) &thr_retval,
		(unsigned long) 0);
	
	/* Return value of thread function. */
	if ((thread_retval) && (result == EOK))
	{
		*thread_retval = (void*) thr_retval;
	}
	
	return result;
}


/*
 * The current thread is blocked until the thread @thr runs. 
 */
int thread_join_timeout (thread_t thr, void** thread_retval, const unsigned int usec)
{
	/* Pointer isn't valid. */
	if(!thr)
	{
		return EINVAL;
	}
	
	/* Return value for the thread function of @thr. */
	int thr_retval = 0;
	
	int result = syscall (SYSCALL_THREAD_JOIN_TIMEOUT, 
		(unsigned long) thr->kernel_thread, 
		(unsigned long) usec,
		(unsigned long) &thr_retval);
	
	/* Return value of thread function. */
	if ((thread_retval) && (result == EOK))
	{
		*thread_retval = (void*) thr_retval;
	}
	
	return result;
}


/*
 * Detach the given user thread @thr.
 */
int thread_detach (thread_t thr)
{
	/* Pointer isn't valid. */
	if(!thr)
	{
		return EINVAL;
	}
	
	int result = syscall (SYSCALL_THREAD_DETACH, 
		(unsigned long) thr->kernel_thread, 
		(unsigned long) 0,
		(unsigned long) 0);
	
	return result;
}


/*
 * Cancel the given user thread @thr.
 */
int thread_cancel (thread_t thr)
{
	/* Pointer isn't valid. */
	if(!thr)
	{
		return EINVAL;
	}
	
	int result = syscall (SYSCALL_THREAD_CANCEL,
		(unsigned long) thr->kernel_thread, 
		(unsigned long) 0,
		(unsigned long) 0);
	
	return result;
}


/*
 * The current user thread is blocked for the given time @sec in seconds.
 */
void thread_sleep (const unsigned int sec)
{
	syscall (SYSCALL_THREAD_SLEEP, 
		(unsigned long) sec, 
		(unsigned long) 0,
		(unsigned long) 0);
}


/*
 * The current user thread is blocked for the given time @usec in microseconds.
 */
void thread_usleep (const unsigned int usec)
{
	syscall (SYSCALL_THREAD_USLEEP, 
		(unsigned long) usec, 
		(unsigned long) 0,
		(unsigned long) 0);
}


/*
 * The current thread is yielded and will be rescheduled.
 */
void thread_yield (void)
{
	syscall (SYSCALL_THREAD_YIELD, 
		(unsigned long) 0, 
		(unsigned long) 0,
		(unsigned long) 0);
}


/*
 * The current thread is blocked until the other thread call thread_wakeup.
 */
void thread_suspend (void)
{
	syscall (SYSCALL_THREAD_SUSPEND, 
		(unsigned long) 0, 
		(unsigned long) 0,
		(unsigned long) 0);
}


/*
 * The given user thread @thr will be unblocked (was suspended).
 */
int thread_wakeup (thread_t thr)
{
	/* Pointer isn't valid. */
	if(!thr)
	{
		return EINVAL;
	}
	
	int result = syscall (SYSCALL_THREAD_WAKEUP, 
		(unsigned long) thr->kernel_thread,
		(unsigned long) 0,
		(unsigned long) 0);
	
	return result;
}


/*
 * Terminates the execution of the calling thread.
 */
void thread_exit (void* thread_retval)
{
	syscall (SYSCALL_THREAD_EXIT, 
		(unsigned long) thread_retval,
		(unsigned long) 0,
		(unsigned long) 0);
}


/*
 * Terminates the execution of the current process.
 */
void exit (void)
{
	syscall (SYSCALL_EXIT,
		(unsigned long) 0,
		(unsigned long) 0,
		(unsigned long) 0);
}


/*
 * Finishing thread according to attached/detached etc.
 */
void thread_finish_over (const int thread_retval)
{
	syscall (SYSCALL_THREAD_FINISH_OVER, 
		(unsigned long) thread_retval, 
		(unsigned long) 0, 
		(unsigned long) 0);
}
