/**
 * @file thread.h
 *
 * User thread management.
 *
 * The thread operations in userspace is provided by syscalls. Every user thread has his own internal kernel thread 
 * structure managed by kernel. For every user thread operation is checked, if the internal kernel thread really exists
 * (it means if it is valid operation) and then is called the related operation to its internal kernel thread. 
 * In the other words, the user threads have the same funcionality as the kernel threads, but user threads are checked,
 * if really exists.
 * @link kernel/proc/thread.h Kernel thread management @endlink
 */


#ifndef U_THREAD_H_
#define U_THREAD_H_

/** Structure for user thread. */
typedef struct user_thread_struct
{
	/** Pointer to kernel thread (can not be accessed from userland). */
	void* kernel_thread;
	/** Thread executing function for user mode. */
	void* (*thread_function) (void*);
	/** Data for executing function. */
	void* data;
} user_thread_struct_t;

/* Only for better managing with pointers. */
typedef user_thread_struct_t* thread_t;


/* All these functions are only users handlers for syscalls commented only in this header file. */

/** 
 * Creates a new user thread. 
 * Function @thread_start will be called in new created user thread.
 *
 * @param thread_ptr Pointer to a thread structure to fill it by proper data.
 * @param thread_start Starting function for the user thread.
 * @param data Data for the starting function.
 * @return EINVAL if given @thread_ptr is bad pointer, ENOMEM if is not enough memory to create new user thread, EOK else
 * @see thread_finish_over 
 */
int thread_create (thread_t* thread_ptr, void* (*thread_start)(void*), void* data);


/**
 * Returns the current running user thread.
 * @return The current running user thread.
 */
thread_t thread_self (void);


/** 
 * The current thread is blocked until the thread @thr runs. 
 * If @thread_retval is a valid pointer, write down the pointer to the return value of the thread @thr to @thread_retval.
 *
 * @param thr Target thread for join.
 * @param thread_retval Return value of target thread.
 * @return EINVAL if @thr is not valid, if the @thr was detached, if any other thread waits for @thr or if calling thread 
 *	is the same as @thr, EOK else
 * @see thread_join_timeout
 */
int thread_join (thread_t thr, void** thread_retval);


/**
 * The current thread is blocked until the thread @thr runs. 
 * If @thread_retval is a valid pointer, write down the pointer to the return value of the thread @thr to @thread_retval.
 * Waits only for a @usec microseconds.
 *
 * @param thr Target thread for join.
 * @param thread_retval Return value of target thread.
 * @param usec The time limit for waiting.
 * @return EINVAL if @thr is not valid, if the @thr was detached, if any other thread waits for @thr or if calling thread 
 *	is the same as @thr, ETIMEDOUT if failed to join in time limit, EOK else
 * @see thread_join 
 */
int thread_join_timeout (thread_t thr, void** thread_retval, const unsigned int usec);


/**
 * Detach the given user thread @thr.
 *
 * @param thr The user thread for detaching.
 * @return EINVAL if @thr is not valid, @thr is already detached, @thr is waiting for join or if any thread is waiting 
 *	for @thr in thread_join, EOK else
 */
int thread_detach (thread_t thr);


/**
 * Cancel the given user thread @thr. 
 * The user thread, waiting for the cancellation the given @thr, will be unblocked.
 *
 * @param thr The user thread to cancellation
 * @return EINVAL if @thr is not valid, EOK else
 */
int thread_cancel (thread_t thr);


/**
 * The current user thread is blocked for the given time @sec in seconds.
 *
 * @param sec Time limit for sleeping (in seconds).
 * @see thread_usleep
 */
void thread_sleep (const unsigned int sec);


/**
 * The current user thread is blocked for the given time @usec in microseconds.
 *
 * @param usec Time limit for sleeping (in microseconds).
 * @see thread_sleep
 */
void thread_usleep (const unsigned int usec);


/**
 * The current thread is yielded and will be rescheduled.
 */
void thread_yield (void);


/**
 * The current thread is blocked until the other thread call thread_wakeup.
 * @see thread_wakeup
 */
void thread_suspend (void);


/**
 * The given user thread @thr will be unblocked (was suspended).
 * The calling user thread is running until are rescheduled.
 *
 * @param thr The user thread for waking up.
 * @return EINVAL if @thr is not valid, EOK else.
 * @see thread_suspend
 */
int thread_wakeup (thread_t thr);


/** 
 * Finishing thread according to attached/detached etc. 
 * It will free the structure of thread.
 *
 * @param thread_retval The return value of the thread function.
 * @see thread_create 
 */
void thread_finish_over (int thread_retval);


/**
 * Terminates the execution of the calling thread. 
 * Ensure access the pointer @thread_retval to the thread, which is calling thread_join.
 *
 * @param thread_retval Return value of the thread.
 * @see exit
 */
void thread_exit (void* thread_retval);


/**
 * Terminates the execution of the current process.
 * @see thread_exit
 */
void exit (void);


#endif